﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Dandelion
{
	/// <summary>
	/// Represents an implementation of network services.
	/// </summary>
	class HttpUtils
	{
		public static int LastResponseCode;
		public static string LastResponseMessage;

		/// <summary>
		/// Sends a HTTP request asynchronously.
		/// </summary>
		/// <param name="method"> The method for the request. </param>
		/// <param name="url"> A value containing the URI of the requested resource. </param>
		/// <param name="contentType"> The value of the Content-type HTTP header. </param>
		/// <param name="httpTimeout"> The value of the http request timeout . </param>
		/// <param name="headers"> Specifies a collection of the name/value pairs that make up the HTTP headers. </param>
		/// <param name="postData"> The value of the request data being sent. </param>
		public async static Task<string> SendHttpRequestAsync(string method,
										string url,
										string contentType,
										int httpTimeout,
										Dictionary<string, string> headers,
										string postData)
		{
			long sl = 0;
			string tmpTime = DateTime.Now.ToString("MMdd HH-mm-ss.fff");
			string logFileName = tmpTime + "_HttpUtils.txt";
			Logger tmpLogger = new Logger(Common.LogFolderPath + logFileName);
			tmpLogger.WriteNewLine("=====");

			LastResponseCode = Common.ResultSuccess;
			LastResponseMessage = string.Empty;

			tmpLogger.WriteNewLine("[SendHttpRequestAsync] url: " + url);
			if (string.IsNullOrEmpty(url))
			{
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] url is null or invalid ");
				LastResponseMessage = "url is null or invalid";
				LastResponseCode = (int)Common.HttpStatusCode.HttpStatusCodeValueInvalid;
				return null;
			}

			if (string.IsNullOrEmpty(postData))
			{
				postData = "";
			}

			string tmpPostData = postData;
			tmpLogger.WriteNewLine("[SendHttpRequestAsync] tmpPostData = " + tmpPostData);
			sl = DateTime.Now.Ticks;

			try
			{
				byte[] byteArray = Encoding.UTF8.GetBytes(tmpPostData);

				HttpWebRequest request = HttpWebRequest.Create(url) as HttpWebRequest;

				request.KeepAlive = false;
				request.Timeout = httpTimeout; //milliseconds
				request.ReadWriteTimeout = httpTimeout;
				request.Method = method;
				request.ContentType = contentType;

				if (headers != null)
				{
					for (int i = 0; i < headers.Count; i++)
					{
						KeyValuePair<string, string> kv = headers.ElementAt(i);
						request.Headers.Set(kv.Key, kv.Value);
						tmpLogger.WriteNewLine("[SendHttpRequestAsync] Headers data, key = " + kv.Key + ", value = " + kv.Value);
					}
				}

				request.ContentLength = byteArray.Length;
				Stream dataStream;

				if (byteArray.Length > 0)
				{
					dataStream = await request.GetRequestStreamAsync().ConfigureAwait(false);

					// Write the data to the request stream.
					await dataStream.WriteAsync(byteArray, 0, byteArray.Length).ConfigureAwait(false);

					// Close the Stream object.
					dataStream.Close();
				}

				WebResponse response = await request.GetResponseAsync().ConfigureAwait(false);

				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebRequest run time(s): " + (double)diffL / 10000000);

				// Get the stream containing content returned by the server.
				dataStream = response.GetResponseStream();

				// Open the stream using a StreamReader for easy access.
				StreamReader reader = new StreamReader(dataStream);

				// Read the content.
				string responseFromServer = await reader.ReadToEndAsync().ConfigureAwait(false);

				// Clean up the streams.
				reader.Close();
				dataStream.Close();
				response.Close();

				LastResponseCode = (int)((HttpWebResponse)response).StatusCode;
				LastResponseMessage = ((HttpWebResponse)response).StatusDescription;
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] StatusCode: " + LastResponseCode +
																	", StatusMessage: " + LastResponseMessage);
				return responseFromServer;
			}
			catch (WebException ex)
			{
				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException run time(s): " + (double)diffL / 10000000);
				if (ex.Status == WebExceptionStatus.ProtocolError)
				{
					var response = ex.Response as HttpWebResponse;
					if (response != null)
					{
						LastResponseCode = (int)response.StatusCode;
						LastResponseMessage = response.StatusDescription;

						var resp = new StreamReader(ex.Response.GetResponseStream()).ReadToEnd();
						tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException response: " + resp.ToString());
						tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException StatusCode: " + LastResponseCode);
						tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException StatusMessage: " + LastResponseMessage);
					}
					else
					{
						LastResponseCode = Common.ResultFail;
						LastResponseMessage = "The WebExceptionStatus is WebExceptionStatus.ProtocolError, but no messae";
						tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException No HTTP Description: " + ex.ToString());
					}
				}
				else
				{
					LastResponseCode = (int)AODCommon.DandelionStatusCode.ControllerStatusCodeInternetFailure;
					LastResponseMessage = "The WebExceptionStatus is " + ex.Status.ToString();
					tmpLogger.WriteNewLine("[SendHttpRequestAsync] WebException No HTTP Description: " + ex.ToString());
				}

				return null;
			}
			catch (Exception e)
			{
				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] Exception run time(s): " + (double)diffL / 10000000);
				LastResponseCode = Common.ResultFail;
				LastResponseMessage = "The Exception is not WebException";
				tmpLogger.WriteNewLine("[SendHttpRequestAsync] Exception Description: " + e.ToString());
				return null;
			}
		}


		/// <summary>
		/// Downloads the file asynchronously.
		/// </summary>
		/// <param name="method"> The method for the request. </param>
		/// <param name="url"> A value containing the URI of the requested resource. </param>
		/// <param name="contentType"> The value of the Content-type HTTP header. </param>
		/// <param name="httpTimeout"> The value of the http request timeout . </param>
		/// <param name="headers"> Specifies a collection of the name/value pairs that make up the HTTP headers. </param>
		/// <param name="postData"> The value of the request data being sent. </param>
		/// <param name="downloadPath"> A relative or absolute path for the file. </param>
		public async static Task<string> DownloadAsync(string method,
										string url,
										string contentType,
										int httpTimeout,
										Dictionary<string, string> headers,
										string postData,
										string downloadPath)
		{
			long sl = 0;
			string tmpTime = DateTime.Now.ToString("MMdd HH-mm-ss.fff");
			string logFileName = tmpTime + "_HttpUtils.txt";
			Logger tmpLogger = new Logger(Common.LogFolderPath + logFileName);
			tmpLogger.WriteNewLine("=====");

			LastResponseCode = Common.ResultSuccess;
			LastResponseMessage = string.Empty;

			tmpLogger.WriteNewLine("[DownloadAsync] url: " + url);
			if (string.IsNullOrEmpty(url))
			{
				tmpLogger.WriteNewLine("[DownloadAsync] url is null or invalid ");
				LastResponseMessage = "url is null or invalid";
				LastResponseCode = (int)Common.HttpStatusCode.HttpStatusCodeValueInvalid;
				return null;
			}

			if (string.IsNullOrEmpty(postData))
			{
				postData = "";
			}

			string tmpPostData = postData;
			tmpLogger.WriteNewLine("[DownloadAsync] tmpPostData = " + tmpPostData);
			sl = DateTime.Now.Ticks;

			try
			{
				byte[] tmpImageBytes;
				byte[] byteArray = Encoding.UTF8.GetBytes(tmpPostData);

				HttpWebRequest request = HttpWebRequest.Create(url) as HttpWebRequest;

				request.KeepAlive = false;
				request.Method = method;
				request.ContentType = contentType;

				if (headers != null)
				{
					for (int i = 0; i < headers.Count; i++)
					{
						KeyValuePair<string, string> kv = headers.ElementAt(i);
						request.Headers.Set(kv.Key, kv.Value);
						tmpLogger.WriteNewLine("[DownloadAsync] Headers data, key = " + kv.Key + ", value = " + kv.Value);
					}
				}

				request.ContentLength = byteArray.Length;
				Stream dataStream;

				if (byteArray.Length > 0)
				{
					dataStream = await request.GetRequestStreamAsync().ConfigureAwait(false);

					// Write the data to the request stream.
					await dataStream.WriteAsync(byteArray, 0, byteArray.Length).ConfigureAwait(false);

					// Close the Stream object.
					dataStream.Close();
				}

				request.Timeout = httpTimeout; //milliseconds
				request.ReadWriteTimeout = httpTimeout;

				WebResponse response = await request.GetResponseAsync().ConfigureAwait(false);

				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[DownloadAsync] WebRequest run time(s): " + (double)diffL / 10000000);

				// Get the stream containing content returned by the server.
				dataStream = response.GetResponseStream();

				if (downloadPath != null && downloadPath.Length > 0)
				{
					using (BinaryReader br = new BinaryReader(dataStream))
					{
						tmpImageBytes = br.ReadBytes(500000);
						br.Close();
					}

					dataStream.Close();
					response.Close();

					FileStream fs = new FileStream(downloadPath, FileMode.Create);
					BinaryWriter bw = new BinaryWriter(fs);
					try
					{
						bw.Write(tmpImageBytes);
					}
					catch (Exception e)
					{
						fs.Close();
						bw.Close();
						tmpLogger.WriteNewLine("[DownloadAsync] BinaryWriter Write Exception: " + e.ToString());
						return null;
					}

					fs.Close();
					bw.Close();

					return downloadPath;
				}
				return null;
			}
			catch (WebException ex)
			{
				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[DownloadAsync] WebException run time(s): " + (double)diffL / 10000000);
				if (ex.Status == WebExceptionStatus.ProtocolError)
				{
					var response = ex.Response as HttpWebResponse;
					if (response != null)
					{
						LastResponseCode = (int)response.StatusCode;
						LastResponseMessage = response.StatusDescription;
						tmpLogger.WriteNewLine("[DownloadAsync] WebException StatusCode: " + LastResponseCode);
						tmpLogger.WriteNewLine("[DownloadAsync] WebException StatusMessage: " + LastResponseMessage);
					}
					else
					{
						LastResponseCode = Common.ResultFail;
						LastResponseMessage = "The WebExceptionStatus is WebExceptionStatus.ProtocolError, but no messae";
						tmpLogger.WriteNewLine("[DownloadAsync] WebException No HTTP Description: " + ex.ToString());
					}
				}
				else if (ex.Status == WebExceptionStatus.Timeout)
				{
					LastResponseCode = (int)AODCommon.DandelionStatusCode.ControllerStatusCodeTimeout;
					LastResponseMessage = "The WebExceptionStatus is " + ex.Status.ToString();
				}
				else
				{
					LastResponseCode = (int)AODCommon.DandelionStatusCode.ControllerStatusCodeInternetFailure;
					LastResponseMessage = "The WebExceptionStatus is " + ex.Status.ToString();
					tmpLogger.WriteNewLine("[DownloadAsync] WebException No HTTP Description: " + ex.ToString());
				}

				return null;
			}
			catch (Exception e)
			{
				long diffL = DateTime.Now.Ticks - sl;
				tmpLogger.WriteNewLine("[DownloadAsync] Exception run time(s): " + (double)diffL / 10000000);
				LastResponseCode = Common.ResultFail;
				LastResponseMessage = "The Exception is not WebException";
				tmpLogger.WriteNewLine("[DownloadAsync] Exception Description: " + e.ToString());
				return null;
			}
		}
	}
}
