﻿using AirFusion;
using System;
using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Threading;

namespace WPFMediaKit.DirectShow.Controls
{
	/// <summary>
	/// The class provides basic functionality needed
	/// to render a D3D surface. This class is abstract.
	/// </summary>
	public abstract class D3DRenderer : FrameworkElement
	{
		#region Enum Definition

		/// <summary>
		/// This type specifies the current rendering source on the WPF Image
		/// </summary>
		private enum D3DImageRenderingSourceType
		{
			D3DImageRenderingSourceTypeDefault,
			D3DImageRenderingSourceTypeSecondary
		};

		#endregion Enum Definition

		#region Members

		/// <summary>
		/// The Primary D3DImage used to render video
		/// </summary>
		private D3DImage m_d3dImagePrimary;

		/// <summary>
		/// The Secondary D3DImage used to render video
		/// </summary>
		private D3DImage m_d3dImageSecondary;

		/// <summary>
		/// The Image control that has the source
		/// to the D3DImage
		/// </summary>
		private Image m_videoImage;

		/// <summary>
		/// We keep reference to the D3D surface so
		/// we can delay loading it to avoid a black flicker
		/// when loading new media
		/// </summary>
		private IntPtr m_pBackBuffer = IntPtr.Zero;

		/// <summary>
		/// Flag to tell us if we have a new D3D
		/// Surface available
		/// </summary>
		private bool m_newSurfaceAvailable;

		/// <summary>
		/// A weak reference of D3DRenderers that have been cloned
		/// </summary>
		private readonly List<WeakReference> m_clonedD3Drenderers = new List<WeakReference>();

		/// <summary>
		/// Backing field for the RenderOnCompositionTargetRendering flag.
		/// </summary>
		private bool m_renderOnCompositionTargetRendering;

		/// <summary>
		/// Temporary storage for the RenderOnCompositionTargetRendering flag.
		/// This is used to remember the value for when the control is loaded and unloaded.
		/// </summary>
		private bool m_renderOnCompositionTargetRenderingTemp;

		/// <summary>
		/// The current source type of WPF Image
		/// </summary>
		private D3DImageRenderingSourceType m_currentD3DImageRenderingSourceType;

		#endregion Members

		#region Dependency Properties

		#region Stretch

		public static readonly DependencyProperty StretchProperty =
			DependencyProperty.Register("Stretch", typeof(Stretch), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(Stretch.Uniform,
					new PropertyChangedCallback(OnStretchChanged)));

		/// <summary>
		/// Defines what rules are applied to the stretching of the video
		/// </summary>
		public Stretch Stretch
		{
			get { return (Stretch)GetValue(StretchProperty); }
			set { SetValue(StretchProperty, value); }
		}

		private static void OnStretchChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			((D3DRenderer)d).OnStretchChanged(e);
		}

		private void OnStretchChanged(DependencyPropertyChangedEventArgs e)
		{
			m_videoImage.Stretch = (Stretch)e.NewValue;
		}

		#endregion Stretch

		#region StretchDirection

		public static readonly DependencyProperty StretchDirectionProperty =
			DependencyProperty.Register("StretchDirection", typeof(StretchDirection), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(StretchDirection.Both,
					new PropertyChangedCallback(OnStretchDirectionChanged)));

		/// <summary>
		/// Gets or Sets the value that indicates how the video is scaled.  This is a dependency property.
		/// </summary>
		public StretchDirection StretchDirection
		{
			get { return (StretchDirection)GetValue(StretchDirectionProperty); }
			set { SetValue(StretchDirectionProperty, value); }
		}

		private static void OnStretchDirectionChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			((D3DRenderer)d).OnStretchDirectionChanged(e);
		}

		protected virtual void OnStretchDirectionChanged(DependencyPropertyChangedEventArgs e)
		{
			m_videoImage.StretchDirection = (StretchDirection)e.NewValue;
		}

		#endregion StretchDirection

		#region IsRenderingEnabled

		public static readonly DependencyProperty IsRenderingEnabledProperty =
			DependencyProperty.Register("IsRenderingEnabled", typeof(bool), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(true));

		/// <summary>
		/// Enables or disables rendering of the video
		/// </summary>
		public bool IsRenderingEnabled
		{
			get { return (bool)GetValue(IsRenderingEnabledProperty); }
			set { SetValue(IsRenderingEnabledProperty, value); }
		}

		#endregion IsRenderingEnabled

		#region NaturalVideoHeight

		private static readonly DependencyPropertyKey NaturalVideoHeightPropertyKey
			= DependencyProperty.RegisterReadOnly("NaturalVideoHeight", typeof(int), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(0));

		public static readonly DependencyProperty NaturalVideoHeightProperty
			= NaturalVideoHeightPropertyKey.DependencyProperty;

		/// <summary>
		/// Gets the natural pixel height of the current media.
		/// The value will be 0 if there is no video in the media.
		/// </summary>
		public int NaturalVideoHeight
		{
			get { return (int)GetValue(NaturalVideoHeightProperty); }
		}

		/// <summary>
		/// Internal method to set the read-only NaturalVideoHeight DP
		/// </summary>
		protected void SetNaturalVideoHeight(int value)
		{
			SetValue(NaturalVideoHeightPropertyKey, value);
		}

		#endregion NaturalVideoHeight

		#region NaturalVideoWidth

		private static readonly DependencyPropertyKey NaturalVideoWidthPropertyKey
			= DependencyProperty.RegisterReadOnly("NaturalVideoWidth", typeof(int), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(0));

		public static readonly DependencyProperty NaturalVideoWidthProperty
			= NaturalVideoWidthPropertyKey.DependencyProperty;

		/// <summary>
		/// Gets the natural pixel width of the current media.
		/// The value will be 0 if there is no video in the media.
		/// </summary>
		public int NaturalVideoWidth
		{
			get { return (int)GetValue(NaturalVideoWidthProperty); }
		}

		/// <summary>
		/// Internal method to set the read-only NaturalVideoWidth DP
		/// </summary>
		protected void SetNaturalVideoWidth(int value)
		{
			SetValue(NaturalVideoWidthPropertyKey, value);
		}

		#endregion NaturalVideoWidth

		#region HasVideo

		private static readonly DependencyPropertyKey HasVideoPropertyKey
			= DependencyProperty.RegisterReadOnly("HasVideo", typeof(bool), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(false));

		public static readonly DependencyProperty HasVideoProperty
			= HasVideoPropertyKey.DependencyProperty;

		/// <summary>
		/// Is true if the media contains renderable video
		/// </summary>
		public bool HasVideo
		{
			get { return (bool)GetValue(HasVideoProperty); }
		}

		/// <summary>
		/// Internal method for setting the read-only HasVideo DP
		/// </summary>
		protected void SetHasVideo(bool value)
		{
			SetValue(HasVideoPropertyKey, value);
		}

		#endregion HasVideo

		#region DeeperColor

		public static readonly DependencyProperty DeeperColorProperty =
			DependencyProperty.Register("DeeperColor", typeof(bool), typeof(D3DRenderer),
				new FrameworkPropertyMetadata(false,
					new PropertyChangedCallback(OnDeeperColorChanged)));

		public bool DeeperColor
		{
			get { return (bool)GetValue(DeeperColorProperty); }
			set { SetValue(DeeperColorProperty, value); }
		}

		private static void OnDeeperColorChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
		{
			((D3DRenderer)d).OnDeeperColorChanged(e);
		}

		protected virtual void OnDeeperColorChanged(DependencyPropertyChangedEventArgs e)
		{
			ToggleDeeperColorEffect((bool)e.NewValue);
		}

		#endregion DeeperColor

		#endregion Dependency Properties

		#region Constructors & Dispose

		protected D3DRenderer()
		{
			InitializeD3DVideo();

			/* Hook into the framework events */
			Loaded += D3DRendererLoaded;
			Unloaded += D3DRendererUnloaded;
		}

		#endregion Constructors & Dispose

		#region Private Methods

		/// <summary>
		/// Effects the image to be deep color
		/// </summary>
		private void ToggleDeeperColorEffect(bool isEnabled)
		{
			//m_videoImage.Effect = isEnabled ? new DeeperColorEffect() : null;
		}

		/// <summary>
		/// Handler for when the D3DRenderer is unloaded
		/// </summary>
		private void D3DRendererUnloaded(object sender, RoutedEventArgs e)
		{
			/* Remember what the property value was */
			m_renderOnCompositionTargetRenderingTemp = RenderOnCompositionTargetRendering;

			/* Make sure to unhook the static event hook because we are unloading */
			RenderOnCompositionTargetRendering = false;
		}

		/// <summary>
		/// Handler for when the D3DRenderer is loaded
		/// </summary>
		private void D3DRendererLoaded(object sender, RoutedEventArgs e)
		{
			/* Restore the property's value */
			RenderOnCompositionTargetRendering = m_renderOnCompositionTargetRenderingTemp;
		}

		/// <summary>
		/// Initializes the D3DRenderer control
		/// </summary>
		private void InitializeD3DVideo()
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Initializing D3D Renderer", "");

			if (m_videoImage != null)
				return;

			/* Initial rendering source type*/
			this.m_currentD3DImageRenderingSourceType = D3DImageRenderingSourceType.D3DImageRenderingSourceTypeDefault;

			/* Create our Image and it's D3DImage source */
			this.m_videoImage = new Image();
			this.m_d3dImagePrimary = new D3DImage();
			this.m_d3dImageSecondary = new D3DImage();

			/* Set our default stretch value of our video */
			m_videoImage.Stretch = (Stretch)StretchProperty.DefaultMetadata.DefaultValue;
			m_videoImage.StretchDirection = (StretchDirection)StretchProperty.DefaultMetadata.DefaultValue;

			/* Our source of the video image is the D3DImage */
			m_videoImage.Source = this.m_d3dImagePrimary;

			/* Register the Image as a visual child */
			AddVisualChild(m_videoImage);

			/* Bind the horizontal alignment dp of this control to that of the video image */
			var horizontalAlignmentBinding = new Binding("HorizontalAlignment") { Source = this };
			m_videoImage.SetBinding(HorizontalAlignmentProperty, horizontalAlignmentBinding);

			/* Bind the vertical alignment dp of this control to that of the video image */
			var verticalAlignmentBinding = new Binding("VerticalAlignment") { Source = this };
			m_videoImage.SetBinding(VerticalAlignmentProperty, verticalAlignmentBinding);

			ToggleDeeperColorEffect((bool)DeeperColorProperty.DefaultMetadata.DefaultValue);
		}

		/// <summary>
		/// This should only fire when a D3D device is lost
		/// </summary>
		private void D3DImageIsFrontBufferAvailableChanged(object sender, DependencyPropertyChangedEventArgs e)
		{
			if (!this.m_d3dImagePrimary.IsFrontBufferAvailable || !this.m_d3dImageSecondary.IsFrontBufferAvailable)
				return;

			/* Flag that we have a new surface, even
			 * though we really don't */
			m_newSurfaceAvailable = true;

			/* Force feed the D3DImage the Surface pointer */
			SetBackBufferInternal(m_pBackBuffer);
		}

		/// <summary>
		/// Renders image to composition target
		/// </summary>
		private void CompositionTargetRendering(object sender, EventArgs e)
		{
			InternalInvalidateVideoImage();
		}

		/// <summary>
		/// Cleans up any dead references we may have to any cloned renderers
		/// </summary>
		private void CleanZombieRenderers()
		{
			lock (m_clonedD3Drenderers)
			{
				var deadObjects = new List<WeakReference>();

				for (int i = 0; i < m_clonedD3Drenderers.Count; i++)
				{
					if (!m_clonedD3Drenderers[i].IsAlive)
						deadObjects.Add(m_clonedD3Drenderers[i]);
				}

				foreach (var deadGuy in deadObjects)
				{
					m_clonedD3Drenderers.Remove(deadGuy);
				}
			}
		}

		/// <summary>
		/// Sets the backbuffer for any cloned D3DRenderers
		/// </summary>
		private void SetBackBufferForClones()
		{
			lock (m_clonedD3Drenderers)
			{
				CleanZombieRenderers();

				foreach (var rendererRef in m_clonedD3Drenderers)
				{
					var renderer = rendererRef.Target as D3DRenderer;
					if (renderer != null)
						renderer.SetBackBuffer(m_pBackBuffer);
				}
			}
		}

		/// <summary>
		/// Configures D3DImage with a new surface.  This happens immediately
		/// </summary>
		private void SetBackBufferInternal(IntPtr backBuffer)
		{
			/* Do nothing if we don't have a new surface available */
			if (!m_newSurfaceAvailable || (backBuffer == IntPtr.Zero))
				return;

			if (!this.m_d3dImagePrimary.Dispatcher.CheckAccess())
			{
				this.m_d3dImagePrimary.Dispatcher.BeginInvoke((Action)(() => SetBackBufferInternal(backBuffer)));
				return;
			}

			if (!this.m_d3dImageSecondary.Dispatcher.CheckAccess())
			{
				this.m_d3dImageSecondary.Dispatcher.BeginInvoke((Action)(() => SetBackBufferInternal(backBuffer)));
				return;
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Setting back buffer", "");

			/* We have this around a try/catch just in case we
			 * lose the device and our Surface is invalid. The
			 * try/catch may not be needed, but testing needs
			 * to take place before it's removed */
			try
			{
				this.m_d3dImagePrimary.Lock();
				this.m_d3dImagePrimary.SetBackBuffer(D3DResourceType.IDirect3DSurface9, backBuffer);
				this.m_d3dImagePrimary.Unlock();
				this.m_d3dImageSecondary.Lock();
				this.m_d3dImageSecondary.SetBackBuffer(D3DResourceType.IDirect3DSurface9, backBuffer);
				this.m_d3dImageSecondary.Unlock();
				SetNaturalWidthHeight();
			}
			catch
			{
			}

			/* Clear our flag, so this won't be ran again
			 * until a new surface is sent */
			m_newSurfaceAvailable = false;
		}

		/// <summary>
		/// Sets the natual width and height of D3D image
		/// </summary>
		private void SetNaturalWidthHeight()
		{
			SetNaturalVideoHeight(this.m_d3dImagePrimary.PixelHeight);
			SetNaturalVideoWidth(this.m_d3dImagePrimary.PixelWidth);
		}

		/// <summary>
		/// Invalidates any possible cloned renderer we may have
		/// </summary>
		private void InvalidateClonedVideoImages()
		{
			lock (m_clonedD3Drenderers)
			{
				CleanZombieRenderers();

				foreach (var rendererRef in m_clonedD3Drenderers)
				{
					var renderer = rendererRef.Target as D3DRenderer;
					if (renderer != null)
						renderer.InvalidateVideoImage();
				}
			}
		}

		/// <summary>
		/// Update D3dImage front buffer from back buffer
		/// </summary>
		private void UpdateD3DImageBufferStatus()
		{
			if (this.m_currentD3DImageRenderingSourceType == D3DImageRenderingSourceType.D3DImageRenderingSourceTypeDefault)
			{
				if (this.m_d3dImageSecondary.IsFrontBufferAvailable && this.m_pBackBuffer != IntPtr.Zero)
				{
					try
					{
						this.m_d3dImageSecondary.Lock();
						this.m_d3dImageSecondary.AddDirtyRect(new Int32Rect(0, 0, this.m_d3dImageSecondary.PixelWidth, this.m_d3dImageSecondary.PixelHeight));
						this.m_d3dImageSecondary.Unlock();
						this.m_videoImage.Source = this.m_d3dImageSecondary;
						this.m_currentD3DImageRenderingSourceType = D3DImageRenderingSourceType.D3DImageRenderingSourceTypeSecondary;
					}
					catch (Exception)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Error updating D3DImage buffer!!!", "");
					}
				}
			}
			else
			{
				if (this.m_d3dImagePrimary.IsFrontBufferAvailable && this.m_pBackBuffer != IntPtr.Zero)
				{
					try
					{
						this.m_d3dImagePrimary.Lock();
						this.m_d3dImagePrimary.AddDirtyRect(new Int32Rect(0, 0, this.m_d3dImagePrimary.PixelWidth, this.m_d3dImagePrimary.PixelHeight));
						this.m_d3dImagePrimary.Unlock();
						this.m_videoImage.Source = this.m_d3dImagePrimary;
						this.m_currentD3DImageRenderingSourceType = D3DImageRenderingSourceType.D3DImageRenderingSourceTypeDefault;
					}
					catch (Exception)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Error updating D3dImage buffer!!!", "");
					}
				}
			}
		}

		/// <summary>
		/// Convert D3DImage to Bitmap.
		/// </summary>
		/// <param name="inputD3DImage">
		/// The D3DImage for convertion.
		/// </param>
		/// <returns>
		/// The bitmap converted from the input D3DImage.
		/// </returns>
		private System.Drawing.Bitmap ConvertD3DImageToBitmap(D3DImage inputD3DImage)
		{
			DrawingVisual drawingVisual = new DrawingVisual();
			DrawingContext drawingContext = drawingVisual.RenderOpen();
			drawingContext.DrawImage(inputD3DImage, new Rect(0, 0, inputD3DImage.PixelWidth, inputD3DImage.PixelHeight));
			drawingContext.Close();
			RenderTargetBitmap renderTargetBitmap = new RenderTargetBitmap(inputD3DImage.PixelWidth, inputD3DImage.PixelHeight, 96, 96, PixelFormats.Default);
			renderTargetBitmap.Render(drawingVisual);
			renderTargetBitmap.Freeze();

			System.Drawing.Bitmap outputBitmap;
			using (MemoryStream memoryStream = new MemoryStream())
			{
				System.Drawing.Bitmap tempBitmap;
				BitmapEncoder bitmapEncoder = new BmpBitmapEncoder();
				bitmapEncoder.Frames.Add(BitmapFrame.Create(renderTargetBitmap));
				bitmapEncoder.Save(memoryStream);
				tempBitmap = new System.Drawing.Bitmap(memoryStream);
				outputBitmap = new System.Drawing.Bitmap(tempBitmap);
			}

			return outputBitmap;
		}

		/// <summary>
		/// Used as a clone for a D3DRenderer
		/// </summary>
		private class ClonedD3DRenderer : D3DRenderer { }

		#endregion Private Methods

		#region Protected Methods

		protected override Size MeasureOverride(Size availableSize)
		{
			m_videoImage.Measure(availableSize);
			return m_videoImage.DesiredSize;
		}

		protected override Size ArrangeOverride(Size finalSize)
		{
			m_videoImage.Arrange(new Rect(finalSize));
			return finalSize;
		}

		protected override int VisualChildrenCount
		{
			get
			{
				return 1;
			}
		}

		protected override Visual GetVisualChild(int index)
		{
			if (index > 0)
				throw new IndexOutOfRangeException();

			return m_videoImage;
		}

		protected Image VideoImage
		{
			get
			{
				return m_videoImage;
			}
		}

		protected short CountOfFrame = 0;       // The count of frame

		private bool _isCanRenderred = true;    // The flag for starting renderring

		protected bool IsCanRenderred
		{
			get
			{
				return _isCanRenderred;
			}

			set
			{
				if (_isCanRenderred != value)
				{
					_isCanRenderred = value;
				}
			}
		}

		/// <summary>
		/// Renders the video with WPF's rendering using the CompositionTarget.Rendering event
		/// </summary>
		protected bool RenderOnCompositionTargetRendering
		{
			get
			{
				return m_renderOnCompositionTargetRendering;
			}
			set
			{
				/* If it is being set to true and it was previously false
				 * then hook into the event */
				if (value && !m_renderOnCompositionTargetRendering)
					CompositionTarget.Rendering += CompositionTargetRendering;
				else if (!value)
					CompositionTarget.Rendering -= CompositionTargetRendering;

				m_renderOnCompositionTargetRendering = value;
				m_renderOnCompositionTargetRenderingTemp = value;
			}
		}

		/// <summary>
		/// Configures D3DImage with a new surface.  The back buffer is
		/// not set until we actually receive a frame, this way we
		/// can avoid a black flicker between media changes
		/// </summary>
		/// <param name="backBuffer">The unmanaged pointer to the Direct3D Surface</param>
		protected void SetBackBuffer(IntPtr backBuffer)
		{
			/* We only do this if target rendering is enabled because we must use an Invoke
			 * instead of a BeginInvoke to keep the Surfaces in sync and Invoke could be dangerous
			 * in other situations */
			if (RenderOnCompositionTargetRendering)
			{
				if (!this.m_d3dImagePrimary.Dispatcher.CheckAccess())
				{
					this.m_d3dImagePrimary.Dispatcher.Invoke((Action)(() => SetBackBuffer(backBuffer)), DispatcherPriority.Render);
					return;
				}

				if (!this.m_d3dImageSecondary.Dispatcher.CheckAccess())
				{
					this.m_d3dImageSecondary.Dispatcher.Invoke((Action)(() => SetBackBuffer(backBuffer)), DispatcherPriority.Render);
					return;
				}
			}

			/* Flag a new surface */
			m_newSurfaceAvailable = true;
			m_pBackBuffer = backBuffer;

			/* Make a special case for target rendering */
			if (RenderOnCompositionTargetRendering || m_pBackBuffer == IntPtr.Zero)
			{
				SetBackBufferInternal(m_pBackBuffer);
			}

			SetBackBufferForClones();
		}

		/// <summary>
		/// Invalidates the video image
		/// </summary>
		protected void InvalidateVideoImage()
		{
			if (!m_renderOnCompositionTargetRendering)
				InternalInvalidateVideoImage();
		}

		/// <summary>
		/// Invalidates the entire Direct3D image, notifying WPF to redraw
		/// </summary>
		protected void InternalInvalidateVideoImage()
		{
			/* Ensure we run on the correct Dispatcher */
			if (!this.m_d3dImagePrimary.Dispatcher.CheckAccess())
			{
				this.m_d3dImagePrimary.Dispatcher.Invoke((Action)(() => InvalidateVideoImage()));
				return;
			}

			if (!this.m_d3dImageSecondary.Dispatcher.CheckAccess())
			{
				this.m_d3dImageSecondary.Dispatcher.Invoke((Action)(() => InvalidateVideoImage()));
				return;
			}

			this.SetBackBufferInternal(this.m_pBackBuffer);
			this.UpdateD3DImageBufferStatus();

			/* Invalidate all of our cloned D3DRenderers */
			InvalidateClonedVideoImages();
		}

		/// <summary>
		/// Snap the current D3DImage and save it to local file as PNG.
		/// </summary>
		/// <param name="outputFolder">
		/// The ouput folder of the snapshot.
		/// </param>
		/// <param name="rotateFlipType">
		/// The rotation type for current D3DImage.
		/// </param>
		/// <returns>
		/// The absolute path of the created snapshot.
		/// </returns>
		protected string SnapCurrentD3DImage(string outputFolder, System.Drawing.RotateFlipType rotateFlipType)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Snap current D3DImage", "");
			try
			{
				if (!Directory.Exists(outputFolder))
				{
					Directory.CreateDirectory(outputFolder);
				}

				System.Drawing.Bitmap bitmap;
				if (this.m_currentD3DImageRenderingSourceType == D3DImageRenderingSourceType.D3DImageRenderingSourceTypeDefault)
				{
					bitmap = this.ConvertD3DImageToBitmap(this.m_d3dImagePrimary);
				}
				else
				{
					bitmap = this.ConvertD3DImageToBitmap(this.m_d3dImageSecondary);
				}

				bitmap.RotateFlip(rotateFlipType);
				string outputFilePath = Path.Combine(outputFolder, string.Format("{0}.png", DateTime.Now.ToString("yyyyMMddHHmmssfff")));
				bitmap.Save(outputFilePath, System.Drawing.Imaging.ImageFormat.Png);

				return outputFilePath;
			}
			catch (Exception exception)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, exception.ToString(), "");
				return string.Empty;
			}
		}

		/// <summary>
		/// Hook/Un-hook <c>IsFrontBufferAvailableChanged</c> event.
		/// </summary>
		/// <param name="isHooked">Is the event hooked.</param>
		protected void SetIsFrontBufferAvailableChangedEventStatus(bool isHooked)
		{
			if (isHooked)
			{
				this.m_d3dImagePrimary.IsFrontBufferAvailableChanged += D3DImageIsFrontBufferAvailableChanged;
				this.m_d3dImageSecondary.IsFrontBufferAvailableChanged += D3DImageIsFrontBufferAvailableChanged;
			}
			else
			{
				this.m_d3dImagePrimary.IsFrontBufferAvailableChanged -= D3DImageIsFrontBufferAvailableChanged;
				this.m_d3dImageSecondary.IsFrontBufferAvailableChanged -= D3DImageIsFrontBufferAvailableChanged;
			}
		}

		#endregion Protected Methods

		#region Public Methods

		/// <summary>
		/// Creates a clone of the D3DRenderer.  This is a work for the visual
		/// brush not working cross-threaded
		/// </summary>
		/// <returns></returns>
		public D3DRenderer CloneD3DRenderer()
		{
			var renderer = new ClonedD3DRenderer();

			lock (m_clonedD3Drenderers)
			{
				m_clonedD3Drenderers.Add(new WeakReference(renderer));
			}

			renderer.SetBackBuffer(m_pBackBuffer);
			return renderer;
		}

		#endregion Public Methods
	}
}