﻿using NamedPipeWrapper;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Common
{
    /// <summary>
    /// Provides the functionality of IPC, inter process communication, with object level via named pipe.
    /// User can register a "delegate object" which will be invoked by users in other processes.
    /// </summary>
    /// <remarks>
    /// The returned and parameter types of the delegate object must be serializable.
    /// </remarks>
    /// <example>
    /// Usage:
    ///     1. Declare delegate type and make sure both caller(client side) and callee(server side) able to reference it.
    ///        For example:
    ///             <code>public delegate int Add(int a,int b);</code>
    ///
    ///     2. Callee(server side)
    ///        Create a instance of Add()
    ///             <code>public int AddInstance(int a, int b){ return a + b; }
    ///             Add add=new Add(AddInstance);</code>
    ///
    ///        Regist "add" object with an unique name
    ///             <code>DelegatePipeManager.RegistDelegate("AddIpc", add);</code>
    ///
    ///     3. Caller(client side)
    ///        Call "add" by the registed name and the declared type with its parameters if any:
    ///             <code>int answer = DelegatePipeManager.Call&ltint&gt("AddIpc", typeof(Add),123,456);</int>
    ///        The returned value will be same as the value from:
    ///             <code>int answer = AddInstance(123,456);</code>
    ///
    ///        Be noted that Call() may timeout because of some reasons which could be:
    ///                     a. it's a time consuming job and remote instance can't finish it in default timeout(5000ms).
    ///                     b. it wasn't registered, or it was but isn't in service yet.
    ///                     c. PC is too busy and with a few resource
    ///                 Here are two ways as advise:
    ///                     a. (Recommend) Use CallWithTimeout() instead and assign a larger value of the timeout(default=5000ms)
    ///                        <code>int answer = DelegatePipeManager.CallWithTimeout&ltint&gt("AddIpc", typeof(Add),1000*60,123,456);</code>
    ///                     b. sleep a while before executeing Call() again
    /// </example>
    public static class DelegatePipeManager
    {
        // store opened clients after it was been invoked first time
        private static object _openedClientMapLock = new object();

        private static SortedDictionary<String, object> _openedClientLockMap = new SortedDictionary<String, object>();
        private static SortedDictionary<String, NamedPipeClient<PipeData, PipeData>> _openedClientMap = new SortedDictionary<String, NamedPipeClient<PipeData, PipeData>>();

        // 1.1 manager store all request data generated once client invokes Call()
        // 1.2 manager let the invoking thread to wait, or said Monitor.Wait(), for server's responce
        // 2.1 When manager gets a reponse from sever, it tyr to get the corresponding request data by id
        // 2.2 manager notify, or said Monitor.Pulse(), the waiting thread which handles client invocation
        // 2.3 manager replaces requset data with the server's response
        // 1.3 The sleeping thread of Call() continues its job
        private static object _pipeDataMapLock = new object();

        private static SortedDictionary<Guid, PipeData> _pipeDataMap = new SortedDictionary<Guid, PipeData>();

        private static int DefaultTimeout = 5000;// 5000 ms
        private static int WaitTimeClientStart = 1000;

        private static void CheckDelegate(Type delegateMethod)
        {
            if (delegateMethod.IsSubclassOf(typeof(Delegate)) == false)
                throw new ArgumentException("The specified type is not a delegate type.");
            MethodInfo methodInfo = delegateMethod.GetMethod("Invoke");
            if (methodInfo.ReturnType.IsSerializable == false)
                throw new ArgumentException("return type of \"" + delegateMethod + "\" is not serializable");
            foreach (ParameterInfo parameterInfo in methodInfo.GetParameters())
                if (parameterInfo.ParameterType.IsSerializable == false)
                    throw new ArgumentException("parameter type of \"" + parameterInfo.ParameterType + "\" in \"" + delegateMethod + "\" is not serializable");
        }

        private static void CheckSerializable(object[] objects)
        {
            objects.ToList().ForEach(o =>
            {
                if (o.GetType().IsSerializable == false)
                    throw new ArgumentException(o.GetType() + " isn't serializable");
            });
        }

        private static void CheckParameterType(Type delegateMethod, object[] objects)
        {
            MethodInfo methodInfo = delegateMethod.GetMethod("Invoke");
            if (objects.Length != methodInfo.GetParameters().Length)
            {
                throw new ArgumentException("number of parameter size is not match");
            }
            for (int i = 0; i < objects.Length; i++)
            {
                if (objects[i].GetType().Equals(methodInfo.GetParameters()[i].ParameterType) == false)
                {
                    if (objects[i].GetType().IsSubclassOf(methodInfo.GetParameters()[i].ParameterType) == false)
                    {
                        throw new ArgumentException("parameter(" + i + ")'s type is \"" + objects[i].GetType() + "\"and not a subclass of \"" + methodInfo.GetParameters()[i].ParameterType + "\"");
                    }
                }
            }
        }

        private static void CheckReturnType(Type delegateMethod, Type returnType)
        {
            MethodInfo methodInfo = delegateMethod.GetMethod("Invoke");
            if (methodInfo.ReturnType.Equals(returnType) == false)
                if (methodInfo.ReturnType.IsSubclassOf(returnType) == false)
                    throw new ArgumentException("\"" + returnType + "\" isn't the return type of " + delegateMethod);
        }

        /// <summary>
        /// Registers a delegate method and serves a host for the method.
        /// </summary>
        /// <seealso cref="DelegatePipeManager"/>
        /// <param name="pipeName">An unique name to registed.</param>
        /// <param name="delegateObject">The delegate object to be registed.</param>
        /// <exception cref="ArgumentException">
        /// Throw if the signature of <paramref name="delegateObject"/> is invalid.
        /// </exception>
        /// <exception cref="DelegatePipeManagerException">
        /// Throw if the signature of <paramref name="delegateObject"/> has been registed.
        /// </exception>
        public static void RegisterDelegate(String pipeName, Delegate delegateObject)
        {
            CheckDelegate(delegateObject.GetType());
            NamedPipeServer<PipeData> server = new NamedPipeServer<PipeData>(pipeName);
            server.ClientMessage += delegate(NamedPipeConnection<PipeData, PipeData> conn, PipeData message)
            {
                try
                {
                    message.ReturnObject = delegateObject.DynamicInvoke(message.ParameterObjectList.ToArray());
                    message.ReturnObjectIsException = false;
                }
                catch (Exception e)
                {
                    message.ReturnObjectIsException = true;
                    Exception outException;
                    if (e.GetType().Equals(typeof(TargetInvocationException)))
                        outException = e.GetBaseException();
                    else
                        outException = e;

                    message.ReturnObject = outException.Message;
                }
                conn.PushMessage(message);
            };
            server.Start();
        }

        /// <summary>
        /// Calls the specified remote method with the given pipe name, delegate type, and parameters if any.
        /// </summary>
        /// <remarks>
        /// Timeout is 5000ms, if the IPC host does not response in time, an exception will be thrown.
        /// </remarks>
        /// <seealso cref="DelegatePipeManager"/>
        /// <typeparam name="TReturnObject">The returned type of this invocation.</typeparam>
        /// <param name="pipeName">The pipe name to be invoked</param>
        /// <param name="delegateMethod">The delegate type to invoked</param>
        /// <param name="parameters">The parameter objects, if any, of the delegate type</param>
        /// <returns>The retuned value of this invocation</returns>
        public static TReturnObject Call<TReturnObject>(string pipeName, Type delegateMethod, params object[] parameters)
        {
            return CallWithTimeout<TReturnObject>(pipeName, delegateMethod, DefaultTimeout, parameters);
        }

        /// <summary>
        /// Calls the specified IPC with the given pipe name, delegate type, and parameters if any.
        /// </summary>
        /// <seealso cref="DelegatePipeManager"/>
        /// <typeparam name="TReturnObject"></typeparam>
        /// <param name="pipeName">The named pipe to be used.</param>
        /// <param name="delegateMethod">The pipe name to be invoked</param>
        /// <param name="timeout">The assigned timeout value of this invocation. Unit micro seconds.</param>
        /// <param name="parameters">The parameter objects, if any, of the delegate type</param>
        /// <returns>The retuned value of this invocation</returns>ArgumentException
        /// <exception cref="ArgumentException">
        /// Throw if the signature of <paramref name="delegateMethod"/> is invalid.
        /// </exception>
        /// <exception cref="DelegatePipeManagerException">
        /// Throw if the request is timeout after 5000 ms
        /// </exception>
        /// <exception cref="System.Exception">
        /// Throw if the remote delegate instance throw an exception in this invocation.
        ///     For example, if the remote delegate throw a SomeKindOfException with a message "I'm the message of SomeKindOfException",
        ///     a <typeparamref name="System.Exception"/> will be throwed here with a message "get exception, nessage:I'm the message of SomeKindOfException",
        /// </exception>
        public static TReturnObject CallWithTimeout<TReturnObject>(string pipeName, Type delegateMethod, int timeout, params object[] parameters)
        {
            Stopwatch stopWatch = new Stopwatch();
            stopWatch.Start();

            CheckDelegate(delegateMethod);
            CheckSerializable(parameters);
            CheckParameterType(delegateMethod, parameters);

            NamedPipeClient<PipeData, PipeData> client = null;
            object clientLock = null;
            lock (_openedClientMapLock)
            {
                _openedClientMap.TryGetValue(pipeName, out client);
                _openedClientLockMap.TryGetValue(pipeName, out clientLock);
                if (client == null)
                {
                    client = new NamedPipeClient<PipeData, PipeData>(pipeName);
                    client.ServerMessage += delegate(NamedPipeConnection<PipeData, PipeData> conn, PipeData serverResponse)
                    {
                        lock (_pipeDataMap)
                        {
                            PipeData clientRequest = null;
                            _pipeDataMap.TryGetValue(serverResponse.Id, out clientRequest);
                            if (clientRequest != null)
                            {
                                _pipeDataMap.Remove(serverResponse.Id);
                                _pipeDataMap.Add(serverResponse.Id, serverResponse);
                                lock (clientRequest)
                                {
                                    Monitor.Pulse(clientRequest);
                                }
                            }
                        }
                    };
                    client.Start();
                    client.WaitForConnection((int)((long)timeout - stopWatch.ElapsedMilliseconds));
                    if (timeout - stopWatch.ElapsedMilliseconds < 0)
                    {
                        client.Stop();
                        throw new DelegatePipeManagerException("timeout when trying to connect to server");
                    }
                    clientLock = new object();
                    _openedClientMap.Add(pipeName, client);
                    _openedClientLockMap.Add(pipeName, clientLock);
                }
            }
            PipeData outPipeData = new PipeData();
            outPipeData.ParameterObjectList = parameters.ToList();
            lock (_pipeDataMap)
            {
                _pipeDataMap.Add(outPipeData.Id, outPipeData);
            }
            Task waitResponseTask = Task.Factory.StartNew(() =>
            {
                lock (outPipeData)
                {
                    Monitor.Wait(outPipeData, (int)((long)timeout - stopWatch.ElapsedMilliseconds));
                }
            });
            lock (clientLock)
            {
                client.PushMessage(outPipeData);
            }
            // wait response
            waitResponseTask.Wait();
            PipeData inPipeData = null;
            lock (_pipeDataMap)
            {
                _pipeDataMap.TryGetValue(outPipeData.Id, out inPipeData);
                _pipeDataMap.Remove(outPipeData.Id);
            }
            if (object.ReferenceEquals(inPipeData, outPipeData))
            {
                // If ReferenceEquals() is true, it means there is no response back from server side and replace the old one in _pipeDataMap
                throw new DelegatePipeManagerException("request timeout after " + timeout + " ms");
            }

            if (inPipeData.ReturnObjectIsException)
            {
                throw new Exception("get exception, message:" + inPipeData.ReturnObject);
            }
            else
            {
                return (TReturnObject)inPipeData.ReturnObject;
            }
        }
    }

    /// <summary>
    /// The exception that is thrown by the <see crefs="DelegatePipeManager" />
    /// </summary>
    [Serializable]
    public class DelegatePipeManagerException : SystemException
    {
        public DelegatePipeManagerException(string message)
            : base(message)
        {
        }
    }

    /// <summary>
    /// Represents the packet that wraps up the request/response through the named pipe.
    /// </summary>
    [Serializable]
    internal class PipeData
    {
        /// <summary>
        /// An uniqe id of this data
        /// </summary>
        public Guid Id { get; private set; }

        /// <summary>
        /// A list of parameter objects to be send from caller(client side) to callee(server side) if any.
        /// </summary>
        public List<object> ParameterObjectList { get; set; }

        /// <summary>
        /// A retuned object of to be send back to caller(client side) from callee(server side) if any.
        /// if <paramref name="ReturnObject"/> is true, it is an object of <typeparamref name="System.Exveption"/>.
        /// </summary>
        public object ReturnObject { get; set; }

        /// <summary>
        /// True if the callee(server side) throw an exception.
        /// </summary>
        public bool ReturnObjectIsException { get; set; }

        public PipeData()
        {
            Id = Guid.NewGuid();
            ReturnObjectIsException = false;
        }
    }
}