﻿using System;

namespace AirFusion.Common
{
	public static partial class Extensions
	{
		/// <summary>
		/// To the hexadecimal string.
		/// </summary>
		/// <param name="byteArray">The byte array.</param>
		/// <param name="delimeter">The delimeter.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentNullException" />
		public static string ToHexString(this byte[] byteArray, string delimeter = "-")
		{
			// check parameters
			if (byteArray == null)
			{
				throw new ArgumentNullException("byteArray");
			}

			return BitConverter.ToString(byteArray).Replace("-", delimeter);
		}

		/// <summary>
		/// To the hexadecimal string.
		/// </summary>
		/// <param name="byteArray">The byte array.</param>
		/// <param name="delimeter">The delimeter.</param>
		/// <param name="offset">The offset.</param>
		/// <param name="length">The length.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentNullException" />
		public static string ToHexString(this byte[] byteArray, int offset, int length, string delimeter = "-")
		{
			// check parameters
			if (byteArray == null)
			{
				throw new ArgumentNullException("byteArray");
			}
			if (offset < 0 || offset >= byteArray.Length)
			{
				throw new ArgumentOutOfRangeException("offset");
			}
			if (length < 0 || offset + length >= byteArray.Length)
			{
				throw new ArgumentOutOfRangeException("length");
			}

			return BitConverter.ToString(byteArray, offset, length).Replace("-", delimeter);
		}

		/// <summary>
		/// Gets string from the hexadecimal string.
		/// </summary>
		/// <typeparam name="T">The target type.</typeparam>
		/// <param name="text">The text.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.FormatException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.OverflowException" />
		public static T FromHexString<T>(this string text)
		{
			var targetType = typeof(T);

			// check parameters
			if (string.IsNullOrWhiteSpace(text))
			{
				throw new ArgumentException("The specified text is empty.");
			}
			if (System.Text.RegularExpressions.Regex.Match(text, @"[^0-9A-Fa-f\- ]").Success)
			{
				throw new FormatException("The specified string is an invalid hex string.");
			}

			// remove non A-F, a-f, 0-9
			text = System.Text.RegularExpressions.Regex.Replace(text, @"[^0-9A-Fa-f]", "", System.Text.RegularExpressions.RegexOptions.Compiled);

			if (targetType.Equals(typeof(byte)))
			{
				return (T)(object)Convert.ToByte(text, 16);
			}
			if (targetType.Equals(typeof(sbyte)))
			{
				return (T)(object)Convert.ToSByte(text, 16);
			}
			if (targetType.Equals(typeof(Int16)))
			{
				return (T)(object)Convert.ToInt16(text, 16);
			}
			if (targetType.Equals(typeof(Int32)))
			{
				return (T)(object)Convert.ToInt32(text, 16);
			}
			if (targetType.Equals(typeof(Int64)))
			{
				return (T)(object)Convert.ToInt64(text, 16);
			}
			if (targetType.Equals(typeof(UInt16)))
			{
				return (T)(object)Convert.ToUInt16(text, 16);
			}
			if (targetType.Equals(typeof(UInt32)))
			{
				return (T)(object)Convert.ToUInt32(text, 16);
			}
			if (targetType.Equals(typeof(UInt64)))
			{
				return (T)(object)Convert.ToUInt64(text, 16);
			}
			if (targetType.Equals(typeof(byte[])))
			{
				int numberChars = text.Length;
				byte[] bytes = new byte[(int)Math.Ceiling((double)numberChars / 2)];
				for (int i = numberChars - 1; i >= 0; i -= 2)
				{
					int start = i - 1;
					int substringLength = 2;
					if (start < 0)
					{
						start = 0;
						substringLength = 1;
					}
					bytes[i / 2] = Convert.ToByte(text.Substring(start, substringLength), 16);
				}
				return (T)(object)bytes;
			}

			throw new NotSupportedException("Failed converting the specified text to the specified target type '" + typeof(T).Name + "'.");
		}
	}
}
