﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net.NetworkInformation;

namespace AirFusion.Common.Network
{
	/// <summary>
	/// Represents an observable class of network information.
	/// </summary>
	public class NetworkObservable : IObservable<ICollection<NetworkInfo>>, IDisposable
	{
		#region Classes

		private class Unsubscriber : IDisposable
		{
			private List<IObserver<ICollection<NetworkInfo>>> _observers;
			private IObserver<ICollection<NetworkInfo>> _observer;

			public Unsubscriber(List<IObserver<ICollection<NetworkInfo>>> observers, IObserver<ICollection<NetworkInfo>> observer)
			{
				this._observers = observers;
				this._observer = observer;
			}

			public void Dispose()
			{
				if (this._observer != null && this._observers.Contains(this._observer))
				{
					this._observers.Remove(this._observer);
				}
			}
		}

		#endregion Classes

		#region Definitions

		// workaround, since NetworkAddressChangedprevent fire multiple times in a single operation,
		// we add this to prevent duplicate event is fired within a given time
		private static TimeSpan _fireTolerance = TimeSpan.FromMilliseconds(1000);

		#endregion Definitions

		#region Private/Protected/Internal Variables

		private List<IObserver<ICollection<NetworkInfo>>> _observers = new List<IObserver<ICollection<NetworkInfo>>>();
		private object _locker = new object();
		private DateTime _lastFireDateTime = DateTime.Now;

		#endregion Private/Protected/Internal Variables

		#region Override IObservable

		/// <summary>
		/// Initializes a new instance of the <see cref="NetworkObservable"/> class.
		/// </summary>
		public NetworkObservable()
		{
			this.RegisterNetworkChangeEvents();
		}

		/// <summary>
		/// Subscribes the specified observer.
		/// </summary>
		/// <param name="observer">The observer.</param>
		/// <returns></returns>
		public IDisposable Subscribe(IObserver<ICollection<NetworkInfo>> observer)
		{
			if (!this._observers.Contains(observer))
			{
				this._observers.Add(observer);
			}

			return new Unsubscriber(this._observers, observer);
		}

		#endregion Override IObservable

		#region Private Methods

		private void RegisterNetworkChangeEvents()
		{
			NetworkChange.NetworkAddressChanged += this.OnNetworkChanged;
		}

		private void UnRegisterNetworkChangedEvents()
		{
			NetworkChange.NetworkAddressChanged -= this.OnNetworkChanged;
		}

		private void OnNetworkChanged(object sender, EventArgs args)
		{
			Debug.Assert(this._locker != null);
			lock (this._locker)
			{
				if (DateTime.Now - this._lastFireDateTime >= _fireTolerance)
				{
					this._lastFireDateTime = DateTime.Now;
					ICollection<NetworkInfo> networkList = NetworkHelper.GetAvailableSupportedNetworkList();

					Debug.Assert(this._observers != null);
					foreach (var observer in this._observers)
					{
						observer.OnNext(networkList);
						observer.OnCompleted();
					}
				}
			}
		}

		#endregion Private Methods

		#region Override IDisposable

		internal bool _disposed = false;

		internal virtual void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					try
					{
						this.UnRegisterNetworkChangedEvents();
					}
					catch { }
				}

				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}

		#endregion Override IDisposable
	}

	/// <summary>
	/// Represents an observer class of network information.
	/// </summary>
	public class NetworkObserver : IObserver<ICollection<NetworkInfo>>
	{
		#region Classes

		/// <summary>
		/// Represents the event arguments of network changed event.
		/// </summary>
		public class NetworkChangedEventArgs : EventArgs
		{
			/// <summary>
			/// Gets the network list.
			/// </summary>
			public ICollection<NetworkInfo> NetworkList { get; private set; }

			internal NetworkChangedEventArgs(ICollection<NetworkInfo> networkList)
			{
				this.NetworkList = networkList;
			}
		}

		#endregion Classes

		#region Delegates

		/// <summary>
		/// Represents the methods that will handle the event of network changed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="NetworkChangedEventArgs"/> instance containing the event data.</param>
		public delegate void NetworkChangedEventHandler(object sender, NetworkChangedEventArgs args);

		#endregion Delegates

		#region Events

		/// <summary>
		/// Occurs when network changed.
		/// </summary>
		public event NetworkChangedEventHandler NetworkChanged;

		#endregion Events

		#region Properties

		/// <summary>
		/// Gets the network list.
		/// </summary>
		public ICollection<NetworkInfo> NetworkList { get; private set; }

		#endregion Properties

		/// <summary>
		/// Notifies the observer that the provider has finished sending push-based notifications.
		/// </summary>
		public virtual void OnCompleted()
		{
			if (this.NetworkChanged != null && this.NetworkList != null)
			{
				CommonUtils.BeginInvokeEvent<NetworkChangedEventArgs>(new EventHandler<NetworkChangedEventArgs>(this.NetworkChanged), this, new NetworkChangedEventArgs(this.NetworkList));
			}
		}

		/// <summary>
		/// Notifies the observer that the provider has experienced an error condition.
		/// </summary>
		/// <param name="e">An object that provides additional information about the error.</param>
		public virtual void OnError(Exception e)
		{
			StaticLogger.LogError("There's something wrong with the network observer, exception:\r\n" + CommonUtils.GetExceptionMessages(e));
		}

		/// <summary>
		/// Provides the observer with new data.
		/// </summary>
		/// <param name="networkList">The current notification information.</param>
		public virtual void OnNext(ICollection<NetworkInfo> networkList)
		{
			this.NetworkList = networkList;
		}
	}
}