﻿using Microsoft.Win32;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Common.RegistryHolders
{
	/// <summary>
	/// Represents a registry holders that able to read properties of a specific object from registry keys or store properties of that object into registry keys.
	/// </summary>
	/// <example>
	/// var instance = new AFUserDevice();
	/// RegistryHolders.Create(instance, RegistryHive.CurrentUser, "Software/OEM/AcerEXTEND/Devices");
	/// RegistryHolders.Load(instance); // load properties from the registry.
	/// RegistryHolders.Store(instance); // store all properties from the registry.
	/// object instanceID = RegistryHolders.LoadProperty(instance, "InstanceID"); // load the specified property.
	/// Guid instanceID = RegistryHolders.LoadProperty&lt;Guid&gt;(instance, "InstanceID"); // load the specified property with the given type.
	/// RegistryHolders.StoreProperty(instance, "InstanceID"); // store the specified property.
	/// </example>
	public static class RegistryHolders
	{
		#region Private/Protected/Internal Static Variables/Methods

		private static ConcurrentDictionary<object, RegistryHoldersData> _holders;

		static RegistryHolders()
		{
			_holders = new ConcurrentDictionary<object, RegistryHoldersData>();
		}

		#endregion Private/Protected/Internal Static Variables/Methods

		#region Public Static Methods

		/// <summary>
		/// Opens or creates registry key for the specified instance.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <param name="regHive">The registry hive.</param>
		/// <param name="regKeyPath">The registry key path.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.IO.IOException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.Security.SecurityException" />
		/// <exception cref="System.UnauthorizedAccessException" />
		public static void Create(object instance, RegistryHive regHive, string regKeyPath)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}
				if (instance.GetType().IsClass == false)
				{
					throw new ArgumentException("The specified instance is not an instance of a class.", "instance");
				}
				if (string.IsNullOrWhiteSpace(regKeyPath))
				{
					throw new ArgumentException("Please specify a valid registry key path.", "regKeyPath");
				}

				Debug.Assert(_holders != null);
				var holderData = _holders.AddOrUpdate(instance, new RegistryHoldersData(instance),
					(obj, data) =>
					{
						return data;
					});

				holderData.OpenOrCreateRegistry(regHive, regKeyPath);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Opens or creates registry key for the specified instance.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <param name="regKey">The registry key.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		public static void Create(object instance, RegistryKey regKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}
				if (instance.GetType().IsClass == false)
				{
					throw new ArgumentException("The specified instance is not an instance of a class.", "instance");
				}
				if (regKey == null)
				{
					throw new ArgumentNullException("regKey");
				}

				Debug.Assert(_holders != null);
				var holderData = _holders.AddOrUpdate(instance, new RegistryHoldersData(instance),
					(obj, data) =>
					{
						return data;
					});

				holderData.SetRegistry(regKey);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads properties of the specified instance from registry.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <param name="throwOnLoadFail">A value indicating whether to throw exceptions when failed loading.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		public static void Load(object instance, bool throwOnLoadFail = false)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}

				Debug.Assert(_holders != null);
				RegistryHoldersData holderData;
				if (_holders.TryGetValue(instance, out holderData) == false)
				{
					throw new InvalidOperationException("Please call Create() first for the specified instance.");
				}
				holderData.LoadProperties(throwOnLoadFail);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads the specified property from the specified instance.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <param name="propertyName">The name of the property.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		public static object LoadProperty(object instance, string propertyName)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}

				Debug.Assert(_holders != null);
				RegistryHoldersData holderData;
				if (_holders.TryGetValue(instance, out holderData) == false)
				{
					throw new InvalidOperationException("Please call Create() first for the specified instance.");
				}
				return holderData.LoadProperty(propertyName);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the properties of the specified instance into registry.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NotSupportedException" />
		public static void Store(object instance)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}

				Debug.Assert(_holders != null);
				RegistryHoldersData holderData;
				if (_holders.TryGetValue(instance, out holderData) == false)
				{
					throw new InvalidOperationException("Please call Create() first for the specified instance.");
				}
				holderData.StoreProperties();
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the specified property of the specified instance into registry.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <param name="propertyName">The name of the property.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		public static void StoreProperty(object instance, string propertyName)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (instance == null)
				{
					throw new ArgumentNullException("instance");
				}

				Debug.Assert(_holders != null);
				RegistryHoldersData holderData;
				if (_holders.TryGetValue(instance, out holderData) == false)
				{
					throw new InvalidOperationException("Please call Create() first for the specified instance.");
				}
				holderData.StoreProperty(propertyName);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Static Methods
	}
}