﻿using System;

namespace AirFusion.Common.Wmi.Cim
{
	/// <summary>
	/// The CIM_OperatingSystem class represents a computer operating system, which is made up of software and firmware that make a computer system's hardware usable.
	/// </summary>
	[ParameterValue("CIM_OperatingSystem")]
	public class WmiOperatingSystem : WmiLogicalElement
	{
		/// <summary>
		/// Gets the name of the class or subclass used in the creation of an instance.
		/// When used with other key properties of the class, this property allows all instances of the class and its subclasses to be uniquely identified.
		/// </summary>
		[ParameterValue("CreationClassName")]
		public string CreationClassName { get; internal set; }

		/// <summary>
		/// Gets the scoping computer system's creation class name.
		/// </summary>
		[ParameterValue("CSCreationClassName")]
		public string CSCreationClassName { get; internal set; }

		/// <summary>
		/// Gets the scoping computer system's name.
		/// </summary>
		[ParameterValue("CSName")]
		public string CSName { get; internal set; }

		/// <summary>
		/// Gets the number of minutes the operating system is offset from Greenwich Mean Time (GMT). The number is positive, negative, or zero.
		/// </summary>
		[ParameterValue("CurrentTimeZone")]
		public Int16? CurrentTimeZone { get; internal set; }

		/// <summary>
		/// Gets a value that indicates if the operating system is distributed across several computer system nodes, which should be grouped as a cluster.
		/// </summary>
		[ParameterValue("Distributed")]
		public bool? Distributed { get; internal set; }

		/// <summary>
		/// Gets the number of kilobytes of physical memory currently unused and available.
		/// </summary>
		[ParameterValue("FreePhysicalMemory")]
		public UInt64? FreePhysicalMemory { get; internal set; }

		/// <summary>
		/// Gets the number of kilobytes that can be mapped into the operating system's paging files without causing other pages to be swapped out.
		/// A value of 0 indicates that there are no paging files.
		/// </summary>
		[ParameterValue("FreeSpaceInPagingFiles")]
		public UInt64? FreeSpaceInPagingFiles { get; internal set; }

		/// <summary>
		/// Gets the number of kilobytes of virtual memory currently unused and available.
		/// For example, this can be calculated by adding the amount of free RAM to the amount of free paging space
		/// (that is, adding the FreePhysicalMemory and FreeSpaceInPagingFiles properties).
		/// </summary>
		[ParameterValue("FreeVirtualMemory")]
		public UInt64? FreeVirtualMemory { get; internal set; }

		/// <summary>
		/// Gets the time when the operating system was last booted.
		/// </summary>
		[ParameterValue("LastBootUpTime"), TypeConverterTypeAttribute(typeof(Converters.WmiDateTimeConverter))]
		public DateTime? LastBootUpTime { get; internal set; }

		/// <summary>
		/// Gets the operating system's notion of the local date and time of day.
		/// </summary>
		[ParameterValue("LocalDateTime"), TypeConverterTypeAttribute(typeof(Converters.WmiDateTimeConverter))]
		public DateTime? LocalDateTime { get; internal set; }

		/// <summary>
		/// Gets the maximum number of process contexts the operating system can support. If there is no fixed maximum, the value should be 0 (zero).
		/// On systems that have a fixed maximum, this object can help diagnose failures that occur when the maximum is reached. If unknown, enter -1.
		/// </summary>
		[ParameterValue("MaxNumberOfProcesses")]
		public UInt32? MaxNumberOfProcesses { get; internal set; }

		/// <summary>
		/// Gets the maximum number of kilobytes of memory that can be allocated to a process. For operating systems with no virtual memory,
		/// this value is typically equal to the total amount of physical memory, minus memory used by the BIOS and the operating system.
		/// For some operating systems, this value can be infinity, in which case 0 should be entered. In other cases, this value can be a constant, for example, 2 GB or 4 GB.
		/// </summary>
		[ParameterValue("MaxProcessMemorySize")]
		public UInt64? MaxProcessMemorySize { get; internal set; }

		/// <summary>
		/// Gets the number of user licenses for the operating system. If unlimited, enter 0, if unknown, enter -1.
		/// </summary>
		[ParameterValue("NumberOfLicensedUsers")]
		public UInt32? NumberOfLicensedUsers { get; internal set; }

		/// <summary>
		/// Gets the number of process contexts currently loaded or running on the operating system.
		/// </summary>
		[ParameterValue("NumberOfProcesses")]
		public UInt32? NumberOfProcesses { get; internal set; }

		/// <summary>
		/// Gets the number of user sessions for which the operating system is currently storing state information.
		/// </summary>
		[ParameterValue("NumberOfUsers")]
		public UInt32? NumberOfUsers { get; internal set; }

		/// <summary>
		/// Gets the type of operating system.
		/// </summary>
		[ParameterValue("OSType")]
		public UInt16? OSType { get; internal set; }

		/// <summary>
		/// Gets the description of the manufacturer and operating system type when the OSType property is set to 1 ("Other").
		/// The format of the string inserted in OtherTypeDescription should be similar to the Values strings defined for OSType.
		/// This property should be set to null when OSType is a value other than 1 (one).
		/// </summary>
		[ParameterValue("OtherTypeDescription")]
		public string OtherTypeDescription { get; internal set; }

		/// <summary>
		/// Gets the number of kilobytes that can be stored in the operating system's paging files.
		/// This number does not represent the actual physical size of the paging file on disk. A value of 0 (zero) indicates that there are no paging files.
		/// </summary>
		[ParameterValue("SizeStoredInPagingFiles")]
		public UInt64? SizeStoredInPagingFiles { get; internal set; }

		/// <summary>
		/// Gets the total swap space, in kilobytes. This value can be null (unspecified) if swap space is not distinguished from page files.
		/// However, some operating systems distinguish these concepts.
		/// For example, entire processes can be "swapped out" in UNIX when the free page list falls and remains below a specified amount.
		/// </summary>
		[ParameterValue("TotalSwapSpaceSize")]
		public UInt64? TotalSwapSpaceSize { get; internal set; }

		/// <summary>
		/// Gets the number of kilobytes of virtual memory. For example, calculate this by adding the amount of total RAM to the amount of paging space
		/// (that is, add the amount of memory in or aggregated by the computer system to the SizeStoredInPagingFiles property).
		/// </summary>
		[ParameterValue("TotalVirtualMemorySize")]
		public UInt64? TotalVirtualMemorySize { get; internal set; }

		/// <summary>
		/// Gets the total amount of physical memory, in kilobytes, available to the operating system.
		/// This value does not necessarily indicate the true amount of physical memory, but what is reported to the operating system as available to it.
		/// </summary>
		[ParameterValue("TotalVisibleMemorySize")]
		public UInt64? TotalVisibleMemorySize { get; internal set; }

		/// <summary>
		/// Gets the version of the operation.
		/// </summary>
		/// <remarks>
		/// The version of the operation should be in one of the following forms:
		/// &lt;major&gt;.&lt;minor&gt;.&lt;revision&gt;
		/// &lt;major&gt;.&lt;minor&gt;&lt;letter&gt;&lt;revision&gt;
		/// </remarks>
		[ParameterValue("Version")]
		public string Version { get; internal set; }

		/// <summary>
		/// Initializes the object.
		/// </summary>
		public WmiOperatingSystem()
		{
			this.CreationClassName = null;
			this.CSCreationClassName = null;
			this.CSName = null;
			this.CurrentTimeZone = null;
			this.Distributed = null;
			this.FreePhysicalMemory = null;
			this.FreeSpaceInPagingFiles = null;
			this.FreeVirtualMemory = null;
			this.LastBootUpTime = null;
			this.LocalDateTime = null;
			this.MaxNumberOfProcesses = null;
			this.MaxProcessMemorySize = null;
			this.NumberOfLicensedUsers = null;
			this.NumberOfProcesses = null;
			this.NumberOfUsers = null;
			this.OSType = null;
			this.OtherTypeDescription = null;
			this.SizeStoredInPagingFiles = null;
			this.TotalSwapSpaceSize = null;
			this.TotalVirtualMemorySize = null;
			this.TotalVisibleMemorySize = null;
			this.Version = null;
		}
	}
}