﻿namespace AirFusion.Common.Wmi.Cim
{
	/// <summary>
	/// The CIM_PhysicalElement subclasses define any component of a system that has a distinct physical identity.
	/// Instances of this class can be defined in terms of labels that can be physically attached to the object.
	/// All processes, files, and logical devices are not considered to be physical elements.
	/// For example, it is not possible to attach a label to a modem; it is only possible to attach a label to the card that implements the modem.
	/// The same card could also implement a LAN adapter. Tangible managed system elements (usually hardware items) have a physical manifestation.
	/// A managed system element is not necessarily a discrete component. For example, it is possible for a single card (a type of physical element) to host more than one logical device.
	/// The card would be represented by a single physical element associated with multiple logical devices.
	/// </summary>
	[ParameterValue("CIM_PhysicalElement")]
	public class WmiPhysicalElement : WmiManagedSystemElement
	{
		/// <summary>
		/// Gets the name of the class or subclass used in the creation of an instance.
		/// When used with other key properties of the class, this property allows all instances of the class and its subclasses to be uniquely identified.
		/// </summary>
		[ParameterValue("CreationClassName")]
		public string CreationClassName { get; internal set; }

		/// <summary>
		/// Gets the name of the organization responsible for producing the physical element.
		/// </summary>
		[ParameterValue("Manufacturer")]
		public string Manufacturer { get; internal set; }

		/// <summary>
		/// Gets the name by which the physical element is generally known.
		/// </summary>
		[ParameterValue("Model")]
		public string Model { get; internal set; }

		/// <summary>
		/// Gets the additional data, beyond asset tag information, that can be used to identify a physical element.
		/// One example is bar-code data that is associated with an element, which also has an asset tag. Note that if only bar-code data is available,
		/// and is unique and able to be used as an element key, this property would be null and the bar-code data would be used as the class key in the Tag property.
		/// </summary>
		[ParameterValue("OtherIdentifyingInfo")]
		public string OtherIdentifyingInfo { get; internal set; }

		/// <summary>
		/// Gets the part number assigned by the organization responsible for producing or manufacturing the physical element.
		/// </summary>
		[ParameterValue("PartNumber")]
		public string PartNumber { get; internal set; }

		/// <summary>
		/// Gets a value that indicates if the physical element is powered on.
		/// </summary>
		[ParameterValue("PoweredOn")]
		public bool? PoweredOn { get; internal set; }

		/// <summary>
		/// Gets the manufacturer-allocated number used to identify the physical element.
		/// </summary>
		[ParameterValue("SerialNumber")]
		public string SerialNumber { get; internal set; }

		/// <summary>
		/// Gets the stock-keeping unit number for this physical element.
		/// </summary>
		[ParameterValue("SKU")]
		public string Sku { get; internal set; }

		/// <summary>
		/// Gets the arbitrary string that uniquely identifies the physical element and serves as the element's key.
		/// This property can contain information, such as asset tag or serial number data.
		/// </summary>
		[ParameterValue("Tag")]
		public string Tag { get; internal set; }

		/// <summary>
		/// Gets the string that indicates the version of the physical element.
		/// </summary>
		[ParameterValue("Version")]
		public string Version { get; internal set; }

		/// <summary>
		/// Initializes the object.
		/// </summary>
		public WmiPhysicalElement()
		{
			this.CreationClassName = null;
			this.Manufacturer = null;
			this.Model = null;
			this.OtherIdentifyingInfo = null;
			this.PartNumber = null;
			this.PoweredOn = null;
			this.SerialNumber = null;
			this.Sku = null;
			this.Tag = null;
			this.Version = null;
		}
	}
}