﻿using System;

namespace AirFusion.Common.Wmi.Win32
{
	/// <summary>
	/// The Win32_OperatingSystem WMI class represents a Windows-based operating system installed on a computer.
	/// </summary>
	[ParameterValue("Win32_OperatingSystem")]
	public class WmiOperatingSystem : Cim.WmiOperatingSystem
	{
		/// <summary>
		/// Gets the name of the disk drive from which the Windows operating system starts.
		/// </summary>
		/// <example>
		/// \\Device\Harddisk0
		/// </example>
		[ParameterValue("BootDevice")]
		public string BootDevice { get; internal set; }

		/// <summary>
		/// Gets the build number of an operating system. It can be used for more precise version information than product release version numbers.
		/// </summary>
		/// <example>
		/// 1381
		/// </example>
		[ParameterValue("BuildNumber")]
		public string BuildNumber { get; internal set; }

		/// <summary>
		/// Gets the type of build used for an operating system.
		/// </summary>
		/// <example>
		/// retail build
		/// </example>
		/// <example>
		/// checked build
		/// </example>
		[ParameterValue("BuildType")]
		public string BuildType { get; internal set; }

		/// <summary>
		/// Gets the code page value an operating system uses. A code page contains a character table that an operating system uses to translate strings for different languages.
		/// The American National Standards Institute (ANSI) lists values that represent defined code pages.
		/// If an operating system does not use an ANSI code page, this member is set to 0 (zero).
		/// The CodeSet string can use a maximum of six characters to define the code page value.
		/// </summary>
		/// <example>
		/// 1255
		/// </example>
		[ParameterValue("CodeSet")]
		public string CodeSet { get; internal set; }

		/// <summary>
		/// Gets the code for the country/region that an operating system uses. Values are based on international phone dialing prefixes—also referred to as IBM country/region codes.
		/// This property can use a maximum of six characters to define the country/region code value.
		/// </summary>
		/// <example>
		/// 1
		/// </example>
		[ParameterValue("CountryCode")]
		public string CountryCode { get; internal set; }

		/// <summary>
		/// Gets the string that indicates the latest service pack installed on a computer. If no service pack is installed, the string is NULL.
		/// </summary>
		/// <example>
		/// Service Pack 3
		/// </example>
		[ParameterValue("CSDVersion")]
		public string CSDVersion { get; internal set; }

		/// <summary>
		/// When the data execution prevention hardware feature is available, this property indicates that the feature is set to work for 32-bit applications if True.
		/// On 64-bit computers, the data execution prevention feature is configured in the Boot Configuration Data (BCD) store and the properties in Win32_OperatingSystem are set accordingly.
		/// </summary>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("DataExecutionPrevention_32BitApplications")]
		public bool? DataExecutionPrevention_32BitApplications { get; internal set; }

		/// <summary>
		/// Data execution prevention is a hardware feature to prevent buffer overrun attacks by stopping the execution of code on data-type memory pages.
		/// If True, then this feature is available. On 64-bit computers, the data execution prevention feature is configured in the BCD store and the properties in Win32_OperatingSystem are set accordingly.
		/// </summary>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("DataExecutionPrevention_Available")]
		public bool? DataExecutionPrevention_Available { get; internal set; }

		/// <summary>
		/// When the data execution prevention hardware feature is available, this property indicates that the feature is set to work for drivers if True.
		/// On 64-bit computers, the data execution prevention feature is configured in the BCD store and the properties in Win32_OperatingSystem are set accordingly.
		/// </summary>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("DataExecutionPrevention_Drivers")]
		public bool? DataExecutionPrevention_Drivers { get; internal set; }

		/// <summary>
		/// Gets a value that indicates which Data Execution Prevention (DEP) setting is applied.
		/// The DEP setting specifies the extent to which DEP applies to 32-bit applications on the system. DEP is always applied to the Windows kernel.
		/// </summary>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("DataExecutionPrevention_SupportPolicy")]
		public byte? DataExecutionPrevention_SupportPolicy { get; internal set; }

		/// <summary>
		/// Gets a value that indicates if the operating system is a checked (debug) build.
		/// If True, the debugging version of User.exe is installed. Checked builds provide error checking, argument verification, and system debugging code.
		/// Additional code in a checked binary generates a kernel debugger error message and breaks into the debugger.
		/// This helps immediately determine the cause and location of the error. Performance may be affected in a checked build due to the additional code that is executed.
		/// </summary>
		[ParameterValue("Debug")]
		public bool? Debug { get; internal set; }

		/// <summary>
		/// Gets the encryption level for secure transactions: 40-bit, 128-bit, or n-bit.
		/// </summary>
		[ParameterValue("EncryptionLevel")]
		public UInt32? EncryptionLevel { get; internal set; }

		/// <summary>
		/// Gets a value that indicates the system boosts the quantum length by 6; if 1, then 12; and if 2 then 18. The default value is 2.
		/// Increase in priority is given to the foreground application. Application boost is implemented by giving an application more execution time slices (quantum lengths).
		/// </summary>
		[ParameterValue("ForegroundApplicationBoost")]
		public byte? ForegroundApplicationBoost { get; internal set; }

		/// <summary>
		/// Gets the language identifier used by the operating system. A language identifier is a standard international numeric abbreviation for a country/region.
		/// Each language has a unique language identifier (LANGID), a 16-bit value that consists of a primary language identifier and a secondary language identifier.
		/// </summary>
		[ParameterValue("Locale")]
		public string Locale { get; internal set; }

		/// <summary>
		/// Gets the name of the operating system manufacturer. For Windows-based systems, this value is "Microsoft Corporation".
		/// </summary>
		[ParameterValue("Manufacturer")]
		public string Manufacturer { get; internal set; }

		/// <summary>
		/// Gets the multilingual User Interface Pack (MUI Pack ) languages installed on the computer. For example, "en-us".
		/// MUI Pack languages are resource files that can be installed on the English version of the operating system.
		/// When an MUI Pack is installed, you can can change the user interface language to one of 33 supported languages.
		/// </summary>
		[ParameterValue("MUILanguages")]
		public string[] MUILanguages { get; internal set; }

		/// <summary>
		/// Gets the stock Keeping Unit (SKU) number for the operating system.
		/// </summary>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("OperatingSystemSKU")]
		public UInt32? OperatingSystemSKU { get; internal set; }

		/// <summary>
		/// Gets the company name for the registered user of the operating system.
		/// </summary>
		/// <example>
		/// Microsoft Corporation
		/// </example>
		[ParameterValue("Organization")]
		public string Organization { get; internal set; }

		/// <summary>
		/// Gets the architecture of the operating system, as opposed to the processor. This property can be localized.
		/// </summary>
		/// <example>
		/// 32-bit
		/// </example>
		/// <remarks>
		/// Windows Server 2003: This property is not available.
		/// </remarks>
		[ParameterValue("OSArchitecture")]
		public string OSArchitecture { get; internal set; }

		/// <summary>
		/// Gets the language version of the operating system installed. The following table lists the possible values. Example: 0x0807 (German, Switzerland).
		/// </summary>
		[ParameterValue("OSLanguage")]
		public UInt32? OSLanguage { get; internal set; }

		/// <summary>
		/// Gets a value that indicates if the physical address extensions (PAE) are enabled by the operating system running on Intel processors.
		/// PAE allows applications to address more than 4 GB of physical memory.
		/// When PAE is enabled, the operating system uses three-level linear address translation rather than two-level.
		/// Providing more physical memory to an application reduces the need to swap memory to the page file and increases performance.
		/// To enable, PAE, use the "/PAE" switch in the Boot.ini file.
		/// For more information about the Physical Address Extension feature, see Http://Go.Microsoft.Com/FWLink/p/?LinkID=45912.
		/// </summary>
		[ParameterValue("PAEEnabled")]
		public bool? PAEEnabled { get; internal set; }

		/// <summary>
		/// Gets a value that indicates whether the operating system booted from an external USB device.
		/// If true, the operating system has detected it is booting on a supported locally connected storage device.
		/// </summary>
		[ParameterValue("PortableOperatingSystem")]
		public bool? PortableOperatingSystem { get; internal set; }

		/// <summary>
		/// Gets a value that indicates whether this is the primary operating system.
		/// </summary>
		[ParameterValue("Primary")]
		public bool? Primary { get; internal set; }

		/// <summary>
		/// Gets the additional system information.
		/// </summary>
		[ParameterValue("ProductType")]
		public UInt32? ProductType { get; internal set; }

		/// <summary>
		/// Gets the name of the registered user of the operating system.
		/// </summary>
		/// <example>
		/// Seoker Wang
		/// </example>
		[ParameterValue("RegisteredUser")]
		public string RegisteredUser { get; internal set; }

		/// <summary>
		/// Gets the operating system product serial identification number.
		/// </summary>
		/// <example>
		/// 10497-OEM-0031416-71674
		/// </example>
		[ParameterValue("SerialNumber")]
		public string SerialNumber { get; internal set; }

		/// <summary>
		/// Gets the major version number of the service pack installed on the computer system. If no service pack has been installed, the value is 0 (zero).
		/// </summary>
		[ParameterValue("ServicePackMajorVersion")]
		public UInt16? ServicePackMajorVersion { get; internal set; }

		/// <summary>
		/// Gets the Minor version number of the service pack installed on the computer system. If no service pack has been installed, the value is 0 (zero).
		/// </summary>
		[ParameterValue("ServicePackMinorVersion")]
		public UInt16? ServicePackMinorVersion { get; internal set; }

		/// <summary>
		/// Gets the bit flags that identify the product suites available on the system. For example, to specify both Personal and BackOffice, set SuiteMask to 4 | 512 or 516.
		/// </summary>
		[ParameterValue("SuiteMask")]
		public UInt32? SuiteMask { get; internal set; }

		/// <summary>
		/// Gets the physical disk partition on which the operating system is installed.
		/// </summary>
		[ParameterValue("SystemDevice")]
		public string SystemDevice { get; internal set; }

		/// <summary>
		/// Gets the system directory of the operating system.
		/// </summary>
		/// <example>
		/// C:\WINDOWS\SYSTEM32
		/// </example>
		[ParameterValue("SystemDirectory")]
		public string SystemDirectory { get; internal set; }

		/// <summary>
		/// Gets the letter of the disk drive on which the operating system resides.
		/// </summary>
		/// <example>
		/// C:
		/// </example>
		[ParameterValue("SystemDrive")]
		public string SystemDrive { get; internal set; }

		/// <summary>
		/// Gets the Windows directory of the operating system.
		/// </summary>
		/// <example>
		/// C:\WINDOWS
		/// </example>
		[ParameterValue("WindowsDirectory")]
		public string WindowsDirectory { get; internal set; }

		/// <summary>
		/// Initializes the object.
		/// </summary>
		public WmiOperatingSystem()
		{
			this.BootDevice = null;
			this.BuildNumber = null;
			this.BuildType = null;
			this.CodeSet = null;
			this.CountryCode = null;
			this.DataExecutionPrevention_32BitApplications = null;
			this.DataExecutionPrevention_Available = null;
			this.DataExecutionPrevention_Drivers = null;
			this.DataExecutionPrevention_SupportPolicy = null;
			this.Debug = null;
			this.EncryptionLevel = null;
			this.ForegroundApplicationBoost = null;
			this.Locale = null;
			this.MUILanguages = null;
			this.OperatingSystemSKU = null;
			this.Organization = null;
			this.OSArchitecture = null;
			this.OSLanguage = null;
			this.PAEEnabled = null;
			this.PortableOperatingSystem = null;
			this.Primary = null;
			this.ProductType = null;
			this.RegisteredUser = null;
			this.SerialNumber = null;
			this.ServicePackMajorVersion = null;
			this.ServicePackMinorVersion = null;
			this.SuiteMask = null;
			this.SystemDevice = null;
			this.SystemDirectory = null;
			this.SystemDrive = null;
			this.WindowsDirectory = null;
		}
	}
}