﻿using AirFusion.Common;
using AirFusion.Common.ProximityDetection;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		#region SoundTrigger

		/// <summary>
		/// The callback of proximity detection.
		/// </summary>
		/// <param name="args">The arguments.</param>
		private void OnProximityDetectionCallback(IProximityDetectionCallbackArgs args)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("args: " + (args != null ? "\r\n" + args.ToString() : "N/A"));

				// check parameters
				if (args == null)
				{
					StaticLogger.LogWarn("The args is null, discard.");
					return;
				}

				// check prerequisite
				if (!this.IsDiscoverable)
				{
					StaticLogger.LogWarn("It's not discoverable currently, discarded.");
					return;
				}

				Guid instanceID = args.GetIdentity();

				if (instanceID.CompareTo(Guid.Empty) != 0)
				{
					AFUserDevice device = this.DeviceStore.GetDevice(instanceID);

					if (device != null) // if it exist in device store
					{
						try
						{
							device.PairingInviteAsync().ContinueWith(delegate(Task task)
							{
								if (task.IsFaulted)
								{
									StaticLogger.LogError("Failed sending pairing invitation, exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception));
								}
							}, TaskContinuationOptions.ExecuteSynchronously).ConfigureAwait(continueOnCapturedContext: false);
						}
						catch (Exception ex)
						{
							StaticLogger.LogError("Failed sending pairing invitation, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						}
					}
					else if (this._dandelionAdapter != null) // otherwise, send invitation through Dandelion adapter
					{
						StaticLogger.LogWarn("The instance ID '" + instanceID.GetStringFormat() + "' doesn't exist in the device store, might due to the device cannot be found, go through Dandelion instead.");

						this._dandelionAdapter.SendPairingInvitationAsync(instanceID).ContinueWith(delegate(Task task)
						{
							if (task.IsFaulted)
							{
								StaticLogger.LogError("Failed sending pairing invitation through Dandelion, exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception));
							}
							else if (task.IsCanceled)
							{
								StaticLogger.LogError("SendPairingInvitationAsync() is cancelled.");
							}
						}, TaskContinuationOptions.ExecuteSynchronously).ConfigureAwait(continueOnCapturedContext: false);
					}
					else // ???
					{
						StaticLogger.LogWarn("The instance ID '" + instanceID.GetStringFormat() + "' doesn't exist in the device store, might due to the device cannot be found.");
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion SoundTrigger

		#region MulticastDnsAdapter

		/// <summary>
		/// Called when the address of an mDNS instance is updated.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="UserDeviceManager.MulticastDnsAdapter.AddressUpdatedEventArgs"/> instance containing the event data.</param>
		private void OnMulticastDnsAdapterAddressUpdated(object sender, UserDeviceManager.MulticastDnsAdapter.AddressUpdatedEventArgs args)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameter
				StaticLogger.LogDebug("args: " + (args != null ? "\r\n" + args.ToString() : "N/A"));

				// check parameters
				if (args == null)
				{
					StaticLogger.LogError("args=NULL");
					return;
				}

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// debug assertion
				Debug.Assert(args != null);
				Debug.Assert(args.Instance != null);

				if (args.Instance.InstanceID.HasValue)
				{
					// if the device exists, update the value; otherwise, create a new one
					Debug.Assert(this.DeviceStore != null);
					var device = this.DeviceStore.AddOrUpdate(args.Instance.ToDevice());
					if (device.PairStatus == AFUserDevice.PairingStatus.Paired)
					{
						Debug.Assert(this.Configuration != null);
						this.Configuration.StoreDevice(device);
					}

					device.PingAsync().ContinueWith(delegate(Task<long> task)
					{
						if (task.IsFaulted || task.Result < 0)
						{
							StaticLogger.LogError("Failed pinging the device '" + device + "', exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception));
						}
						else
						{
							StaticLogger.LogInfo("The device '" + device + "' is connectable in " + task.Result + "ms.");

							// fire DiscoveryEvent
							if (this.IsDiscovering && this.DiscoveryEvent != null)
							{
								CommonUtils.InvokeEvent<DiscoveryEventArgs>(
									new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
									this,
									new DiscoveryEventArgs(device, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
							}

							// fire DeviceUpdate event
							if (device.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceUpdate != null)
							{
								CommonUtils.InvokeEvent<DeviceUpdateEventArgs>(
									new EventHandler<DeviceUpdateEventArgs>(this.DeviceUpdate),
									this,
									new DeviceUpdateEventArgs(device));
							}
						}
					}).ConfigureAwait(continueOnCapturedContext: false);
				}
				else if (args.Instance.IsConnectable()) // no instance ID, query all the information through web service
				{
					StaticLogger.LogWarn("The instance ID of instance '" + args.Instance.InstanceName + "' is unspecified, get information through web service...");
					this.QueryWebServiceForInformationAsync(args.Instance).ContinueWith(delegate(Task<AFUserDevice> task)
					{
						if (task.IsFaulted)
						{
							StaticLogger.LogError("Failed getting information of instance '" + args.Instance.InstanceName + "' through web service, exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception));
						}
						else
						{
							// fire DiscoveryEvent
							if (this.IsDiscovering && this.DiscoveryEvent != null)
							{
								CommonUtils.InvokeEvent<DiscoveryEventArgs>(
									new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
									this,
									new DiscoveryEventArgs(task.Result, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
							}

							// fire DeviceFound event
							if (task.Result.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceFound != null)
							{
								CommonUtils.InvokeEvent<DeviceFoundEventArgs>(
									new EventHandler<DeviceFoundEventArgs>(this.DeviceFound),
									this,
									new DeviceFoundEventArgs(task.Result));
							}

							// TODO:
							// fire specific property changed event
							// fire DeviceUpdate event
							if (task.Result.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceUpdate != null)
							{
								CommonUtils.InvokeEvent<DeviceUpdateEventArgs>(
									new EventHandler<DeviceUpdateEventArgs>(this.DeviceUpdate),
									this,
									new DeviceUpdateEventArgs(task.Result));
							}
						}
					}).ConfigureAwait(continueOnCapturedContext: false);
				}
				else
				{
					StaticLogger.LogWarn("The instance ID of instance '" + args.Instance.InstanceName + "' is unspecified and the instance is unconnectable, skipped.");
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed handling address updated, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when an mDNS instance is found.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="AirFusion.Manager.DeviceManager.UserDeviceManager.MulticastDnsAdapter.ServiceFoundEventArgs"/> instance containing the event data.</param>
		private void OnMulticastDnsAdapterServiceFound(object sender, UserDeviceManager.MulticastDnsAdapter.ServiceFoundEventArgs args)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameter
				StaticLogger.LogDebug("args: " + (args != null ? "\r\n" + args.ToString() : "N/A"));

				// check parameters
				if (args == null)
				{
					StaticLogger.LogError("args=NULL");
					return;
				}

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// debug assertion
				Debug.Assert(args != null);
				Debug.Assert(args.Instance != null);

				if (args.Instance.InstanceID.HasValue)
				{
					// if the device exists, update the value; otherwise, create a new one
					Debug.Assert(this.DeviceStore != null);
					AFUserDevice device = this.DeviceStore.AddOrUpdate(args.Instance.ToDevice());

					// fire DiscoveryEvent
					if (this.IsDiscovering && this.DiscoveryEvent != null)
					{
						CommonUtils.InvokeEvent<DiscoveryEventArgs>(
							new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
							this,
							new DiscoveryEventArgs(device, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
					}

					// fire DeviceFound event
					if (device.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceFound != null)
					{
						CommonUtils.InvokeEvent<DeviceFoundEventArgs>(
							new EventHandler<DeviceFoundEventArgs>(this.DeviceFound),
							this,
							new DeviceFoundEventArgs(device));
					}
				}
				else if (args.Instance.IsConnectable()) // no instance ID, query all the information through web service
				{
					StaticLogger.LogWarn("The instance ID of instance '" + args.Instance.InstanceName + "' is unspecified, get information through web service...");
					this.QueryWebServiceForInformationAsync(args.Instance).ContinueWith(delegate(Task<AFUserDevice> task)
					{
						if (task.IsFaulted)
						{
							StaticLogger.LogError("Failed getting information of instance '" + args.Instance.InstanceName + "' through web service, exception:\r\n" + CommonUtils.GetExceptionMessages(task.Exception));
						}
						else
						{
							// fire DiscoveryEvent
							if (this.IsDiscovering && this.DiscoveryEvent != null)
							{
								CommonUtils.InvokeEvent<DiscoveryEventArgs>(
									new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
									this,
									new DiscoveryEventArgs(task.Result, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
							}

							// fire DeviceFound event
							if (task.Result.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceFound != null)
							{
								CommonUtils.InvokeEvent<DeviceFoundEventArgs>(
									new EventHandler<DeviceFoundEventArgs>(this.DeviceFound),
									this,
									new DeviceFoundEventArgs(task.Result));
							}
						}
					}).ConfigureAwait(continueOnCapturedContext: false);
				}
				else
				{
					StaticLogger.LogWarn("The instance ID of instance '" + args.Instance.InstanceName + "' is unspecified and the instance is unconnectable, skipped.");
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed handling service found, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when an mDNS instance is lost.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="AirFusion.Manager.DeviceManager.UserDeviceManager.MulticastDnsAdapter.ServiceLostEventArgs"/> instance containing the event data.</param>
		private void OnMulticastDnsAdapterServiceLost(object sender, UserDeviceManager.MulticastDnsAdapter.ServiceLostEventArgs args)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameter
				StaticLogger.LogDebug("args: " + (args != null ? "\r\n" + args.ToString() : "N/A"));

				// check parameters
				if (args == null)
				{
					StaticLogger.LogError("args=NULL");
					return;
				}

				// check prerequisite
				if (this._mdnsAdapter == null)
				{
					StaticLogger.LogInfo("Skip since the mDNS adpater is not initialized.");
					return;
				}

				// debug assertion
				Debug.Assert(args != null);
				Debug.Assert(args.Instance != null);

				if (args.Instance.InstanceID.HasValue)
				{
					// if the device exists, update the value; otherwise, create a new one
					Debug.Assert(this.DeviceStore != null);
					AFUserDevice device = this.DeviceStore.GetDevice(args.Instance.InstanceID.Value) ?? args.Instance.ToDevice();

					// TODO:
					//   remove the address from the address pool
					//

					// fire DiscoveryEvent
					if (this.IsDiscovering && this.DiscoveryEvent != null)
					{
						CommonUtils.InvokeEvent<DiscoveryEventArgs>(
							new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent),
							this,
							new DiscoveryEventArgs(device, DiscoveryEventArgs.DiscoveryEventType.DeviceLost));
					}

					// fire DeviceFound event
					if (device.PairStatus == AFUserDevice.PairingStatus.Paired && this.DeviceLost != null)
					{
						CommonUtils.InvokeEvent<DeviceLostEventArgs>(
							new EventHandler<DeviceLostEventArgs>(this.DeviceLost),
							this,
							new DeviceLostEventArgs(device));
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed handling service found, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion MulticastDnsAdapter

		#region DandelionAdapter

		/// <summary>
		/// Called when nearby scan event is callback.
		/// </summary>
		/// <param name="args">The <see cref="DandelionAdapter.NearbyScanCallbackArg"/> instance containing the callback argument data.</param>
		private void OnDandelionAdapterNearbyScanCallback(DandelionAdapter.NearbyScanCallbackArg args)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("args: " + (args != null ? "\r\n" + args.ToString() : "N/A"));

				// check parameters
				if (args == null)
				{
					StaticLogger.LogError("args=null");
					return;
				}
				if (args.IsFaulted || args.Exception != null)
				{
					StaticLogger.LogError("Skipped this callback since the nearby scanning is faulted, exception:\r\n" + CommonUtils.GetExceptionMessages(args.Exception));
					return;
				}

				// check prerequisite
				if (!this.IsDiscovering) // if it's not discovering, skip
				{
					return;
				}

				Debug.Assert(this.DeviceStore != null);
				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
				{
					if (args.InstanceID.HasValue && args.Profile != null)
					{
						// updates values or add the device into the device store if it's not present yet
						Debug.Assert(this.DeviceStore != null);
						AFUserDevice device = this.DeviceStore.AddOrUpdate(args.Profile.ToDevice());

						// fire DiscoveryEvent
						if (this.DiscoveryEvent != null)
						{
							StaticLogger.LogTrace("Fire DiscoveryEvent event...");
							CommonUtils.InvokeEvent<DiscoveryEventArgs>(new EventHandler<DiscoveryEventArgs>(this.DiscoveryEvent), this, new DiscoveryEventArgs(device, DiscoveryEventArgs.DiscoveryEventType.DeviceFoundOrUpdate));
						}
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed processing the nearby scan event, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion DandelionAdapter

		#region Web Service Callbacks

		/// <summary>
		/// Called when a device has sent out the pairing invitation.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="invitationData">The invitation data.</param>
		/// <returns>The response to the request.</returns>
		private PairingInvitationAdapter.WebServiceResponse OnPairingInvited(IPAddress clientIP, PairingInvitationAdapter.PairingInvitationInfo invitationData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("invitationData: " + (invitationData != null ? invitationData.ToString() : "N/A"));

				// check prerequisite
				if (invitationData == null)
				{
					StaticLogger.LogError("The invite data is unspecified.");
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The invitation data is unspecified.");
				}
				if (!invitationData.InstanceID.HasValue)
				{
					StaticLogger.LogError("The instance ID is unspecified.");
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID is unspecified.");
				}
				if (invitationData.InstanceID.Value.CompareTo(Guid.Empty) == 0 || invitationData.InstanceID.Value.CompareTo(this.LocalDevice.InstanceID) == 0)
				{
					StaticLogger.LogError("The instance ID '" + invitationData.InstanceID.Value.GetStringFormat() + "' is invalid.");
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + invitationData.InstanceID.Value.GetStringFormat() + "' is invalid.");
				}

				// get or create the device by instance ID
				AFUserDevice device = null;
				Debug.Assert(this.DeviceStore != null);
				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
				{
					device = this.DeviceStore.GetOrCreateDevice(invitationData.InstanceID.Value);
				}

				// debug assertion
				Debug.Assert(device != null);
				Debug.Assert(invitationData != null);

				// if it's already paired, return accepted
				if (device.PairStatus == AFUserDevice.PairingStatus.Paired)
				{
					StaticLogger.LogWarn("The device '" + invitationData.InstanceID.Value.GetStringFormat() + "' is already paired.");
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}

				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, device)))
				{
					// fill/update values if such values are specified
					{
						device.Address = clientIP;
						if (invitationData.DeviceName != null)
						{
							device.DeviceName = invitationData.DeviceName;
						}
						if (invitationData.ModelName != null)
						{
							device.ModelName = invitationData.ModelName;
						}
						if (invitationData.UserName != null)
						{
							device.UserName = invitationData.UserName;
						}
						if (invitationData.CommunicationPort.HasValue)
						{
							device.WebServicePort = invitationData.CommunicationPort.Value;
						}
					}
				}

				// fire PairingInvited event
				if (this.PairingInvited != null)
				{
					StaticLogger.LogTrace("Fire PairingInvited event...");
					CommonUtils.InvokeEvent<PairingInvitedEventArgs>(new EventHandler<PairingInvitedEventArgs>(this.PairingInvited), this, new PairingInvitedEventArgs(device));
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a device has sent out the pairing request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="requestData">The request data.</param>
		/// <returns>The response to the request.</returns>
		private PairingRequestAdapter.WebServiceResponse OnPairingRequested(IPAddress clientIP, PairingRequestAdapter.PairingRequestInfo requestData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("requestData: " + (requestData != null ? "\r\n" + requestData.ToString() : "N/A"));

				// check prerequisite
				if (requestData == null)
				{
					StaticLogger.LogError("The request data is unspecified.");
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The request data is unspecified.");
				}
				if (!requestData.InstanceID.HasValue)
				{
					StaticLogger.LogError("The instance ID is unspecified.");
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID is unspecified.");
				}
				if (requestData.InstanceID.Value.Equals(Guid.Empty) || requestData.InstanceID.Value.Equals(this.LocalDevice.InstanceID))
				{
					StaticLogger.LogError("The instance ID '" + requestData.InstanceID.Value.GetStringFormat() + "' is invalid.");
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + requestData.InstanceID.Value.GetStringFormat() + "' is invalid.");
				}

				// get or create the device by instance ID
				AFUserDevice device = null;
				Debug.Assert(this.DeviceStore != null);
				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
				{
					device = this.DeviceStore.GetOrCreateDevice(requestData.InstanceID.Value);
				}

				// debug assertion
				Debug.Assert(device != null);
				Debug.Assert(requestData != null);

				var originalPairingStatus = device.PairStatus;

				// fill/update values if such values are specified
				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, device)))
				{
					device.Address = clientIP;
					device.PairStatus = AFUserDevice.PairingStatus.PairingPending; // set PairStatus=PairingPending
					if (requestData.DeviceName != null)
					{
						device.DeviceName = requestData.DeviceName;
					}
					if (requestData.ModelName != null)
					{
						device.ModelName = requestData.ModelName;
					}
					if (requestData.UserName != null)
					{
						device.UserName = requestData.UserName;
					}
					if (requestData.CommunicationPort.HasValue)
					{
						device.WebServicePort = requestData.CommunicationPort.Value;
					}
				}

				if (originalPairingStatus == AFUserDevice.PairingStatus.Paired) // if it's already paired, return accepted
				{
					StaticLogger.LogWarn("The device '" + requestData.InstanceID.Value.GetStringFormat() + "' is already paired, wait 1 second to response with accepted.");

					// response with accepted in 1 second
					Task.Delay(1000).ContinueWith(
						async (task) =>
						{
							try
							{
								var response = PairingResponse.CreatePairingResponse(true);
								response.SessionId = device.SessionID;
								await device.PairingResponseAsync(response).ConfigureAwait(continueOnCapturedContext: false);
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Failed responding the request, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						}, TaskContinuationOptions.ExecuteSynchronously);

					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
				else // otherwise, fire the event
				{
					// fire PairingRequested event
					if (this.PairingRequested != null)
					{
						StaticLogger.LogTrace("Fire PairingRequested event...");
						CommonUtils.BeginInvokeEvent<PairingRequestedEventArgs>(new EventHandler<PairingRequestedEventArgs>(this.PairingRequested), this, new PairingRequestedEventArgs(device));

						return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
					}
					else
					{
						return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing request is responsed.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="responseData">The response data.</param>
		/// <returns>The response to the request.</returns>
		private PairingResponseAdapter.WebServiceResponse OnPairingResponsed(IPAddress clientIP, PairingResponseAdapter.PairingResponseInfo responseData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("responseData: " + (responseData != null ? "\r\n" + responseData.ToString() : "N/A"));

				// check prerequisite
				if (responseData == null)
				{
					StaticLogger.LogError("The request data is unspecified.");
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The response data is unspecified.");
				}
				if (!responseData.InstanceID.HasValue)
				{
					StaticLogger.LogError("The 'InstanceID' is unspecified.");
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The 'InstanceID' is unspecified.");
				}
				if (responseData.InstanceID.Value.CompareTo(Guid.Empty) == 0 || responseData.InstanceID.Value.CompareTo(this.LocalDevice.InstanceID) == 0)
				{
					StaticLogger.LogError("The instance ID '" + responseData.InstanceID.Value.GetStringFormat() + "' is invalid.");
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + responseData.InstanceID.Value.GetStringFormat() + "' is invalid.");
				}
				if (!responseData.Accepted.HasValue)
				{
					StaticLogger.LogError("The 'Accepted' is unspecified.");
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The 'Accepted' is unspecified.");
				}

				// get device from device list
				AFUserDevice device = null;
				Debug.Assert(this.DeviceStore != null);
				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
				{
					device = this.DeviceStore.GetDevice(responseData.InstanceID.Value);

					// check if the device exists
					if (device == null)
					{
						StaticLogger.LogWarn("The device '" + responseData.InstanceID.Value.GetStringFormat() + "' is unrecognized.");
						return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "The device '" + responseData.InstanceID.Value.GetStringFormat() + "' is unrecognized. Please make sure you have a pairing request sent out previously.");
					}
				}

				// debug assertion
				Debug.Assert(device != null);
				Debug.Assert(responseData != null);
				Debug.Assert(responseData.Accepted.HasValue);

				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, device)))
				{
					// check the pairing status
					if (device.PairStatus != AFUserDevice.PairingStatus.PairingPending)
					{
						StaticLogger.LogWarn("The device '" + responseData.InstanceID.Value.GetStringFormat() + "' is not in pairing pending.");
						return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The device '" + responseData.InstanceID.Value.GetStringFormat() + "' is not in pairing pending. Please make sure you have a pairing request sent out previously.");
					}

					// fill/update values
					device.Address = clientIP;
					device.SessionID = (responseData.SessionID != null) ? PairingSessionId.CreateSessionId(responseData.SessionID) : null;
					device.PairStatus = (responseData.Accepted.Value) ? AFUserDevice.PairingStatus.Paired : AFUserDevice.PairingStatus.UnPaired;
				}

				// fire PairingResponsed event
				if (this.PairingResponsed != null)
				{
					StaticLogger.LogTrace("Fire PairingResponsed event...");
					CommonUtils.BeginInvokeEvent<PairingResponsedEventArgs>(new EventHandler<PairingResponsedEventArgs>(this.PairingResponsed), this, new PairingResponsedEventArgs(device, PairingResponsedEventArgs.PairingResponseType.Reserved, responseData.Accepted));

					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing status request is received.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns>The response to the request.</returns>
		private PairingStatusRequestAdapter.WebServiceResponse OnPairingStatusRequested(IPAddress clientIP, Guid instanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("instanceID: " + instanceID.GetStringFormat());

				// check prerequisite
				if (instanceID.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogError("The instance ID '" + instanceID.GetStringFormat() + "' is invalid.");
					return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + instanceID.GetStringFormat() + "' is invalid.");
				}

				Debug.Assert(this.DeviceStore != null);
				var device = this.DeviceStore.GetDevice(instanceID);
				var pairingStatus = new PairingStatusInfo()
				{
					PairingStatus = device != null && device.PairStatus == AFUserDevice.PairingStatus.Paired
				};

				return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.OK, pairingStatus);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when the pairing status of a remote device is updated.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of target device.</param>
		/// <param name="pairingStatusData">The pairing status data.</param>
		/// <returns>The response to the request.</returns>
		private PairingStatusUpdateAdapter.WebServiceResponse OnPairingStatusUpdated(IPAddress clientIP, Guid targetInstanceID, PairingStatusInfo pairingStatusData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("targetInstanceID: " + targetInstanceID.GetStringFormat());
				StaticLogger.LogDebug("pairingStatusData: " + (pairingStatusData != null ? "\r\n" + pairingStatusData.ToString() : "N/A"));

				// check prerequisite
				if (targetInstanceID.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogError("The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
				}
				if (pairingStatusData == null)
				{
					StaticLogger.LogError("The pairing status data is unspecified.");
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The pairing status is unspecified.");
				}
				if (!pairingStatusData.PairingStatus.HasValue)
				{
					StaticLogger.LogError("The 'PairingStatus' is unspecified.");
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The 'PairingStatus' is unspecified.");
				}

				// debug assertion
				Debug.Assert(pairingStatusData != null);
				Debug.Assert(pairingStatusData.PairingStatus.HasValue);
				Debug.Assert(this.DeviceStore != null);
				Debug.Assert(this.Configuration != null);

				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, this.DeviceStore)))
				{
					AFUserDevice device = this.DeviceStore.GetDevice(targetInstanceID);
					if (device != null)
					{
						StaticLogger.LogInfo("Set pairing status of '" + targetInstanceID.GetStringFormat() + "' to '" + (pairingStatusData.PairingStatus.Value ? AFUserDevice.PairingStatus.Paired.ToString() : AFUserDevice.PairingStatus.UnPaired.ToString()) + "'...");
						device.PairStatus = pairingStatusData.PairingStatus.Value ? AFUserDevice.PairingStatus.Paired : AFUserDevice.PairingStatus.UnPaired;

						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
					}
					else if (pairingStatusData.PairingStatus.Value == true)
					{
						StaticLogger.LogInfo("The device '" + targetInstanceID.GetStringFormat() + "' doesn't exist, make it paired it on background...");
						device = this.DeviceStore.GetOrCreateDevice(targetInstanceID);
						device.PairStatus = AFUserDevice.PairingStatus.Paired;

						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
					}
					else
					{
						StaticLogger.LogWarn("The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized.");

						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized. Please make sure the both two device is paired.");
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when some device is requesting the device information of some specified instance ID. Typically, the instance ID will be the same as the one of current device.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID of requested device.</param>
		/// <returns>The response to the request.</returns>
		private DeviceInformationRequestAdapter.WebServiceResponse OnDeviceInformationRequested(IPAddress clientIP, Guid? instanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("instanceID: " + (instanceID.HasValue ? instanceID.Value.GetStringFormat() : "N/A"));

				// debug assertion
				Debug.Assert(this.LocalDevice != null);

				// get device
				AFUserDevice device = (!instanceID.HasValue) || (this.LocalDevice != null && instanceID.Value.CompareTo(this.LocalDevice.InstanceID) == 0) ? this.LocalDevice : this.DeviceStore.GetDevice(instanceID.Value);

				// create response and return
				if (device != null)
				{
					var response = new DeviceInformation();
					device.FillTo(response);

					return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.OK, response);
				}
				else
				{
					if (instanceID.HasValue)
					{
						return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "Cannot find the device '" + instanceID.Value.GetStringFormat() + "'.");
					}
					else
					{
						return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Failed getting information of local device.");
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when the device information of a specific device is updated.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of device to be updated.</param>
		/// <param name="information">The updated device information.</param>
		/// <returns>The response to the request.</returns>
		private DeviceInformationUpdateAdapter.WebServiceResponse OnDeviceInformationUpdated(IPAddress clientIP, Guid targetInstanceID, DeviceInformation information)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("clientIP: " + (clientIP != null ? clientIP.ToString() : "N/A"));
				StaticLogger.LogDebug("targetInstanceID: " + targetInstanceID.GetStringFormat());
				StaticLogger.LogDebug("information:\r\n" + information.ToString());

				// check prerequisite
				if (!information.InstanceID.HasValue)
				{
					StaticLogger.LogError("The instance ID is unspecified.");
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID is unspecified");
				}
				if (targetInstanceID.Equals(information.InstanceID) == false)
				{
					StaticLogger.LogError("Ambiguous instance ID '" + targetInstanceID.GetStringFormat() + "' & '" + (information.InstanceID.HasValue ? information.InstanceID.Value.GetStringFormat() : "N/A") + "'.");
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "Ambiguous instance ID '" + targetInstanceID.GetStringFormat() + "' & '" + (information.InstanceID.HasValue ? information.InstanceID.Value.GetStringFormat() : "N/A") + "'.");
				}
				if (targetInstanceID.Equals(Guid.Empty))
				{
					StaticLogger.LogError("The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The instance ID '" + targetInstanceID.GetStringFormat() + "' is invalid.");
				}
				if (targetInstanceID.CompareTo(this.LocalDevice.InstanceID) == 0)
				{
					StaticLogger.LogError("The device information of local device can't be updated by remote device.");
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.Unauthorized, "The device information of local device can't be updated by remote device.");
				}

				// get device from device list
				Debug.Assert(this.DeviceStore != null);
				AFUserDevice device = this.DeviceStore.GetDevice(targetInstanceID);

				// check if the device exists
				if (device == null)
				{
					StaticLogger.LogWarn("The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized.");
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotFound, "The device '" + targetInstanceID.GetStringFormat() + "' is unrecognized. Please make sure you have both devices paired previously.");
				}

				// debug assertion
				Debug.Assert(device != null);

				lock (CommonUtils.GetLocker(new Tuple<object, object>(this, device)))
				{
					// update the address if it's not specified
					if ((device.Address == null || device.Address.Equals(IPAddress.None)) && clientIP != null && !clientIP.Equals(IPAddress.None))
					{
						device.Address = clientIP;
					}

					// fill/update values if such values are specified
					device.FillFromIfSpecified(information);

					// update the configuration file if the PairStatus=Paired
					if (device.PairStatus == AFUserDevice.PairingStatus.Paired && this.Configuration != null)
					{
						this.Configuration.StoreDevice(device);
					}
				}

				// fire DeviceUpdate event
				if (this.DeviceUpdate != null)
				{
					// TODO:
					// fire event only when updated

					StaticLogger.LogTrace("Fire DeviceUpdate event...");
					CommonUtils.BeginInvokeEvent<DeviceUpdateEventArgs>(new EventHandler<DeviceUpdateEventArgs>(this.DeviceUpdate), this, new DeviceUpdateEventArgs(device));

					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
				}
				else
				{
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.Accepted);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Web Service Callbacks

		#region AFUserDevice

		/// <summary>
		/// Called when the current device is paired with another device.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="eventArgs">The <see cref="DevicePairedEventArgs" /> instance containing the event data.</param>
		internal void OnDevicePaired(object sender, DevicePairedEventArgs eventArgs)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("sender: " + (sender != null ? sender.ToString() : "N/A"));
				StaticLogger.LogDebug("eventArgs: " + (eventArgs != null ? eventArgs.ToString() : "N/A"));

				// check prerequisite
				if (eventArgs == null || eventArgs.Device == null)
				{
					StaticLogger.LogWarn("The parameters is incorrect.");
				}

				// update configuration
				if (this.Configuration != null)
				{
					this.Configuration.StoreDevice(eventArgs.Device);
				}

				// fire DevicePaired event
				if (this.DevicePaired != null)
				{
					StaticLogger.LogTrace("Fire DevicePaired event...");
					CommonUtils.InvokeEvent<DevicePairedEventArgs>(new EventHandler<DevicePairedEventArgs>(this.DevicePaired), this, new DevicePairedEventArgs(eventArgs.Device));
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Unhandled exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when the current device is un-paired with another device.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="eventArgs">The <see cref="DeviceUnPairedEventArgs" /> instance containing the event data.</param>
		internal void OnDeviceUnPaired(object sender, DeviceUnPairedEventArgs eventArgs)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("sender: " + (sender != null ? sender.ToString() : "N/A"));
				StaticLogger.LogDebug("eventArgs: " + (eventArgs != null ? eventArgs.ToString() : "N/A"));

				// check prerequisite
				if (eventArgs == null || eventArgs.Device == null)
				{
					StaticLogger.LogWarn("The parameters is incorrect.");
				}

				// update configuration
				try
				{
					if (this.Configuration != null)
					{
						this.Configuration.RemoveDevice(eventArgs.Device);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed removing device from Registry, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}

				// fire DeviceUnPaired event
				if (this.DeviceUnPaired != null)
				{
					StaticLogger.LogTrace("Fire DeviceUnPaired event...");
					CommonUtils.InvokeEvent<DeviceUnPairedEventArgs>(new EventHandler<DeviceUnPairedEventArgs>(this.DeviceUnPaired), this, new DeviceUnPairedEventArgs(eventArgs.Device));
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Unhandled exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion AFUserDevice
	}
}
