﻿using AirFusion.Common;
using Bonjour;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net.Sockets;
using System.Threading;

namespace AirFusion.Manager.DeviceManager
{
	#region Properties

	public partial class DnsSdHandler
	{
		/// <summary>
		/// Gets the service name of the DNS-SD service. ex: _air-fusion
		/// </summary>
		public string ServiceName { get; protected set; }

		/// <summary>
		/// Gets the root domain of the DNS-SD service.
		/// </summary>
		public string Domain { get; protected set; }

		/// <summary>
		/// Gets the service type of the DNS-SD service. ex: _air-fusion._tcp
		/// </summary>
		public string ServiceType { get { return this.ServiceName + "." + this.ConnProtocol.GetStringValue(); } }

		/// <summary>
		/// Gets the connection protocol of the DNS-SD service. ex: _tcp
		/// </summary>
		public ConnectionProtocol ConnProtocol { get; protected set; }

		/// <summary>
		/// Gets the collection of registered instance names.
		/// </summary>
		public List<string> RegisteredInstanceNames { get; protected set; }

		/// <summary>
		/// Gets or sets the indication whether to hide self service.
		/// </summary>
		public bool HideSelf { get; set; }
	}

	#endregion Properties

	#region Public Methods

	/// <summary>
	/// Provides the functionality of DNS service discovery, registeration, &amp; query.
	/// </summary>
	public partial class DnsSdHandler
	{
		/// <summary>
		/// Initializes a instance of the DNS-SD handler.
		/// </summary>
		/// <param name="serviceName">The service name. e.g. _air-fusion</param>
		/// <param name="connProtocol">The connection protocol. (_tcp or _udp)</param>
		/// <param name="domain">The domain name. Specify null for default domain.</param>
		/// <exception cref="System.ArgumentException" />
		/// <remarks>
		/// - Suggest not to carry '.' in instanceName since Android library (jmdns) cannot handle this case.
		/// - Please instantiate the instance in the main thread since an STA COM is utilized.
		/// </remarks>
		public DnsSdHandler(string serviceName, ConnectionProtocol connProtocol, string domain)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("serviceName: " + serviceName);
				StaticLogger.LogDebug("connProtocol: " + connProtocol);
				StaticLogger.LogDebug("domain: " + domain);

				// check parameters
				if (string.IsNullOrWhiteSpace(serviceName))
				{
					throw new ArgumentException("Please specify a valid service name.", "serviceName");
				}

				this.ServiceName = serviceName;
				this.ConnProtocol = connProtocol;
				this.Domain = (domain ?? DefaultDomain).TrimEnd();
				this.RegisteredInstanceNames = new List<string>();
				this.HideSelf = false;
				this._eventManager = null;
				this._processorID = 0;
				this._processorDic = new ConcurrentDictionary<uint, ProcessorData>();
				this._synchronizationContext = null;

				if (SynchronizationContext.Current == null)
				{
					throw new InvalidOperationException("Please make sure the instance is initialized by main thread with message loop.");
				}

				this._synchronizationContext = SynchronizationContext.Current;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Initializes this instance.
		/// </summary>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		public void Initialize()
		{
			try
			{
				StaticLogger.TrackEnter();

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						handler._eventManager = new DNSSDEventManager();
						handler._eventManager.ServiceRegistered += handler.OnServiceRegistered;
						handler._eventManager.ServiceFound += (browser, flags, ifIndex, instanceName, serviceType, dom) =>
						{
							handler.OnBrowserEvent(browser, flags, ifIndex, instanceName, serviceType, dom, BrowserCallbackArgs.BrowserEventType.Found);
						};
						handler._eventManager.ServiceLost += (browser, flags, ifIndex, instanceName, serviceType, dom) =>
						{
							handler.OnBrowserEvent(browser, flags, ifIndex, instanceName, serviceType, dom, BrowserCallbackArgs.BrowserEventType.Lost);
						};
						handler._eventManager.ServiceResolved += handler.OnServiceResolved;
						handler._eventManager.AddressFound += handler.OnAddressFound;
						handler._eventManager.QueryRecordAnswered += handler.OnQueryRecordAnswered;
						handler._eventManager.OperationFailed += handler.OnOperationFailed;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed initializing, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Registers a service.
		/// </summary>
		/// <param name="instanceName">The instance name. Specify null for default.</param>
		/// <param name="host">The host name. Specify null for default host name.</param>
		/// <param name="port">The port number of the service.</param>
		/// <param name="txtRecord">The txt record.</param>
		/// <param name="interfaceIndex">The index of interface, 0 for all interfaces.</param>
		/// <param name="callback">The callback that will be invoked for the operation.</param>
		/// <returns>The processor ID assoicated with the registrar.</returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		/// <remarks>Only one instance can be registered in a <see cref="DnsSdHandler"/> instance.</remarks>
		public virtual uint RegisterService(string instanceName, string host, ushort port, TxtRecord txtRecord, uint interfaceIndex, RegistrarCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("instanceName: " + instanceName);
				StaticLogger.LogDebug("host: " + host);
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("txtRecord:" + (txtRecord != null ? "\r\n" + txtRecord.ToString() : "N/A"));
				StaticLogger.LogDebug("interfaceIndex: " + interfaceIndex);

				// check parameters
				if (port == 0)
				{
					throw new ArgumentException("Please specify the port.", "port");
				}
				if (txtRecord == null)
				{
					throw new ArgumentNullException("txtRecord");
				}

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// get available processor ID
				uint processorID = uint.MaxValue;
				if (this.GetAnAvailableProcessorID(out processorID, new RegistrarData()
				{
					Service = null,
					Processor = null,
					InstanceName = instanceName,
					Hostname = host,
					Port = port,
					InterfaceIndex = interfaceIndex,
					TxtRecord = txtRecord,
					Callback = callback
				}) == false)
				{
					throw new Exception("All processor slot are in use.");
				}

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						if (handler._processorDic.ContainsKey(processorID) && handler._processorDic[processorID] != null && handler._processorDic[processorID] is RegistrarData)
						{
							if (handler._processorDic[processorID].State == ProcessorState.ScheduleToRun)
							{
								// register the service
								handler._processorDic[processorID].Service = new DNSSDService();
								StaticLogger.LogDebug("HashCode (service): " + handler._processorDic[processorID].Service.GetHashCode());
								handler._processorDic[processorID].Processor = handler._processorDic[processorID].Service.Register(0, interfaceIndex, instanceName, handler.ServiceType, handler.Domain + ".", host, port, txtRecord.Transform(), handler._eventManager);
								StaticLogger.LogDebug("HashCode (registrar): " + handler._processorDic[processorID].Processor.GetHashCode());

								handler._processorDic[processorID].State = ProcessorState.Running;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed registering the service instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						handler.RemoveProcessorID(processorID);
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);

				return processorID;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts the browse of service instances.
		/// </summary>
		/// <param name="interfaceIndex">The index of interface, 0 for all interfaces.</param>
		/// <param name="callback">The callback that will be invoked for the operation.</param>
		/// <returns>The processor ID assoicated with the browser.</returns>
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		public virtual uint StartBrowseService(uint interfaceIndex, BrowserCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("interfaceIndex: " + interfaceIndex);

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// get available processor ID
				uint processorID = uint.MaxValue;
				if (this.GetAnAvailableProcessorID(out processorID, new BrowserData()
				{
					Service = null,
					Processor = null,
					InterfaceIndex = interfaceIndex,
					Callback = callback
				}) == false)
				{
					throw new Exception("All processor slot are in use.");
				}

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						if (handler._processorDic.ContainsKey(processorID) && handler._processorDic[processorID] != null && handler._processorDic[processorID] is BrowserData)
						{
							if (handler._processorDic[processorID].State == ProcessorState.ScheduleToRun)
							{
								// create a new service and start browsing
								handler._processorDic[processorID].Service = new DNSSDService();
								StaticLogger.LogDebug("HashCode (service): " + handler._processorDic[processorID].Service.GetHashCode());
								handler._processorDic[processorID].Processor = handler._processorDic[processorID].Service.Browse(0, interfaceIndex, handler.ServiceType, handler.Domain, handler._eventManager);
								StaticLogger.LogDebug("HashCode (browser): " + handler._processorDic[processorID].Processor.GetHashCode());

								handler._processorDic[processorID].State = ProcessorState.Running;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed starting a browser, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						handler.RemoveProcessorID(processorID);
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);

				return processorID;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts to resolve a specific service instance.
		/// </summary>
		/// <param name="instanceName">The instance name to query.</param>
		/// <param name="interfaceIndex">The index of interface, 0 for all interfaces.</param>
		/// <param name="forceMulticast">A value indicates whether to perform a multicast query.</param>
		/// <param name="callback">The callback that will be invoked for the operation.</param>
		/// <returns>The processor ID assoicated with the resolver.</returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		public virtual uint StartResolveService(string instanceName, uint interfaceIndex, bool forceMulticast, ResolverCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("instanceName: " + instanceName);
				StaticLogger.LogDebug("interfaceIndex: " + interfaceIndex);
				StaticLogger.LogDebug("forceMulticast: " + forceMulticast);

				// check parameters
				if (string.IsNullOrEmpty(instanceName))
				{
					throw new ArgumentException("Please specify the instance name to resolve.", "instanceName");
				}
				if (callback == null)
				{
					throw new ArgumentNullException("callback");
				}

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// get available processor ID
				uint processorID = uint.MaxValue;
				if (this.GetAnAvailableProcessorID(out processorID, new ResolverData()
				{
					Service = null,
					Processor = null,
					InstanceName = instanceName,
					InterfaceIndex = interfaceIndex,
					ForceMulticast = forceMulticast,
					Callback = callback
				}) == false)
				{
					throw new Exception("All processor slot are in use.");
				}

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						// flags
						DNSSDFlags resolverFlags = 0;
						if (forceMulticast)
						{
							resolverFlags |= DNSSDFlags.kDNSSDFlagsForceMulticast;
						}

						if (handler._processorDic.ContainsKey(processorID) && handler._processorDic[processorID] != null && handler._processorDic[processorID] is ResolverData)
						{
							if (handler._processorDic[processorID].State == ProcessorState.ScheduleToRun)
							{
								// create a new service and start resolving
								handler._processorDic[processorID].Service = new DNSSDService();
								StaticLogger.LogDebug("HashCode (service): " + handler._processorDic[processorID].Service.GetHashCode());
								handler._processorDic[processorID].Processor = handler._processorDic[processorID].Service.Resolve(resolverFlags, interfaceIndex, instanceName, handler.ServiceType, handler.Domain, handler._eventManager);
								StaticLogger.LogDebug("HashCode (resolver): " + handler._processorDic[processorID].Processor.GetHashCode());

								handler._processorDic[processorID].State = ProcessorState.Running;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed resolving the service instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						handler.RemoveProcessorID(processorID);
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);

				return processorID;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts to query the specific record of the specified service instance.
		/// </summary>
		/// <param name="instanceName">The service instance name to query.</param>
		/// <param name="rrType">The RR type to query.</param>
		/// <param name="method">The query method.</param>
		/// <param name="interfaceIndex">The index of interface, 0 for all interfaces.</param>
		/// <param name="callback">The callback that will be invoked for the operation.</param>
		/// <returns>The processor ID assoicated with the querier.</returns>
		/// <remarks>'A' &amp; 'AAAA' records cannot be queried by this method, please use <see cref="StartQueryAddress"/> instead.</remarks>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		public virtual uint StartQueryRecord(string instanceName, DnsSdRRType rrType, QueryRecordMethod method, uint interfaceIndex, RecordQuerierCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("instanceName: " + instanceName);
				StaticLogger.LogDebug("rrType: " + rrType);
				StaticLogger.LogDebug("method: " + method);
				StaticLogger.LogDebug("interfaceIndex: " + interfaceIndex);

				// check parameters
				if (string.IsNullOrEmpty(instanceName))
				{
					throw new ArgumentException("Please specify the instance name to query resource record.", "instanceName");
				}
				if (rrType == DnsSdRRType.Unknown)
				{
					throw new ArgumentException("Please specify a valid RR type.", "rrType");
				}
				if (rrType == DnsSdRRType.TypeA || rrType == DnsSdRRType.TypeAAAA)
				{
					throw new ArgumentException("'A' & 'AAAA' are not supported.");
				}
				if (callback == null)
				{
					throw new ArgumentNullException("callback");
				}

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// get available processor ID
				uint processorID = uint.MaxValue;
				if (this.GetAnAvailableProcessorID(out processorID, new RecordQuerierData()
				{
					Service = null,
					Processor = null,
					InstanceName = instanceName,
					InterfaceIndex = interfaceIndex,
					RRType = rrType,
					QueryMethod = method,
					Callback = callback
				}) == false)
				{
					throw new Exception("All processor slot are in use.");
				}

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						// construct the full name
						string fullName = EscapeInstanceName(instanceName) + "." + handler.ServiceType + "." + handler.Domain + ".";

						// flags
						DNSSDFlags queryFlags = DNSSDFlags.kDNSSDFlagsReturnIntermediates;
						if (method == QueryRecordMethod.ForceMulticast)
						{
							queryFlags |= DNSSDFlags.kDNSSDFlagsForceMulticast;
						}
						else if (method == QueryRecordMethod.LongLivedQuery)
						{
							queryFlags |= DNSSDFlags.kDNSSDFlagsLongLivedQuery;
						}

						if (handler._processorDic.ContainsKey(processorID) && handler._processorDic[processorID] != null && handler._processorDic[processorID] is RecordQuerierData)
						{
							if (handler._processorDic[processorID].State == ProcessorState.ScheduleToRun)
							{
								// create a new service and start querying
								handler._processorDic[processorID].Service = new DNSSDService();
								StaticLogger.LogDebug("HashCode (service): " + handler._processorDic[processorID].Service.GetHashCode());
								handler._processorDic[processorID].Processor = handler._processorDic[processorID].Service.QueryRecord(queryFlags, interfaceIndex, fullName, rrType.Transform(), DNSSDRRClass.kDNSSDClass_IN, handler._eventManager);
								StaticLogger.LogDebug("HashCode (querier): " + handler._processorDic[processorID].Processor.GetHashCode());

								handler._processorDic[processorID].State = ProcessorState.Running;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed querying '" + rrType.ToString() + "' records of instance '" + instanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						handler.RemoveProcessorID(processorID);
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);

				return processorID;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts to query the address of the specified hostname and the address family.
		/// </summary>
		/// <param name="hostname">The fully-qualified hostname.</param>
		/// <param name="addressFamily">The address family.</param>
		/// <param name="method">The query method.</param>
		/// <param name="interfaceIndex">The index of interface, 0 for all interfaces.</param>
		/// <param name="callback">The callback that will be invoked for the operation.</param>
		/// <returns>The processor ID assoicated with the querier.</returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		public virtual uint StartQueryAddress(string hostname, AddressFamily addressFamily, AddrQueryMethod method, uint interfaceIndex, AddressQuerierCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("hostname: " + hostname);
				StaticLogger.LogDebug("addressFamily: " + addressFamily);
				StaticLogger.LogDebug("method: " + method);
				StaticLogger.LogDebug("interfaceIndex: " + interfaceIndex);

				// check parameters
				if (string.IsNullOrEmpty(hostname))
				{
					throw new ArgumentException("Please specify the hostname to query address.", "hostname");
				}
				if (addressFamily != AddressFamily.InterNetwork && addressFamily != AddressFamily.InterNetworkV6)
				{
					throw new ArgumentException("Please specify a valid address family (must be IPv4 or IPv6).", "addressFamily");
				}
				if (callback == null)
				{
					throw new ArgumentNullException("callback");
				}

				// make sure the service is running
				CommonUtils.MakeSureServiceIsRunning(BonjourServiceName, ServiceTimeout);

				// get available processor ID
				uint processorID = uint.MaxValue;
				if (this.GetAnAvailableProcessorID(out processorID, new AddrQuerierData()
				{
					Service = null,
					Processor = null,
					Hostname = hostname,
					InterfaceIndex = interfaceIndex,
					AddressFamily = addressFamily,
					QueryMethod = method,
					Callback = callback
				}) == false)
				{
					throw new Exception("All processor slot are in use.");
				}

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					try
					{
						StaticLogger.TrackEnter();

						// flags
						DNSSDFlags queryFlags = 0;
						if (method == AddrQueryMethod.ForceMulticast)
						{
							queryFlags |= DNSSDFlags.kDNSSDFlagsForceMulticast;
						}
						else if (method == AddrQueryMethod.LongLivedQuery)
						{
							queryFlags |= DNSSDFlags.kDNSSDFlagsLongLivedQuery;
						}

						if (handler._processorDic.ContainsKey(processorID) && handler._processorDic[processorID] != null && handler._processorDic[processorID] is AddrQuerierData)
						{
							if (handler._processorDic[processorID].State == ProcessorState.ScheduleToRun)
							{
								// create a new service and start querying
								handler._processorDic[processorID].Service = new DNSSDService();
								StaticLogger.LogDebug("HashCode (service): " + handler._processorDic[processorID].Service.GetHashCode());
								handler._processorDic[processorID].Processor = handler._processorDic[processorID].Service.GetAddrInfo(queryFlags, interfaceIndex, addressFamily.Transform(), hostname, handler._eventManager);
								StaticLogger.LogDebug("HashCode (querier): " + handler._processorDic[processorID].Processor.GetHashCode());

								handler._processorDic[processorID].State = ProcessorState.Running;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed querying '" + addressFamily.ToString() + "' address of host '" + hostname + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						handler.RemoveProcessorID(processorID);
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}, this);

				return processorID;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Determines if the specified ID is a valid processor ID and is under browsing.
		/// </summary>
		/// <param name="processorID">The processor ID.</param>
		/// <returns>An indication of whether the specified ID is under processing.</returns>
		public virtual bool IsBrowsing(uint processorID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("processorID: " + processorID);

				Debug.Assert(this._processorDic != null);
				return this._processorDic.ContainsKey(processorID) && this._processorDic[processorID] != null && this._processorDic[processorID] is BrowserData;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Determines if the specified ID is a valid processor ID and is under resolving.
		/// </summary>
		/// <param name="processorID">The processor ID.</param>
		/// <returns>An indication of whether the specified ID is under processing.</returns>
		public virtual bool IsResolving(uint processorID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("processorID: " + processorID);

				Debug.Assert(this._processorDic != null);
				return this._processorDic.ContainsKey(processorID) && this._processorDic[processorID] != null && this._processorDic[processorID] is ResolverData;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Determines if the specified ID is a valid processor ID and is under record querying.
		/// </summary>
		/// <param name="processorID">The processor ID.</param>
		/// <returns>An indication of whether the specified ID is under processing.</returns>
		public virtual bool IsRecordQuerying(uint processorID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("processorID: " + processorID);

				Debug.Assert(this._processorDic != null);
				return this._processorDic.ContainsKey(processorID) && this._processorDic[processorID] != null && this._processorDic[processorID] is RecordQuerierData;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Determines if the specified ID is a valid processor ID and is under address querying.
		/// </summary>
		/// <param name="processorID">The processor ID.</param>
		/// <returns>An indication of whether the specified ID is under processing.</returns>
		public virtual bool IsAddressQuerying(uint processorID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("processorID: " + processorID);

				Debug.Assert(this._processorDic != null);
				return this._processorDic.ContainsKey(processorID) && this._processorDic[processorID] != null && this._processorDic[processorID] is AddrQuerierData;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets an enumerator of the processor ID of browsers.
		/// </summary>
		/// <returns></returns>
		public virtual IEnumerable<uint> GetBrowsers()
		{
			Debug.Assert(this._processorDic != null);
			return this._processorDic.Where(data => data.Value is BrowserData).Select(data => data.Key);
		}

		/// <summary>
		/// Gets an enumerator of the processor ID of resolvers.
		/// </summary>
		/// <returns></returns>
		public virtual IEnumerable<uint> GetResolvers()
		{
			Debug.Assert(this._processorDic != null);
			return this._processorDic.Where(data => data.Value is ResolverData).Select(data => data.Key);
		}

		/// <summary>
		/// Gets an enumerator of the processor ID of record queriers.
		/// </summary>
		/// <returns></returns>
		public virtual IEnumerable<uint> GetRecordQueriers()
		{
			Debug.Assert(this._processorDic != null);
			return this._processorDic.Where(data => data.Value is RecordQuerierData).Select(data => data.Key);
		}

		/// <summary>
		/// Gets an enumerator of the processor ID of address queriers.
		/// </summary>
		/// <returns></returns>
		public virtual IEnumerable<uint> GetAddrQueriers()
		{
			Debug.Assert(this._processorDic != null);
			return this._processorDic.Where(data => data.Value is AddrQuerierData).Select(data => data.Key);
		}

		/// <summary>
		/// Stops the specified processor with the processor ID.
		/// </summary>
		/// <param name="processorID">The processor ID to stop.</param>
		/// <exception cref="System.ArgumentException" />
		public virtual void StopProcessor(uint processorID)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("processorID: " + processorID);

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					if (this._processorDic.ContainsKey(processorID) == false)
					{
						StaticLogger.LogWarn("The processor ID '" + processorID + "' doesn't exist.");
					}

					if (this._processorDic[processorID] != null)
					{
						try
						{
							if (this._processorDic[processorID].Service != null)
							{
								this._processorDic[processorID].Service.Stop();
								StaticLogger.LogTrace("DNSSDService.Stop() (Service, HasCode=" + this._processorDic[processorID].Service.GetHashCode() + ") finished.");
							}
							if (this._processorDic[processorID].Processor != null)
							{
								this._processorDic[processorID].Processor.Stop();
								StaticLogger.LogTrace("DNSSDService.Stop() (Processor, HasCode=" + this._processorDic[processorID].Processor.GetHashCode() + ") finished.");
							}
						}
						catch (Exception ex)
						{
							StaticLogger.LogWarn("DNSSDService.Stop() failed, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						}
					}

					this.RemoveProcessorID(processorID);
				}, this);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
	}

	#endregion Public Methods

	#region Disposable

	public partial class DnsSdHandler : IDisposable
	{
		internal bool _disposed = false;

		internal virtual void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					try
					{
						this.StopAllProcessors();
					}
					catch { }
				}

				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}
	}

	#endregion Disposable
}