﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service manager of device information (/information/*).
	/// </summary>
	internal sealed class DeviceInformationService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle device information request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID of requested device.</param>
		/// <returns>The response to the request.</returns>
		public delegate DeviceInformationRequestAdapter.WebServiceResponse DeviceInformationRequestCallback(IPAddress clientIP, Guid? instanceID);

		/// <summary>
		/// Represents the callback function that will handle device information update request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of device to be updated.</param>
		/// <param name="information">The updated device information.</param>
		public delegate DeviceInformationUpdateAdapter.WebServiceResponse DeviceInformationUpdateCallback(IPAddress clientIP, Guid targetInstanceID, DeviceInformation information);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private DeviceInformationRequestAdapter _requestAdapter;
		private DeviceInformationUpdateAdapter _updateAdapter;
		private DeviceInformationRequestCallback _requestCallback;
		private DeviceInformationUpdateCallback _updateCallback;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="DeviceInformationService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public DeviceInformationService(WebServicesManager webServiceManager)
		{
			this._requestAdapter = new DeviceInformationRequestAdapter(webServiceManager);
			this._updateAdapter = new DeviceInformationUpdateAdapter(webServiceManager);

			this._requestAdapter.RegisterRequestCallback(this.OnRequestAdapterRequestReceived);
			this._updateAdapter.RegisterRequestCallback(this.OnUpdateAdapterRequestReceived);

			this._requestCallback = null;
			this._updateCallback = null;
		}

		/// <summary>
		/// Registers the callback function of device information request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterRequestCallback(DeviceInformationRequestCallback callback)
		{
			this._requestCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of device information update request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterUpdateCallback(DeviceInformationUpdateCallback callback)
		{
			this._updateCallback = callback;
		}

		/// <summary>
		/// Requests the device information from the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="instanceID">The instance ID of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<DeviceInformationRequestAdapter.WebServiceResponse> RequestAsync(IPAddress address, ushort port, Guid? instanceID = null)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._requestAdapter != null);
				return await this._requestAdapter.RequestAsync(address, port, instanceID).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Updates the device information to the target device asynchronously.
		/// </summary>
		/// <param name="address">The address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="instanceID">The instance ID of the target device.</param>
		/// <param name="information">The information of the device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<DeviceInformationUpdateAdapter.WebServiceResponse> UpdateAsync(IPAddress address, ushort port, Guid instanceID, DeviceInformation information)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._updateAdapter != null);
				return await this._updateAdapter.UpdateAsync(address, port, information).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a device information request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private DeviceInformationRequestAdapter.WebServiceResponse OnRequestAdapterRequestReceived(IPAddress address, ushort port, DeviceInformationRequestAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._requestCallback == null)
				{
					return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}

				// get instance ID
				Guid? instanceID = null;
				{
					string webServiceUriTrimSlashes = DeviceInformationRequestAdapter.WebServiceUri.TrimStart('/');
					string requestUriTrimSlashes = remoteRequest.RequestUri.ToString().TrimStart('/');

					if (!requestUriTrimSlashes.StartsWith(webServiceUriTrimSlashes))
					{
						Trace.Fail("Shouldn't be here, there should be something wrong.");
						return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Unknown error.");
					}

					string postfix = requestUriTrimSlashes.Substring(webServiceUriTrimSlashes.Length).Trim('/');
					var postfixTokens = postfix.Split('/');

					if (postfixTokens != null && !string.IsNullOrEmpty(postfixTokens[0]))
					{
						Guid guid;
						if (!Guid.TryParse(postfixTokens[0], out guid))
						{
							StaticLogger.LogWarn("Bad instance ID: " + postfixTokens[0]);
							return DeviceInformationRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is invalid.");
						}
						instanceID = guid;
					}
				}

				// invoke callback function
				return this._requestCallback(address, instanceID);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a device information update is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private DeviceInformationUpdateAdapter.WebServiceResponse OnUpdateAdapterRequestReceived(IPAddress address, ushort port, DeviceInformationUpdateAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				// check prerequisite
				if (this._updateCallback == null)
				{
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}

				// get instance ID
				Guid instanceID;
				{
					string webServiceUriTrimSlashes = DeviceInformationUpdateAdapter.WebServiceUri.TrimStart('/');
					string requestUriTrimSlashes = remoteRequest.RequestUri.ToString().TrimStart('/');

					if (!requestUriTrimSlashes.StartsWith(webServiceUriTrimSlashes))
					{
						StaticLogger.LogError("Unknown error.");
						return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Unknown error.");
					}

					string postfix = requestUriTrimSlashes.Substring(webServiceUriTrimSlashes.Length).Trim('/');
					var postfixTokens = postfix.Split('/');

					if (postfixTokens == null || string.IsNullOrEmpty(postfixTokens[0]))
					{
						StaticLogger.LogWarn("No instance ID is specified.");
						return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified.");
					}

					if (!Guid.TryParse(postfixTokens[0], out instanceID))
					{
						StaticLogger.LogWarn("Bad instance ID: " + postfixTokens[0]);
						return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is invalid.");
					}
				}

				// invoke callback function
				try
				{
					return this._updateCallback(address, instanceID, remoteRequest.Request);
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Unhandled exception of callback function, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					return DeviceInformationUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Failed processing the request.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
