﻿using AirFusion.Common;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of device information (update) (PUT /information/{InstanceID}).
	/// </summary>
	internal sealed class DeviceInformationUpdateAdapter : WebServiceAdapter<DeviceInformation, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/information";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="DeviceInformationUpdateAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public DeviceInformationUpdateAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Updates the device information to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="information">The information of the device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> UpdateAsync(IPAddress address, ushort port, DeviceInformation information)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (information == null)
			{
				throw new ArgumentNullException("information");
			}
			if (information.InstanceID.HasValue == false)
			{
				throw new ArgumentNullException("information.InstanceID", "The instance ID of the specified information is unspecified.");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + "/" + information.InstanceID.Value.GetStringFormat()).Uri,
				information);

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
