﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service manager of device pairing (/pairing/*).
	/// </summary>
	internal sealed partial class PairingService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle pairing invitation.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="invitationData">The invitation data.</param>
		/// <returns>The response to the request.</returns>
		public delegate PairingInvitationAdapter.WebServiceResponse PairingInvitationCallback(IPAddress clientIP, PairingInvitationAdapter.PairingInvitationInfo invitationData);

		/// <summary>
		/// Represents the callback function that will handle pairing request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="requestData">The request data.</param>
		/// <returns>The response to the request.</returns>
		public delegate PairingRequestAdapter.WebServiceResponse PairingRequestCallback(IPAddress clientIP, PairingRequestAdapter.PairingRequestInfo requestData);

		/// <summary>
		/// Represents the callback function that will handle pairing response.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="responseData">The response data.</param>
		/// <returns>The response to the request.</returns>
		public delegate PairingResponseAdapter.WebServiceResponse PairingResponseCallback(IPAddress clientIP, PairingResponseAdapter.PairingResponseInfo responseData);

		/// <summary>
		/// Represents the callback function that will handle pairing status request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns>The response to the request.</returns>
		public delegate PairingStatusRequestAdapter.WebServiceResponse PairingStatusRequestCallback(IPAddress clientIP, Guid instanceID);

		/// <summary>
		/// Represents the methods that will handle the event of pairing status received.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="targetInstanceID">The instance ID of target device.</param>
		/// <param name="pairingStatusData">The pairing status data.</param>
		/// <returns>The response to the request.</returns>
		public delegate PairingStatusUpdateAdapter.WebServiceResponse PairingStatusUpdateCallback(IPAddress clientIP, Guid targetInstanceID, PairingStatusInfo pairingStatusData);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private PairingInvitationAdapter _invitationAdapter;
		private PairingRequestAdapter _requestAdapter;
		private PairingResponseAdapter _responseAdapter;
		private PairingStatusRequestAdapter _statusRequestAdapter;
		private PairingStatusUpdateAdapter _statusUpdateAdapter;
		private PairingInvitationCallback _invitationCallback;
		private PairingRequestCallback _requestCallback;
		private PairingResponseCallback _responseCallback;
		private PairingStatusRequestCallback _statusRequestCallback;
		private PairingStatusUpdateCallback _statusUpdateCallback;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="PairingService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public PairingService(WebServicesManager webServiceManager)
		{
			this._invitationAdapter = new PairingInvitationAdapter(webServiceManager);
			this._requestAdapter = new PairingRequestAdapter(webServiceManager);
			this._responseAdapter = new PairingResponseAdapter(webServiceManager);
			this._statusRequestAdapter = new PairingStatusRequestAdapter(webServiceManager);
			this._statusUpdateAdapter = new PairingStatusUpdateAdapter(webServiceManager);

			this._invitationAdapter.RegisterRequestCallback(this.OnInvitationAdapterRequestReceived);
			this._requestAdapter.RegisterRequestCallback(this.OnRequestAdapterRequestReceived);
			this._responseAdapter.RegisterRequestCallback(this.OnResponseAdapterRequestReceived);
			this._statusRequestAdapter.RegisterRequestCallback(this.OnStatusRequestAdapterRequestReceived);
			this._statusUpdateAdapter.RegisterRequestCallback(this.OnStatusUpdateAdapterRequestReceived);

			this._invitationCallback = null;
			this._requestCallback = null;
			this._responseCallback = null;
			this._statusRequestCallback = null;
			this._statusUpdateCallback = null;
		}

		/// <summary>
		/// Registers the callback function of pairing invitation.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterPairingInvitationCallback(PairingInvitationCallback callback)
		{
			this._invitationCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of pairing request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterPairingRequestCallback(PairingRequestCallback callback)
		{
			this._requestCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of pairing response.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterPairingResponseCallback(PairingResponseCallback callback)
		{
			this._responseCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of pairing status request.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterPairingStatusRequestCallback(PairingStatusRequestCallback callback)
		{
			this._statusRequestCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of pairing status update.
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterPairingStatusUpdateCallback(PairingStatusUpdateCallback callback)
		{
			this._statusUpdateCallback = callback;
		}

		/// <summary>
		/// Sends a pairing invitation to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="invitationData">The data of the pairing invitation.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingInvitationAdapter.WebServiceResponse> InviteAsync(IPAddress address, ushort port, PairingInvitationAdapter.PairingInvitationInfo invitationData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._invitationAdapter != null);
				return await this._invitationAdapter.InviteAsync(address, port, invitationData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a pairing request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="requestData">The data of the pairing request.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingRequestAdapter.WebServiceResponse> RequestAsync(IPAddress address, ushort port, PairingRequestAdapter.PairingRequestInfo requestData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._requestAdapter != null);
				return await this._requestAdapter.RequestAsync(address, port, requestData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a pairing response to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="responseData">The data of the pairing response.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingResponseAdapter.WebServiceResponse> ResponseAsync(IPAddress address, ushort port, PairingResponseAdapter.PairingResponseInfo responseData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._responseAdapter != null);
				return await this._responseAdapter.ResponseAsync(address, port, responseData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Requests the pairing status from the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="selfInstanceID">The instance ID of self device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingStatusRequestAdapter.WebServiceResponse> RequestStatusAsync(IPAddress address, ushort port, Guid selfInstanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._statusRequestAdapter != null);
				return await this._statusRequestAdapter.RequestStatusAsync(address, port, selfInstanceID).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Updates the pairing status to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="selfInstanceID">The instance ID of self device.</param>
		/// <param name="pairingStatusData">The pairing status data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingStatusUpdateAdapter.WebServiceResponse> UpdateStatusAsync(IPAddress address, ushort port, Guid selfInstanceID, PairingStatusInfo pairingStatusData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._statusUpdateAdapter != null);
				return await this._statusUpdateAdapter.UpdateStatusAsync(address, port, selfInstanceID, pairingStatusData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a pairing invitation is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PairingInvitationAdapter.WebServiceResponse OnInvitationAdapterRequestReceived(IPAddress address, ushort port, PairingInvitationAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._invitationCallback == null)
				{
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// check the instance ID
				if (!remoteRequest.Request.InstanceID.HasValue || remoteRequest.Request.InstanceID.Value.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogWarn("Invalid instance ID.");
					return PairingInvitationAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified or invalid.");
				}

				// invoke callback function
				return this._invitationCallback(address, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PairingRequestAdapter.WebServiceResponse OnRequestAdapterRequestReceived(IPAddress address, ushort port, PairingRequestAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._requestCallback == null)
				{
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// check the instance ID
				if (!remoteRequest.Request.InstanceID.HasValue || remoteRequest.Request.InstanceID.Value.Equals(Guid.Empty))
				{
					StaticLogger.LogWarn("Invalid instance ID.");
					return PairingRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified or invalid.");
				}

				// invoke callback function
				return this._requestCallback(address, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing response is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PairingResponseAdapter.WebServiceResponse OnResponseAdapterRequestReceived(IPAddress address, ushort port, PairingResponseAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._responseCallback == null)
				{
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// check the instance ID
				if (!remoteRequest.Request.InstanceID.HasValue || remoteRequest.Request.InstanceID.Value.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogWarn("Invalid instance ID.");
					return PairingResponseAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified or invalid.");
				}

				// invoke callback function
				return this._responseCallback(address, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing status request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PairingStatusRequestAdapter.WebServiceResponse OnStatusRequestAdapterRequestReceived(IPAddress address, ushort port, PairingStatusRequestAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._statusRequestCallback == null)
				{
					return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}

				// get instance ID
				Guid instanceID;
				{
					string webServiceUriTrimSlashes = PairingStatusRequestAdapter.WebServiceUri.TrimStart('/');
					string requestUriTrimSlashes = remoteRequest.RequestUri.ToString().TrimStart('/');

					if (!requestUriTrimSlashes.StartsWith(webServiceUriTrimSlashes))
					{
						Trace.Fail("Shouldn't be here, there should be something wrong.");
						return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Unknown error.");
					}

					string postfix = requestUriTrimSlashes.Substring(webServiceUriTrimSlashes.Length).Trim('/');
					var postfixTokens = postfix.Split('/');

					if (postfixTokens == null || string.IsNullOrEmpty(postfixTokens[0]))
					{
						StaticLogger.LogWarn("No instance ID is specified.");
						return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified.");
					}

					if (!Guid.TryParse(postfixTokens[0], out instanceID))
					{
						StaticLogger.LogWarn("Bad instance ID: " + postfixTokens[0]);
						return PairingStatusRequestAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is invalid.");
					}
				}

				// invoke callback function
				return this._statusRequestCallback(address, instanceID);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a pairing status update is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PairingStatusUpdateAdapter.WebServiceResponse OnStatusUpdateAdapterRequestReceived(IPAddress address, ushort port, PairingStatusUpdateAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._statusUpdateCallback == null)
				{
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered callback function.");
				}
				if (remoteRequest == null)
				{
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// get instance ID
				Guid instanceID;
				{
					string webServiceUriTrimSlashes = PairingStatusUpdateAdapter.WebServiceUri.TrimStart('/');
					string requestUriTrimSlashes = remoteRequest.RequestUri.ToString().TrimStart('/');

					if (!requestUriTrimSlashes.StartsWith(webServiceUriTrimSlashes))
					{
						Trace.Fail("Shouldn't be here, there should be something wrong.");
						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "Unknown error.");
					}

					string postfix = requestUriTrimSlashes.Substring(webServiceUriTrimSlashes.Length).Trim('/');
					var postfixTokens = postfix.Split('/');

					if (postfixTokens == null || string.IsNullOrEmpty(postfixTokens[0]))
					{
						StaticLogger.LogWarn("No instance ID is specified.");
						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified.");
					}

					if (!Guid.TryParse(postfixTokens[0], out instanceID))
					{
						StaticLogger.LogWarn("Bad instance ID: " + postfixTokens[0]);
						return PairingStatusUpdateAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is invalid.");
					}
				}

				// invoke callback function
				return this._statusUpdateCallback(address, instanceID, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
