﻿using AirFusion.Common;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of pairing status (request) (GET /pairing/status/{InstanceID}).
	/// </summary>
	internal sealed class PairingStatusRequestAdapter : WebServiceAdapter<DBNull, PairingStatusInfo>
	{
		#region Definitions

		internal const string WebServiceUri = "/pairing/status";
		internal static HttpMethod WebServiceMethod = HttpMethod.Get;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="PairingStatusRequestAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public PairingStatusRequestAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Requests the pairing status from the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="selfInstanceID">The instance ID of self device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestStatusAsync(IPAddress address, ushort port, Guid selfInstanceID)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (selfInstanceID.Equals(Guid.Empty))
			{
				throw new ArgumentException("Please specify a valid instance ID.");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + "/" + selfInstanceID.GetStringFormat()).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
