﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.IncomingCall
{
	/// <summary>
	/// Represents the manager of incoming call
	/// </summary>
	public sealed partial class IncomingCallManager : IIncomingCallManager
	{
		#region Protected/Private/Internal Variables

		private bool _enableFeatureFlag = false;

		internal WebServicesManager _webServiceManager;
		internal IncomingCallService _incomingCallService;
		internal UserDeviceManager _userDeviceManager;

		#endregion Protected/Private/Internal Variables

		#region IIncomingCallManager Events

		/// <summary>
		/// Occurs when PC receives a incoming call notification from phone
		/// </summary>
		public event ICCallReceivedEventHandler CallReceived;

		/// <summary>
		/// Occurs when the state of phone call is changed
		/// </summary>
		public event ICStateChangedEventHandler StateChanged;

		#endregion IIncomingCallManager Events

		#region Initialize Functions

		/// <summary>
		/// Initialize a new instance of the <see cref="IncomingCallManager"/> class.
		/// </summary>
		public IncomingCallManager()
		{
			this._webServiceManager = null;
			this._userDeviceManager = null;
		}

		/// <summary>
		/// Sets up the user device manager.
		/// </summary>
		/// <param name="userDeviceManager">The user device manager.</param>
		public void SetUserDeviceManager(UserDeviceManager userDeviceManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (userDeviceManager == null)
				{
					throw new ArgumentNullException("userDeviceManager");
				}

				this._userDeviceManager = userDeviceManager;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sets up the web service manager.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public void SetWebService(WebServicesManager webServiceManager)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (webServiceManager == null)
				{
					throw new ArgumentNullException("webServiceManager");
				}

				this._webServiceManager = webServiceManager;

				// incomingcall service
				{
					this._incomingCallService = new IncomingCallService(webServiceManager);
					this._incomingCallService.RegisterIncomingCallPostRequestCallback(OnICPostRequestReceived);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Initialize Functions

		#region IIncomingCallManager Methods

		/// <summary>
		/// Enables/Disables the feature of Answer phone call on PC
		/// </summary>
		/// <param name="enable">
		/// - True : Enable the feature of Answer phone call on PC
		/// - False: Disable the feature of Answer phone call on PC
		/// </param>
		public void Enable(bool enable)
		{
			if (enable)
			{
				this._enableFeatureFlag = true;
			}
			else
			{
				this._enableFeatureFlag = false;
			}
		}

		/// <summary>
		/// Answers the incoming call on PC.
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		public void Answer(AFUserDevice device, AnswerCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// [1] Check if the bluetooth is enabled
				if (PCBluetoothHandler.Instance.IsBluetoothEnabled())
				{
					// [2] Check if it has the mac address of phone BT
					if (device.Bluetooth.Address == null)
					{
						if (callback != null)
							callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.No_Phone_BT_Address));
					}
					else
					{
						// [3] Check if the phone bluetooth is paired on PC
						string phoneBluetoothAddress = device.Bluetooth.Address.ToString("C");
						if (PCBluetoothHandler.Instance.IsPairedBluetoothDevice(phoneBluetoothAddress))
						{
							#region Start to send "Answer" request to Phone

							IncomingCallReplyAdapter.IncomingCallReplyData replyData = new IncomingCallReplyAdapter.IncomingCallReplyData();
							//replyData.DeviceInstanceID = device.InstanceID;
							replyData.BluetoothMacAddr = this._userDeviceManager.LocalDevice.Bluetooth.Address.ToString("C"); ;

							Task<IncomingCallReplyAdapter.Response> task = this._incomingCallService.AnswerPhoneCallAsync(device.Address, device.WebServicePort, replyData);
							task.ContinueWith(delegate(Task t)
							{
								if (t.IsCompleted)
								{
									if (t.IsFaulted)
									{
										if (t.Exception != null && t.Exception.InnerException != null)
										{
											StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
										}
										else
										{
											StaticLogger.LogError("Error statusCode.");
										}

										if (callback != null)
											callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.WebService_Failed));
									}
									else
									{
										IncomingCallReplyAdapter.IncomingCallActionResponse response = task.Result.ResponseData;
										StaticLogger.LogDebug("Response's error code = " + response.PhoneCallError.ToString());
										if (callback != null)
										{
											switch (response.PhoneCallError)
											{
												case IncomingCallReplyAdapter.IncomingCallActionResponse.PhoneError.No_Error:
													callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Success));
													break;

												case IncomingCallReplyAdapter.IncomingCallActionResponse.PhoneError.BT_TurnOff:
													callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.Phone_BT_Disabled));
													break;

												case IncomingCallReplyAdapter.IncomingCallActionResponse.PhoneError.BT_Unpaired:
													callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.Phone_BT_NotPaired));
													break;
											}
										}
									}
								}
							});

							#endregion Start to send "Answer" request to Phone
						}
						else
						{
							if (callback != null)
								callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.PC_BT_NotPaired));
						}
					}
				}
				else
				{
					if (callback != null)
						callback(new AnswerStatus(device, ICBaseStatus.StatusType.Completed_Error, AnswerStatus.Error.PC_BT_Disabled));
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Rejects the incoming call or hangs up the phone call.
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		public void Decline(AFUserDevice device, DeclineCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				IncomingCallReplyAdapter.IncomingCallReplyData replyData = new IncomingCallReplyAdapter.IncomingCallReplyData();
				replyData.BluetoothMacAddr = this._userDeviceManager.LocalDevice.Bluetooth.Address.ToString("C");

				Task<IncomingCallReplyAdapter.Response> task = this._incomingCallService.DeclinePhoneCallAsync(device.Address, device.WebServicePort, replyData);
				task.ContinueWith(delegate(Task t)
				{
					if (t.IsCompleted)
					{
						if (t.IsFaulted)
						{
							if (t.Exception != null && t.Exception.InnerException != null)
							{
								StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
							}
							else
							{
								StaticLogger.LogError("Error statusCode.");
							}

							if (callback != null)
								callback(new DeclineStatus(device, ICBaseStatus.StatusType.Completed_Error));
						}
						else
						{
							if (callback != null)
								callback(new DeclineStatus(device, ICBaseStatus.StatusType.Completed_Success));
						}
					}
				});
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Hangs up the phone call
		/// </summary>
		/// <param name="device">The phone device with incoming call</param>
		/// <param name="callback">Infrom progress/error message by running callback function</param>
		public void HangUp(AFUserDevice device, DeclineCallback callback)
		{
			// *Note: This function will be modified when QCA provides Bluetooth APIs
			try
			{
				StaticLogger.TrackEnter();

				IncomingCallReplyAdapter.IncomingCallReplyData replyData = new IncomingCallReplyAdapter.IncomingCallReplyData();
				replyData.BluetoothMacAddr = this._userDeviceManager.LocalDevice.Bluetooth.Address.ToString("C");

				Task<IncomingCallReplyAdapter.Response> task = this._incomingCallService.DeclinePhoneCallAsync(device.Address, device.WebServicePort, replyData);
				task.ContinueWith(delegate(Task t)
				{
					if (t.IsCompleted)
					{
						if (t.IsFaulted)
						{
							if (t.Exception != null && t.Exception.InnerException != null)
							{
								StaticLogger.LogError("Error statusCode. \r\n" + CommonUtils.GetExceptionMessages(t.Exception.InnerException));
							}
							else
							{
								StaticLogger.LogError("Error statusCode.");
							}

							if (callback != null)
								callback(new DeclineStatus(device, ICBaseStatus.StatusType.Completed_Error));
						}
						else
						{
							if (callback != null)
								callback(new DeclineStatus(device, ICBaseStatus.StatusType.Completed_Success));
						}
					}
				});
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#region [TBD] The Functions will be implemented in the future

		///// <summary>
		///// Rejects the incoming call and asks phone to send one short message to caller
		///// </summary>
		///// <param name="device">The phone device with incoming call</param>
		///// <param name="callback">Infrom progress/error message by running callback function</param>
		///// <param name="message">Short message</param>
		//public void DeclineWithSms(AFUserDevice device, DeclineCallback callback, string message);

		///// <summary>
		///// Turns off the microphone of PC
		///// </summary>
		///// <param name="mute">
		///// - True : Turn off
		///// - False: Turn on
		///// </param>
		//public void MuteMicrophone(bool mute);

		///// <summary>
		///// Switch the communication device
		///// </summary>
		///// <param name="device">
		///// - PC: Connect hands-free connection and it will return to call on PC
		///// - Phone: Disconnect hands-free connection and the voice will output from phone
		///// </param>
		///// <param name="target">If device == PC, the target is the smartphone</param>
		//public void SwitchCommunDevice(AnswerPhoneDevice device, AFUserDevice target = null);

		#endregion [TBD] The Functions will be implemented in the future

		#endregion IIncomingCallManager Methods

		#region Callbacks

		/// <summary>
		/// Called when a incoming call post request is received.
		/// </summary>
		/// <param name="sourceIP">The remote device IP address</param>
		/// <param name="postData">The post request data</param>
		/// <returns>The response to the request</returns>
		private IncomingCallPostAdapter.WebServiceResponse OnICPostRequestReceived(IPAddress sourceIP, IncomingCallPostAdapter.IncomingCallPostData postData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (postData == null)
				{
					StaticLogger.LogError("The request data is unspecified.");
					return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotAcceptable, "The request data is unspecified.");
				}

				if (this._enableFeatureFlag)
				{
					if (postData.State == IncomingCallPostAdapter.IncomingCallPostData.NOTIFICATION_VALUE)
					{
						// Incoming Call Notification
						TriggerCallReceivedEvent(postData);
					}
					else
					{
						TriggerStateChangedEvent(postData);
					}
				}

				return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks

		#region Trigger Events

		/// <summary>
		/// Triggers the CallReceived event
		/// </summary>
		/// <param name="postData"></param>
		private void TriggerCallReceivedEvent(IncomingCallPostAdapter.IncomingCallPostData postData)
		{
			try
			{
				StaticLogger.LogDebug("[TriggerCallReceivedEvent] postData => " + postData.State);

				if (this.CallReceived != null)
				{
					AFUserDevice userDevice = this._userDeviceManager.DeviceStore.GetDevice(postData.DeviceInstanceID.Value);
					CloseQCAPhoneWindows(userDevice, 3000, true);

					string callerPhone = postData.PhoneNumber;
					string callerName = postData.ContactName;
					Bitmap callerPhoto = null;

					if (postData.CallerPhoto == null || postData.CallerPhoto == string.Empty)
					{
						// Set default caller's photo
					}
					else
					{
						// Set the caller's photo that was sent from phone
						callerPhoto = ConvertBitmapImage(postData.CallerPhoto);
					}

					CommonUtils.BeginInvokeEvent<ICCallReceivedEventArgs>(
						new EventHandler<ICCallReceivedEventArgs>(this.CallReceived),
						this,
						new ICCallReceivedEventArgs(userDevice, callerPhone, callerName, callerPhoto)
						);
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("[TriggerCallReceivedEvent] Exception => " + ex.ToString());
			}
		}

		/// <summary>
		/// Triggers the StateChanged event
		/// </summary>
		/// <param name="postData"></param>
		private void TriggerStateChangedEvent(IncomingCallPostAdapter.IncomingCallPostData postData)
		{
			try
			{
				StaticLogger.LogDebug("[TriggerStateChangedEvent] postData => " + postData.State);
				if (this.StateChanged != null)
				{
					AFUserDevice userDevice = this._userDeviceManager.DeviceStore.GetDevice(postData.DeviceInstanceID.Value);
					ICStateChangeEventArgs.LineState curState = ICStateChangeEventArgs.LineState.Unknown;
					uint curTalkTime = 0;

					switch (postData.State)
					{
						case IncomingCallPostAdapter.IncomingCallPostData.ON_LINE_VALUE:
							curState = ICStateChangeEventArgs.LineState.AnswerOnPC;
							CloseQCAPhoneWindows(userDevice, 10000, false);
							break;

						case IncomingCallPostAdapter.IncomingCallPostData.OFF_LINE_VALUE:
							curState = ICStateChangeEventArgs.LineState.HangUpFromPC;
							//curTalkTime = postData.time
							break;

						default:
							break;
					}

					if (curState != ICStateChangeEventArgs.LineState.Unknown)
					{
						CommonUtils.BeginInvokeEvent<ICStateChangeEventArgs>(
							new EventHandler<ICStateChangeEventArgs>(this.StateChanged),
							this,
							new ICStateChangeEventArgs(userDevice, curState, curTalkTime)
							);
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("[TriggerStateChangedEvent] Exception => " + ex.ToString());
			}
		}

		#endregion Trigger Events

		#region Private Functions

		/// <summary>
		/// Converts the string data to a bitmap photo
		/// </summary>
		/// <param name="photoData">The string data of photo</param>
		/// <returns></returns>
		private Bitmap ConvertBitmapImage(string photoData)
		{
			Bitmap retBitmap = null;

			if (photoData != null && photoData != string.Empty)
			{
				byte[] byteBuffer = Convert.FromBase64String(photoData);
				MemoryStream memoryStream = new MemoryStream(byteBuffer);
				memoryStream.Position = 0;

				retBitmap = (Bitmap)Bitmap.FromStream(memoryStream);

				memoryStream.Close();
				memoryStream = null;
				byteBuffer = null;
			}

			return retBitmap;
		}

		#endregion Private Functions

		#region Workaround Methods

		/// <summary>
		/// Check the QCA's window and foce to close it
		/// </summary>
		/// <param name="device">The phone device</param>
		/// <param name="max_time_ms">The max time (milliseconds) that to find the windows of QCA</param>
		/// <param name="hide_only">
		///  - True: Only to hide windows, not to close
		///  - False: To close the windows
		/// </param>
		private void CloseQCAPhoneWindows(AFUserDevice device, int max_time_ms, bool hide_only)
		{
			try
			{
				string phone_name = device.DeviceName;
				string phone_bt_mac = device.Bluetooth.Address.ToString("C");

				Thread t = new Thread(() => PCBluetoothHandler.Instance.CloseQcaBluetoothWindows(phone_name, phone_bt_mac, max_time_ms, hide_only));
				t.Start();
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("[IncomingCallManager] CloseQCAPhoneWindows(): Exception => " + ex.ToString());
			}
		}

		#endregion Workaround Methods

		#region Test Function

		////private AFUserDevice _testDevice;
		////private int _testCaseNum = 0;
		////private void TestAnswerPhone(AFUserDevice device)
		////{
		////	this._testDevice = device;
		////	this._testCaseNum = 0;
		////	Thread t = new Thread(ThreadTestFun);
		////	t.Start();
		////}

		////private void TestDeclinePhone(AFUserDevice device)
		////{
		////	this._testDevice = device;
		////	this._testCaseNum = 1;
		////	Thread t = new Thread(ThreadTestFun);
		////	t.Start();
		////}

		////private void TestHangUpPhone(AFUserDevice device)
		////{
		////	this._testDevice = device;
		////	this._testCaseNum = 2;
		////	Thread t = new Thread(ThreadTestFun);
		////	t.Start();
		////}

		////private void ThreadTestFun()
		////{
		////	switch (this._testCaseNum)
		////	{
		////		case 0:
		////			this.Answer(this._testDevice, null);
		////			break;
		////		case 1:
		////			this.Decline(this._testDevice, null);
		////			break;
		////		case 2:
		////			this.HangUp(this._testDevice, null);
		////			break;
		////		default:
		////			break;
		////	}
		////}

		#endregion Test Function
	}
}
