﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of incoming call post from Android side (Phone => PC)
	/// </summary>
	internal sealed class IncomingCallPostAdapter : WebServiceAdapter<IncomingCallPostAdapter.IncomingCallPostData, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/incomingcall";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Class

		/// <summary>
		/// Represents the request data of incoming call post.
		/// </summary>
		public sealed class IncomingCallPostData
		{
			/// <summary>
			/// Incoming Call value that Android defines
			/// </summary>
			public const int NOTIFICATION_VALUE = 1;

			/// <summary>
			/// Online state value that Android defines
			/// </summary>
			public const int ON_LINE_VALUE = 2;

			/// <summary>
			/// Offline state value that Android defines
			/// </summary>
			public const int OFF_LINE_VALUE = 0;

			#region Properties

			/// <summary>
			/// Gets the device id
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(GuidConverter))]
			public Guid? DeviceInstanceID { get; internal set; }

			/// <summary>
			/// Gets the phone number of incoming call
			/// </summary>
			[JsonProperty("phone_number")]
			public string PhoneNumber { get; internal set; }

			/// <summary>
			/// Gets the contact name of incoming call
			/// </summary>
			[JsonProperty("contact_name")]
			public string ContactName { get; internal set; }

			/// <summary>
			/// Gets the current state of incoming call
			/// </summary>
			[JsonProperty("state")]
			public int State { get; internal set; }

			/// <summary>
			/// Gets the poto of caller
			/// </summary>
			[JsonProperty("contact_photo")]
			public string CallerPhoto { get; internal set; }

			#endregion Properties

			/// <summary>
			/// Initializes a new instance of the <see cref="IncomingCallPostData"/> class.
			/// </summary>
			public IncomingCallPostData()
			{
				this.DeviceInstanceID = null;
				this.PhoneNumber = null;
				this.ContactName = null;
				this.State = -1;
				this.CallerPhoto = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".DeviceInstanceID: " + (this.DeviceInstanceID.HasValue ? this.DeviceInstanceID.Value.GetStringFormat() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".PhoneNumber: " + (string.IsNullOrEmpty(this.PhoneNumber) ? this.PhoneNumber : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ContactName: " + (string.IsNullOrEmpty(this.ContactName) ? this.ContactName : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".State: " + this.State.ToString());

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Class

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="IncomingCallPostAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public IncomingCallPostAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		#endregion Methods
	}
}
