﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of incoming call reply (POST /incomingcall/answer or /incomingcall/hangup)
	/// </summary>
	internal sealed class IncomingCallReplyAdapter : WebServiceAdapter<IncomingCallReplyAdapter.IncomingCallReplyData, IncomingCallReplyAdapter.IncomingCallActionResponse>
	{
		#region Definitions

		internal const string WebServiceAnswerUri = "/incomingcall/answer";
		internal const string WebServiceHangupUri = "/incomingcall/hangup";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Class

		/// <summary>
		/// Represents the request data of answering/declining the incoming call.
		/// </summary>
		public sealed class IncomingCallReplyData
		{
			#region Properties

			/*
			/// <summary>
			/// Gets the instance ID of the device.
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(GuidConverter))]
			public Guid? DeviceInstanceID { get; set; }
			*/

			/// <summary>
			/// Gets the Bluetooth MAC address of local PC.
			/// </summary>
			[JsonProperty("bt_mac_addr")]
			public string BluetoothMacAddr { get; set; }

			#endregion Properties

			/// <summary>
			/// Initializes a new instance of the <see cref="IncomingCallReplyData"/> class.
			/// </summary>
			public IncomingCallReplyData()
			{
				//this.DeviceInstanceID = null;
				this.BluetoothMacAddr = string.Empty;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				//builder.AppendLine(this.GetType().GetFullName() + ".DeviceInstanceID: " + (this.DeviceInstanceID.HasValue ? this.DeviceInstanceID.Value.GetStringFormat() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".BluetoothMacAddr: " + (this.BluetoothMacAddr != string.Empty ? this.BluetoothMacAddr : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		/// <summary>
		/// Represents the customize response data for the request of answering/declining the incoming call.
		/// </summary>
		public sealed class IncomingCallActionResponse
		{
			/// <summary>
			/// Specifies the error codes from phone for answer/decline/hang up request
			/// </summary>
			public enum PhoneError
			{
				/// <summary>
				/// Success (No any error)
				/// </summary>
				No_Error = 1000,

				/// <summary>
				/// Phone's BT is OFF
				/// </summary>
				BT_TurnOff = 1001,

				/// <summary>
				/// The paired BT device of Phone has no PC's BT
				/// </summary>
				BT_Unpaired = 1002
			}

			#region Properties

			/// <summary>
			/// Gets the error code from phone (For answer/hang up request)
			/// </summary>
			[JsonProperty("phone_call_err")]
			public PhoneError PhoneCallError { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="IncomingCallActionResponse"/> class.
			/// </summary>
			public IncomingCallActionResponse()
			{
				this.PhoneCallError = PhoneError.No_Error;
			}

			#endregion Properties
		}

		#endregion Class

		/// <summary>
		/// Initializes a new instance of the <see cref="IncomingCallReplyAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public IncomingCallReplyAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Sends a answer request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="replyData">The required replay data to ask phone to answer the incoming call.</param>
		public async Task<WebServiceResponse> RequestAnswerPhoneAsync(IPAddress address, ushort port, IncomingCallReplyData replyData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (replyData == null)
			{
				throw new ArgumentNullException("replyData");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceAnswerUri).Uri,
				replyData);

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		/// <summary>
		/// Sends a hang up request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="replyData">The required replay data to ask phone to hang up the incoming/phone call.</param>
		public async Task<WebServiceResponse> RequestHangUpPhoneAsync(IPAddress address, ushort port, IncomingCallReplyData replyData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (replyData == null)
			{
				throw new ArgumentNullException("replyData");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceHangupUri).Uri,
				replyData);

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}
	}
}
