﻿using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace AirFusion.Manager.Notification
{
	/// <summary>
	/// Represents the contants definitions for the notification
	/// </summary>
	public static class Definitions
	{
		internal static string ACEREXTEND_PHONE_PACKAGENAME = "com.acer.android.airfusion";
		internal static string GMAIL_PHONE_PACKAGENAME = "com.google.android.gm";

		internal static int NOTIFICATION_ID_SHARING_LINK = 0x0100;
		internal static int NOTIFICATION_ID_OPEN_REMOTE_DOWNLOADS = 0x1000;
		internal static int NOTIFICATION_ID_OPEN_REMOTE_UPLOADED_FILE = 0x10000;
	}

	/// <summary>
	/// Represents a notification help management
	/// </summary>
	public static class NotificationHelper
	{
		/// <summary>
		/// Gets the image cache folder
		/// </summary>
		/// <returns></returns>
		public static string GetIconCacheFolder()
		{
			string path = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData) + @"\OEM\AcerEXTEND\NotificationImageStore";
			if (!System.IO.Directory.Exists(path))
			{
				System.IO.Directory.CreateDirectory(path);
			}
			return path;
		}
	}

	/// <summary>
	/// Represents the notification ignore/permit status of the device
	/// </summary>
	public enum DeviceIgnoreStatus
	{
		/// <summary>
		/// The notification of the device is ignored
		/// </summary>
		Ignore = 0,

		/// <summary>
		/// The notification of the device is permitted to show
		/// </summary>
		Permit = 1
	}

	/// <summary>
	/// Represents a notification key management.
	/// </summary>
	internal class NotificationKeyMaker
	{
		/// <summary>
		/// Gets the notification key by notification data
		/// </summary>
		/// <param name="postData">The received notification data.</param>
		public static string GetNotificationKeyByPostData(NotificationPostAdapter.NotificationPostInfo postData)
		{
			Debug.Assert(postData != null);
			Debug.Assert(postData.DeviceInstanceID != null);

			return GenerateHashCode(postData.DeviceInstanceID, postData.PackageName, postData.Id, postData.UserId, postData.Tag);
		}

		/// <summary>
		/// Gets the notification key by removal request uri
		/// </summary>
		/// <param name="removeUri">The notification removal uri</param>
		public static string GetNotificationKeyByRemovalUri(string removeUri)
		{
			Debug.Assert(removeUri != null);

			//expect: /notification/instance_id/user_id/package_name/id/tag
			string delimStr = "/";
			char[] delimiter = delimStr.ToCharArray();
			string[] splits = removeUri.Split(delimiter, 7, StringSplitOptions.None);

			string instanceId = null;
			int userId = 0;
			string packageName = null;
			int id = 0;
			string tag = null;

			if (splits.Length >= 3)
				instanceId = splits[2].ToLowerInvariant();

			if (splits.Length >= 4)
				userId = Convert.ToInt32(splits[3]);

			if (splits.Length >= 5)
				packageName = splits[4];

			if (splits.Length >= 6)
				id = Convert.ToInt32(splits[5]);

			if (splits.Length >= 7)
				tag = (splits[6] == string.Empty) ? null : splits[6];

			return GenerateHashCode(instanceId, packageName, id, userId, tag);
		}

		/// <summary>
		/// Generates the hash code
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="objs"></param>
		/// <returns></returns>
		private static string GenerateHashCode(object obj, params object[] objs)
		{
			Debug.Assert(obj != null);

			string divider = "-";
			string identity = obj.ToString();
			foreach (object value in objs)
			{
				if (value != null)
				{
					identity += divider + value.ToString();
				}
			}

			return identity.GetHashCode().ToString();
		}
	}

	/// <summary>
	/// Represents the local launch constructor
	/// </summary>
	internal sealed class LocalLaunchFactory
	{
		/// <summary>
		/// Gets the local launch handler.
		/// </summary>
		/// <param name="toastType">The notification toast type.</param>
		/// <param name="postData">The post data.</param>
		public static LocalLaunch GetLocalLaunchHandler(NotificationType toastType, NotificationPostAdapter.NotificationPostInfo postData)
		{
			LocalLaunch launchItem = null;

			if (toastType == NotificationType.UrlShare || toastType == NotificationType.UrlShareWithLaunch)
			{
				// make sure the url is not null
				if (string.IsNullOrEmpty(postData.ContentView.BottomText))
				{
					throw new ArgumentNullException("postData.ContentView.BottomText");
				}
				launchItem = new UrlShareLocalLaunch(postData.ContentView.BottomText);
			}
			else if (toastType == NotificationType.FileShare || toastType == NotificationType.FileShareWithLaunch)
			{
				// make sure the file name is not null
				if (string.IsNullOrEmpty(postData.ContentView.BottomText))
				{
					throw new ArgumentNullException("postData.ContentView.BottomText");
				}
				launchItem = new FileShareLocalLaunch(postData.ContentView.BottomText);
			}
			else
			{
				// gmail
				if (postData.PackageName.Equals(Definitions.GMAIL_PHONE_PACKAGENAME, StringComparison.InvariantCultureIgnoreCase))
				{
					launchItem = new GmailLocalLaunch(postData.BigContentView.BottomText.TrimEnd());
				}
			}

			return launchItem;
		}
	}

	/// <summary>
	/// Represents the local launch manager.
	/// </summary>
	public abstract class LocalLaunch
	{
		/// <summary>
		/// Does the action.
		/// </summary>
		/// <param name="id">The identifier.</param>
		public abstract void DoAction(int id = 0);
	}

	/// <summary>
	/// Represents the Gmail local launch manager.
	/// </summary>
	public class GmailLocalLaunch : LocalLaunch
	{
		private string _gmailAccount;

		/// <summary>
		/// Initialize a new instance of the <see cref="GmailLocalLaunch"/> class.
		/// </summary>
		/// <param name="gmailAccount">the gmail account</param>
		public GmailLocalLaunch(string gmailAccount)
			: base()
		{
			this._gmailAccount = gmailAccount;
		}

		/// <summary>
		/// Launchs the local Gmail
		/// </summary>
		public override void DoAction(int id = 0)
		{
			System.Diagnostics.Process.Start("https://mail.google.com/mail/u/?authuser=" + this._gmailAccount);
		}
	}

	/// <summary>
	/// Represents the local launch manager for url (open the browser).
	/// </summary>
	public class UrlShareLocalLaunch : LocalLaunch
	{
		private string _url;

		/// <summary>
		/// Initialize a new instance of the <see cref="UrlShareLocalLaunch"/> class.
		/// </summary>
		/// <param name="url">The url.</param>
		public UrlShareLocalLaunch(string url)
		{
			this._url = url;
		}

		/// <summary>
		/// Opens the url on browser.
		/// </summary>
		/// <param name="id"></param>
		public override void DoAction(int id = 0)
		{
			System.Diagnostics.Process.Start(this._url);
		}
	}

	/// <summary>
	/// Represents the local launch manager for files (open the folder or file).
	/// </summary>
	public class FileShareLocalLaunch : LocalLaunch
	{
		private string _files;

		[DllImport("shell32.dll", CharSet = CharSet.Unicode)]
		private static extern int SHGetKnownFolderPath([MarshalAs(UnmanagedType.LPStruct)] Guid rfid, uint dwFlags, IntPtr hToken, out string pszPath);

		private static readonly Guid DownloadsGuid = new Guid("374DE290-123F-4565-9164-39C4925E467B");

		/// <summary>
		/// Initialize a new instance of the <see cref="FileShareLocalLaunch"/> class.
		/// </summary>
		/// <param name="fileName">The file name</param>
		public FileShareLocalLaunch(string fileName)
		{
			this._files = fileName;
		}

		/// <summary>
		/// Opens the folder or the file.
		/// </summary>
		/// <param name="id">
		/// The id to represent which action to do. 0 => Open the folder; 1 => Open the file directly
		/// </param>
		public override void DoAction(int id = 0)
		{
			try
			{
				string downloadsFolder = string.Empty;
				string filePath = string.Empty;
				int retCode = SHGetKnownFolderPath(DownloadsGuid, 0, IntPtr.Zero, out downloadsFolder);
				if (retCode == 0 && !string.IsNullOrEmpty(downloadsFolder))
				{
					downloadsFolder = System.IO.Path.Combine(downloadsFolder, "AcerEXTEND");
					if (!System.IO.Directory.Exists(downloadsFolder))
					{
						System.IO.Directory.CreateDirectory(downloadsFolder);
					}

					filePath = System.IO.Path.Combine(downloadsFolder, this._files);
				}

				if (id == 0)
				{
					// open folder
					System.Diagnostics.Process.Start(downloadsFolder);
				}
				else if (id == 1)
				{
					// open file directly
					if (System.IO.File.Exists(filePath))
					{
						System.Diagnostics.Process.Start(filePath);
					}
				}
			}
			catch (Exception ex)
			{
				Trace.Fail("DoAction() failed:\r\n" + ex.Message);
			}
		}
	}
}
