﻿using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;

namespace AirFusion.Manager.Notification
{
	/// <summary>
	/// Represents the style of notification.
	/// </summary>
	internal abstract class NotificationStyle
	{
		/// <summary>
		/// Gets the post data.
		/// </summary>
		public NotificationPostAdapter.NotificationPostInfo PostData { get; internal set; }

		/// <summary>
		/// Gets or sets the icon path.
		/// </summary>
		public string IconPath { get; internal set; }

		/// <summary>
		/// Gets the view.
		/// </summary>
		public NotificationPostAdapter.NotificationContentView View { get; internal set; }

		/// <summary>
		/// Gets the name of the application.
		/// </summary>
		public string AppName
		{
			get
			{
				return this.PostData != null ? (!string.IsNullOrEmpty(this.PostData.AppName) ? this.PostData.AppName : this.PostData.PackageName) : string.Empty;
			}
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationStyle"/> class.
		/// </summary>
		public NotificationStyle()
		{
			this.PostData = null;
			this.IconPath = null;
		}

		/// <summary>
		/// Gets the toast element used by UI.
		/// </summary>
		/// <returns>A instance of the <see cref="ToastItem"/> class</returns>
		public abstract ToastItem GetContent();
	}

	/// <summary>
	/// Represents the type of the notification content view.
	/// </summary>
	public enum ContentViewType
	{
		/// <summary>
		/// The basic content view
		/// </summary>
		Basic = 0,

		/// <summary>
		/// The big content view
		/// </summary>
		Big = 1
	}

	/// <summary>
	/// Represents a toast factory.
	/// </summary>
	internal static class ToastFactory
	{
		/// <summary>
		/// Creates the toast element used by UI
		/// </summary>
		/// <param name="postData">The received notification data from phone</param>
		/// <param name="iconPath">The path of the notification icon</param>
		/// <returns>A instance of the <see cref="ToastItem"/> class</returns>
		public static ToastItem CreateToast(NotificationPostAdapter.NotificationPostInfo postData, string iconPath)
		{
			ToastItem item = null;
			if (postData.BigContentView != null &&
				postData.BigContentView.Style != null &&
				postData.BigContentView.Style.Value >= 0 &&
				postData.BigContentView.Style.Value <= 8)
			{
				NotificationStyle content = GetContentParser(postData.BigContentView.Style.Value);
				if (content != null)
				{
					content.PostData = postData;
					content.IconPath = iconPath;
					content.View = postData.BigContentView;
					item = content.GetContent();
				}
			}

			if (item == null || item.ToastContent == null || item.ToastContent.Length == 0)
			{
				if (postData.ContentView != null &&
					postData.ContentView.Style != null &&
					postData.ContentView.Style.Value >= 0 &&
					postData.ContentView.Style.Value <= 5)
				{
					NotificationStyle content = GetContentParser(postData.ContentView.Style.Value);
					if (content != null)
					{
						content.PostData = postData;
						content.IconPath = iconPath;
						content.View = postData.ContentView;
						item = content.GetContent();
					}
				}
			}

			return item;
		}

		/// <summary>
		/// Gets the parser of the notification by notificaiton style
		/// </summary>
		/// <param name="style">The content style of the notification</param>
		/// <returns>A instance of the <see cref="NotificationStyle"/> class</returns>
		private static NotificationStyle GetContentParser(int style)
		{
			NotificationStyle content = null;
			switch (style)
			{
				case 0:
				case 1:
				case 2:
				case 3:
				// Style 6 7 8 are media style
				case 6:
				case 7:
				case 8:
					content = new NotificationContentBasicStyle();
					break;

				case 4:
					content = new NotificationContentInboxStyle();
					break;

				case 5:
					content = new NotificationContentRemoteViewStyle();
					break;

				default:
					content = new NotificationContentBasicStyle();
					break;
			}

			return content;
		}
	}

	/// <summary>
	/// Represents the inbox style of notification.
	/// </summary>
	internal sealed class NotificationContentInboxStyle : NotificationStyle
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationContentInboxStyle"/> class.
		/// </summary>
		public NotificationContentInboxStyle()
			: base()
		{
		}

		/// <summary>
		/// Gets the toast element from the inboxed-style notification.
		/// </summary>
		/// <returns>An instance of the <see cref="ToastItem"/> class.</returns>
		public override ToastItem GetContent()
		{
			ToastItem item = new ToastItem(this.PostData, this.IconPath);

			// get the toast title
			item.ToastInfo = this.View.Info;
			item.ToastTitle = string.Empty;
			item.AppName = this.AppName;

			if (!string.IsNullOrEmpty(this.View.Title))
			{
				if (item.ToastTitle.IndexOf(this.View.Title, 0, StringComparison.InvariantCultureIgnoreCase) != 0)
				{
					item.ToastTitle += " " + this.View.Title;
				}
			}

			// get the toast content
			if (this.View.InboxTexts.Length >= 1)
			{
				// inboxed style may support "text", the last(bottom) text of the notification
				if (string.IsNullOrEmpty(this.View.BottomText))
				{
					item.ToastContent = new string[this.View.InboxTexts.Length];
				}
				else
				{
					item.ToastContent = new string[this.View.InboxTexts.Length + 1];
					item.ToastContent[this.View.InboxTexts.Length] = this.View.BottomText;
				}

				for (int index = 0; index < this.View.InboxTexts.Length; ++index)
				{
					item.ToastContent[index] = this.View.InboxTexts[index];
				}
			}

			return item;
		}
	}

	/// <summary>
	/// Represents the basic style of notification.
	/// </summary>
	internal sealed class NotificationContentBasicStyle : NotificationStyle
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationContentBasicStyle"/> class.
		/// </summary>
		public NotificationContentBasicStyle()
			: base()
		{
		}

		/// <summary>
		/// Gets the toast element from the basic-style notification.
		/// </summary>
		/// <returns>An instance of the <see cref="ToastItem"/> class.</returns>
		public override ToastItem GetContent()
		{
			ToastItem item = new ToastItem(this.PostData, this.IconPath);

			// get the toast title
			item.ToastInfo = this.View.Info;
			item.ToastTitle = string.Empty;
			item.AppName = this.AppName;

			if (!string.IsNullOrEmpty(this.View.Title))
			{
				if (item.ToastTitle.IndexOf(this.View.Title, 0, StringComparison.InvariantCultureIgnoreCase) != 0)
				{
					item.ToastTitle += " " + this.View.Title;
				}
			}

			// get the toast content
			// other styles may support "text2", "text" as the toast content
			bool middleTextNull = string.IsNullOrEmpty(this.View.MiddleText);
			bool bottomTextNull = string.IsNullOrEmpty(this.View.BottomText);

			if (!middleTextNull && !bottomTextNull)
			{
				item.ToastContent = new string[2];
				item.ToastContent[0] = this.View.MiddleText;
				item.ToastContent[1] = this.View.BottomText;
			}
			else if (!middleTextNull || !bottomTextNull)
			{
				item.ToastContent = new string[1];
				item.ToastContent[0] = middleTextNull ? this.View.BottomText : this.View.MiddleText;
			}

			return item;
		}
	}

	/// <summary>
	/// Represents the remove-view style of notification.
	/// </summary>
	internal sealed class NotificationContentRemoteViewStyle : NotificationStyle
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationContentRemoteViewStyle"/> class.
		/// </summary>
		public NotificationContentRemoteViewStyle()
			: base()
		{
		}

		/// <summary>
		/// Gets the toast element from the remote-view-style notification.
		/// </summary>
		/// <returns>An instance of the <see cref="ToastItem"/> class.</returns>
		public override ToastItem GetContent()
		{
			ToastItem item = new ToastItem(this.PostData, this.IconPath);
			item.AppName = this.AppName;

			// get the toast title
			bool noDefaultAppTitle = false;

			if (string.IsNullOrEmpty(this.AppName))
			{
				noDefaultAppTitle = true;
				if (this.View.CustomTexts.Length >= 1)
				{
					item.ToastTitle = this.View.CustomTexts[0];
				}
			}
			else
			{
				item.ToastTitle = string.Empty;
			}

			// get the toast content
			if (noDefaultAppTitle)
			{
				if (this.View.CustomTexts.Length >= 2)
				{
					item.ToastContent = new string[this.View.CustomTexts.Length - 1];
					for (int index = 1; index < this.View.CustomTexts.Length; ++index)
					{
						item.ToastContent[index - 1] = this.View.CustomTexts[index];
					}
				}
			}
			else
			{
				if (this.View.CustomTexts.Length >= 1)
				{
					item.ToastContent = new string[this.View.CustomTexts.Length];
					for (int index = 0; index < this.View.CustomTexts.Length; ++index)
					{
						item.ToastContent[index] = this.View.CustomTexts[index];
					}
				}
			}

			return item;
		}
	}
}