﻿using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Manager.Notification
{
	/// <summary>
	/// Represents a quick action item
	/// </summary>
	public class QuickActionItem
	{
		/// <summary>
		/// Gets the it of the quick action
		/// </summary>
		public int? QAIndex { get; internal set; }

		/// <summary>
		/// Gets the title of the quick action
		/// </summary>
		public string QATitle { get; internal set; }

		/// <summary>
		/// Gets if the action intent will pop-up another display
		/// </summary>
		public bool QAActivityPopup { get; internal set; }

		/// <summary>
		/// Gets the icon path of the quick action
		/// </summary>
		public string QAIconPath { get; internal set; }

		/// <summary>
		/// Initialize a new instance of the <see cref="QuickActionItem"/> class.
		/// </summary>
		internal QuickActionItem()
		{
			this.QAIndex = null;
			this.QATitle = String.Empty;
			this.QAIconPath = null;
			this.QAActivityPopup = false;
		}
	}

	/// <summary>
	/// Specifies the type of the notification
	/// </summary>
	public enum NotificationType
	{
		/// <summary>
		/// Basic type with only 'ignore', 'dismiss', 'quick actions', or 'local launch' buttons
		/// </summary>
		Basic = 1,

		/// <summary>
		/// Type with only 'ignore', 'dismiss', 'launch', 'quick actions', or 'local launch' buttons
		/// </summary>
		BasicWithLaunch = 2,

		/// <summary>
		/// Type with only 'Open browser' button, used when sharing the website link from phone device
		/// </summary>
		UrlShare = 3,

		/// <summary>
		/// Type with only 'Open Folder' and 'Open File' button, used when sharing the file from phone device
		/// </summary>
		FileShare = 4,

		/// <summary>
		/// Type with only 'Open browser' button, used when sharing the website link from phone device
		/// </summary>
		UrlShareWithLaunch = 5,

		/// <summary>
		/// Type with only 'Open Folder' and 'Open File' button, used when sharing the file from phone device
		/// </summary>
		FileShareWithLaunch = 6,
	}

	/// <summary>
	/// Represents a toast item
	/// </summary>
	public class ToastItem
	{
		/// <summary>
		/// Gets the path of the cached icon.
		/// </summary>
		public string IconPath { get; internal set; }

		/// <summary>
		/// Gets the app name of the notification.
		/// </summary>
		public string AppName { get; internal set; }

		/// <summary>
		/// Gets the title of the toast.
		/// </summary>
		public string ToastTitle { get; internal set; }

		/// <summary>
		/// Gets the content of the toast.
		/// </summary>
		public string[] ToastContent { get; internal set; }

		/// <summary>
		/// Gets the info of the toast.
		/// </summary>
		public string ToastInfo { get; internal set; }

		/// <summary>
		/// Gets the quick action elements
		/// </summary>
		public List<QuickActionItem> QAItems { get; internal set; }

		/// <summary>
		/// Gets the type of the toast: Basic, file share, url share
		/// </summary>
		public NotificationType Type { get; internal set; }

		/// <summary>
		/// Check if the notification is media style
		/// </summary>
		public bool isMediaStyle { get; internal set; }

		/// <summary>
		/// Check if the notification is allow to be muted
		/// </summary>
		public bool isAllowMute { get; internal set; }

		/// <summary>
		/// Gets a value that indicates whether the notification intent could popup another display.
		/// </summary>
		public bool? ActivityPopup { get; internal set; }

		/// <summary>
		/// Gets a value that indicates whether the notification is removable.
		/// </summary>
		public bool Removable { get; internal set; }

		/// <summary>
		/// Gets a value which indicates whether the notification is launchable on local pc.
		/// </summary>
		public LocalLaunch LocalLaunchItem { get; internal set; }

		/// <summary>
		/// Gets the instance ID of the source device that posted the notification.
		/// </summary>
		public Guid? DeviceInstanceId { get; internal set; }

		/// <summary>
		/// Initialize a new instance of the <see cref="ToastItem"/> class.
		/// </summary>
		internal ToastItem()
		{
			this.IconPath = String.Empty;
			this.AppName = String.Empty;
			this.ToastTitle = String.Empty;
			this.DeviceInstanceId = null;
			this.ActivityPopup = null;
			this.Removable = false;
			this.ToastContent = null;
			this.ToastInfo = null;
			this.QAItems = null;
			this.LocalLaunchItem = null;
			this.isMediaStyle = false;
			this.isAllowMute = true;
		}

		/// <summary>
		/// Initialize a new instance of the <see cref="ToastItem"/> class.
		/// </summary>
		/// <param name="postData">The received post data</param>
		/// <param name="iconPath">The notification icon path</param>
		internal ToastItem(NotificationPostAdapter.NotificationPostInfo postData, string iconPath)
		{
			this.IconPath = iconPath;
			this.AppName = null;
			this.DeviceInstanceId = postData.DeviceInstanceID;
			this.ActivityPopup = postData.ActivityPopup;
			this.Removable = postData.Removable.Value;
			this.ToastContent = null;
			this.ToastTitle = null;

			if (postData.BigContentView != null &&
				postData.BigContentView.Style != null &&
				postData.BigContentView.Style.Value <= 8 &&
				postData.BigContentView.Style.Value >= 6)
			{
				this.isMediaStyle = true;
			}

			if (postData.QuickActions != null)
			{
				this.QAItems = new List<QuickActionItem>();
				int index = 0;
				foreach (NotificationPostAdapter.NotificationQuickActionContent qa in postData.QuickActions)
				{
					QuickActionItem item = new QuickActionItem();
					item.QAIndex = index++;
					item.QATitle = qa.QATitle;
					item.QAActivityPopup = (qa.QAActivityPopup != null) ? qa.QAActivityPopup.Value : false;
					item.QAIconPath = NotificationHelper.GetIconCacheFolder() + @"\" + qa.QAIcon + ".png";

					/*
					byte[] byteBuffer = Convert.FromBase64String(qa.QAIcon);
					using (MemoryStream ms = new MemoryStream(byteBuffer))
					{
						ms.Position = 0;
						item.QABitmap = (Bitmap)Bitmap.FromStream(ms);
						byteBuffer = null;
					}
					*/

					this.QAItems.Add(item);
				}
			}

			//Get the notification type
			this.Type = NotificationType.Basic;
			this.isAllowMute = true;

			if (!string.IsNullOrEmpty(postData.PackageName)
				&& postData.PackageName.Equals(Definitions.ACEREXTEND_PHONE_PACKAGENAME, StringComparison.InvariantCultureIgnoreCase))
			{
				this.isAllowMute = false;

				if (postData.Id != null)
				{
					if (postData.Id.Value == Definitions.NOTIFICATION_ID_SHARING_LINK) //Open Windows Browser
					{
						this.Type = NotificationType.UrlShare;

						if (this.ActivityPopup != null && this.ActivityPopup.Value == true)
						{
							this.Type = NotificationType.UrlShareWithLaunch;
						}
					}
					else if (postData.Id.Value ==
						(Definitions.NOTIFICATION_ID_OPEN_REMOTE_DOWNLOADS | Definitions.NOTIFICATION_ID_OPEN_REMOTE_UPLOADED_FILE)) //Open download folder & download file
					{
						this.Type = NotificationType.FileShare;

						if (this.ActivityPopup != null && this.ActivityPopup.Value == true)
						{
							this.Type = NotificationType.FileShareWithLaunch;
						}
					}
					else
					{
						if (this.ActivityPopup != null && this.ActivityPopup.Value == true)
						{
							this.Type = NotificationType.BasicWithLaunch;
						}
					}
				}
			}
			else if (this.ActivityPopup != null && this.ActivityPopup.Value == true)
			{
				this.Type = NotificationType.BasicWithLaunch;
			}

			//Compare Local List
			this.LocalLaunchItem = LocalLaunchFactory.GetLocalLaunchHandler(this.Type, postData);
		}
	}
}