﻿using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of notification ignore (PC => phone)
	/// (PUT /notification/ignore/{Package_name})
	/// </summary>
	internal sealed class NotificationIgnoreAdapter : WebServiceAdapter<DBNull, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification/ignore";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationIgnoreAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public NotificationIgnoreAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationIgnoreAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		/// <param name="timeout">The custom web service timeout.</param>
		public NotificationIgnoreAdapter(WebServicesManager webServiceManager, TimeSpan? timeout)
			: base(webServiceManager, timeout)
		{
		}

		/// <summary>
		/// Sends a notification ignore request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="extendIgnoreRequestUri">
		/// The required forbidden string to specify which notification should be forbidden forever
		/// The request string should follow the format
		/// /notification/ignore/{Package_name}.
		/// </param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestIgnoreAsync(IPAddress address, ushort port, string extendIgnoreRequestUri)
		{
			// Seoker comment: the request URI should be composed in this object, the responsibility doesn't belong to upper-layer.

			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + extendIgnoreRequestUri).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
