﻿using AirFusion.Common;
using System;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService
{
	/// <summary>
	/// Extensions.
	/// </summary>
	internal static partial class Extensions
	{
		/// <summary>
		/// Determines whether the specified status code is success.
		/// </summary>
		/// <param name="statusCode">The status code.</param>
		/// <returns>True if the status code indicates success.</returns>
		public static bool IsSuccess(this HttpStatusCode statusCode)
		{
			return (int)statusCode >= 200 && (int)statusCode < 300;
		}

		/// <summary>
		/// Gets the presentation of the specified HTTP content.
		/// </summary>
		/// <param name="content">The content.</param>
		/// <returns>The string presentation.</returns>
		public static async Task<string> GetPresentationAsync(this HttpContent content)
		{
			if (content is StringContent)
			{
				var text = await (content as StringContent).ReadAsStringAsync();
				if (!string.IsNullOrEmpty(text))
				{
					return text.GetSubString();
				}
			}
			else if (content is ByteArrayContent)
			{
				var byteArrayContent = await (content as ByteArrayContent).ReadAsByteArrayAsync();
				if (byteArrayContent != null)
				{
					return byteArrayContent.Length > 32 ? byteArrayContent.ToHexString(0, 32) : byteArrayContent.ToHexString();
				}
			}
			else if (content is StreamContent)
			{
				return "<<stream>>";
			}
			else if (content is MultipartContent)
			{
				var builder = new StringBuilder();
				builder.AppendLine("<<Multipart>>");
				int index = 0;
				foreach (var c in (content as MultipartContent))
				{
					builder.AppendLine("[" + (index + 1) + "]\r\n" + (await c.GetPresentationAsync()));
					index++;
				}

				return builder.ToString().Trim();
			}

			return "<<N/A>>";
		}

		/// <summary>
		/// Gets the presentation of the specified headers.
		/// </summary>
		/// <param name="response">The headers.</param>
		/// <returns>The string presentation.</returns>
		public static string GetPresentation(this HttpResponseMessage response)
		{
			// check parameters
			if (response == null)
			{
				return string.Empty;
			}

			var builder = new StringBuilder();
			builder.AppendLine("HttpResponseMessage.Version: " + (response.Version != null ? response.Version.ToString() : "N/A"));
			builder.AppendLine("HttpResponseMessage.StatusCode: " + (int)response.StatusCode + " " + response.ReasonPhrase);
			builder.AppendLine("HttpResponseMessage.Headers: " + (response.Headers != null ? "\r\n" + response.Headers.ToString() : "N/A"));

			return builder.ToString().Trim();
		}

		/// <summary>
		/// Gets the presentation of the specified headers.
		/// </summary>
		/// <param name="response">The headers.</param>
		/// <returns>The string presentation.</returns>
		public static async Task<string> GetPresentationAsync(this HttpResponseMessage response)
		{
			// check parameters
			if (response == null)
			{
				return string.Empty;
			}

			var builder = new StringBuilder();
			builder.AppendLine("HttpResponseMessage.Version: " + (response.Version != null ? response.Version.ToString() : "N/A"));
			builder.AppendLine("HttpResponseMessage.StatusCode: " + (int)response.StatusCode + " " + response.ReasonPhrase);
			builder.AppendLine("HttpResponseMessage.Headers: " + (response.Headers != null ? "\r\n" + response.Headers.ToString() : "N/A"));
			builder.AppendLine("HttpResponseMessage.Content: " + (response.Content != null ? "\r\n" + (await response.Content.GetPresentationAsync()) : "N/A"));

			return builder.ToString().Trim();
		}
	}
}
