// This header is used by both the DLL and the module
#ifndef VNCHANDLER_H__
#define VNCHANDLER_H__

#pragma once

#ifdef VNCHANDLER_EXPORTS
#define VNCAPI __declspec(dllexport)
#else
#define VNCAPI __declspec(dllimport)
#endif

// 
#ifdef __cplusplus
#   define EXTERN_C     extern "C"
#else
#   define EXTERN_C
#endif // __cplusplus

#include "res/resource.h"
#include "ClientConnection.h"
#include "Exception.h"

#define IPV4_ADDRESS_SIZE 16
#define PASSWORD_SIZE 32

extern "C"
{
	typedef void (__stdcall * VNCMessageCallback)(int vnc_id, int vnc_msg_type, int vnc_fun_code, int vnc_err_code, char* msg);
	VNCAPI void WINAPI SetVNCCallbackFun(VNCMessageCallback callbackPointer);

	typedef void (__stdcall * FileUploadProCallback)(int vnc_id, int file_id, int progress, int state);
	typedef void (__stdcall * FileDownloadProCallback)(int vnc_id, wchar_t* file_name, int progress, int state);

	VNCAPI void WINAPI SetFileTransferCallbackFun(FileUploadProCallback uploadCallback, FileDownloadProCallback downloadCallback);

	VNCAPI void WINAPI SetAFMiracastScreenInfo(BYTE bIndex, int iWidth, int iHeight, int iDeltaX, int iDeltaY, float fDpiRate);

	VNCAPI void WINAPI SetDownloadFolderPath(BYTE bIndex, wchar_t* folder_path);

	VNCAPI void WINAPI EnableDebugLog(char* logfile_path, int iLogLevel);
	VNCAPI void WINAPI WriteDebugLog(int iLogLevel, char* sDebugMsg);
	VNCAPI void WINAPI CloseDebugLog();
}

struct VNCConnectInfo
{
	char sIPv4[IPV4_ADDRESS_SIZE];
	char sPassword[PASSWORD_SIZE];
	unsigned int uiPortNum;
};

enum VNCMsgType
{
	MT_CommonMessage,
	MT_ErrorMessage
};

enum VNCFunctionCode
{
	FC_InitialVNC = 0,
	FC_StartRunVNC = 1,
	FC_HandleWinMsg = 2,
	FC_SendMouseInput = 3,
	FC_ClipboardText = 4,
	FC_FileUpload = 5,
	FC_FileUPloadCancel = 6,
	FC_FileDownload = 7,
	FC_FileTransferProc = 8,
	FC_SendIMEString = 9,
	FC_SendKeyboard = 10,

	FC_ServerRfbMsgHandler = 11,
	FC_ColourMapEntries = 12
};

enum VNCErrorCode
{
	EC_NoError = 0,
	EC_FailedConnect = 1,				// Connecting to VNC server was failed for unknow reason
	EC_FailedConnect_SocketClose = 2,	// Windows Socket Error = 10038 (WSAENOTSOCK => Socket operation on nonsocket.)
	EC_FailedConnect_InvalidIP = 3,		// Windows Socket Error = 10049 (WSAEADDRNOTAVAIL => Cannot assign requested address.)
	EC_FailedConnect_TimeOut,		// Windows Socket Error = 10060 (WSAETIMEDOUT => Connection timed out.)
	EC_FailedConnect_NoServer,		// Windows Socket Errpr = 10061 (WSAECONNREFUSED => Connection refused.)
	EC_FailedConnect_NoNetwork,		// Windows Socket Error = 10065 (WSAEHOSTUNREACH => No route to host.)
	EC_ConnectionClosed,
	EC_InvalidRFBProtocol,
	EC_RFBMessageError,
	EC_InvalidVncSecurity,
	EC_SocketError,
	EC_CreateFileFailed,
	EC_VNCStopRunning,
	EC_InvalidFileSize,
	EC_OtherError
};


/// <summary>
/// Create one new VNCHandler
/// </summary>
/// <returns>
/// -1: Failed to create VNCHandler
/// >0: VNC Index (unique key, for support multi connection)
/// </returns>
EXTERN_C VNCAPI int WINAPI CreateNewVNCHandler();

/// <summary>
/// Ask the specified VNC Handle to enter running mode (Make connection with server & monitor the commands that are received from server)
/// </summary>
/// <param name="bIndex">The index of VNC Handler</param>
/// <param name="mConnInfo">Making connection needs information (i.e. IP, Port, and PW)</param>
/// <returns>True: Success; False: Failed</returns>
EXTERN_C VNCAPI DWORD WINAPI RunVNCHandler(BYTE bIndex, VNCConnectInfo mConnInfo);

/// <summary>
/// Initialize the Clipboard function (Set UI handle)
/// </summary>
/// <returns>True: Successfully; False: Failed</returns>
EXTERN_C VNCAPI bool WINAPI InitVNCClipboard(BYTE bIndex, HWND hWnd);

/// <summary>
/// Removes the hWnd from Clipboard Chain (hWnd is UI window handle)
/// </summary>
EXTERN_C VNCAPI void WINAPI RemoveVNCClipboard(HWND hWnd);

/// <summary>
/// This function is used for handling the windows message of WPF UI
/// </summary>
EXTERN_C VNCAPI void WINAPI VncWinProc(BYTE bIndex, HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);

/// <summary>
/// Send the unicode chars to Server
/// </summary>
/// <param name="wcBuffer">Unicode string</param>
/// <param name="iWsLeng">The length of unicode string</param>
EXTERN_C VNCAPI void WINAPI SendIMEChars(BYTE bIndex, wchar_t* wsBuffer, int iWsLeng);

/// <summary>
/// Send a virtual key code to VNC Server
/// </summary>
/// <param name="bIndex"></param>
/// <param name="uiVirtualKeyCode">Virtual Key Code (ex: Backspace = VK_BACK = 0x08)</param>
/// <param name="uiKeyStateFlg">KEYEVENTF_KEYDOWN = 0x0000, KEYEVENTF_EXTENDEDKEY = 0x0001, KEYEVENTF_KEYUP = 0x0002,</param>
EXTERN_C VNCAPI void WINAPI SendVirtualKey(BYTE bIndex, unsigned int uiVirtualKeyCode, unsigned int uiKeyStateFlg);

/// <summary>
/// Sends the touch input event (Down/Up/Move) to VNC Server
/// </summary>
/// <param name="bIndex">VNC Client ID</param>
/// <param name="uiTouchEventType">TouchEventType.Download, TouchEventType.Move, TouchEventType.Up</param>
/// <param name="point_x">location X coordinate value</param>
/// <param name="point_y">location Y coordinate value</param>
EXTERN_C VNCAPI void WINAPI SendTouchControl(BYTE bIndex, unsigned int uiTouchEventType, long point_x, long point_y);

/// <summary>
/// Ask VNC Handler to send one local file to Server (ex: Smart Phone)
/// </summary>
/// <param name="bIndex">VNC index</param>
/// <param name="wsClientFilePath">The path of file that will be sent to server</param>
/// <param name="iFileId"></param>
EXTERN_C VNCAPI void WINAPI TransferFileToServer(BYTE bIndex, wchar_t* wsClientFilePath, int iFileId);

/// <summary>
/// Close and release one VNCHandler (according to the value of VNC Index)
/// </summary>
EXTERN_C VNCAPI void WINAPI CloseVNCHandler(int iVncIndex);

/// <summary>
/// Close and release the all VNCHandlers
/// </summary>
EXTERN_C VNCAPI void WINAPI ReleaseVNCHandlers();

/// <summary>
/// Right-clik to paste the clipboard
/// </summary>
/// <param name="bIndex">VNC index</param>
EXTERN_C VNCAPI void WINAPI RightClickClipboard(BYTE bIndex);

#endif	// VNCHANDLER_H__
